#!/usr/bin/perl
use Scalar::Util qw(looks_like_number);
use Getopt::Long qw(GetOptions);
use List::Util qw[min];
binmode (STDOUT,":utf8");

# Program AIS_menu
# (c) Gary C. Kessler, 2018-2024

## See README.TXT for background commentary and version information. ##

$build_date = "04/04/2024";
$version = "4.8.9";

# Set default value for auto-mode parameter and then see if the switch is provided

$auto = 0;

GetOptions
  (
  "auto" => \$auto,
  );

#####
# STEP 1: Determine the message type and initialize the $command command string

do
  {
  print "\033[2J";    # Clear the screen
  print "\033[0;0H";  # Place cursor at 0,0 (upper left-hand corner)
  print "\n";
  print "                AIS Sentence Preprocessor (Build: $build_date Version: $version)\n\n";
  print "                                   MENU\n\n";
  print "  Type  1: Position Report Class A\n";
  print "  Type  2: Position Report Class A (Assigned Schedule)\n";
  print "  Type  3: Position Report Class A (Response to Interrogation)\n";
  print "  Type  4: Base Station Report\n";
  print "  Type  5: Static and Voyage Related Data\n";
  print "  Type  6: Binary Addressed Message\n";
  print "  Type  7: Binary Acknowledge Message\n";
  print "  Type  8: Binary Broadcast Message\n";
  print "  Type  9: Standard SAR Aircraft Position Report\n";
  print "  Type 10: UTC/Date Inquiry\n";
  print "  Type 11: UTC/Date Response\n";
  print "  Type 12: Addressed Safety-Related Message\n";
  print "  Type 13: Safety-Related Acknowledgement\n";
  print "  Type 14: Safety-Related Broadcast Message\n";
  print "  Type 15: Interrogation Message\n";
  print "  Type 16: Assignment Mode Command Message\n";
  print "  Type 17: DGNSS Broadcast Binary Message\n";
  print "  Type 18: Standard Class B CS Position Report\n";
  print "  Type 19: Extended Class B CS Position Report\n";
  print "  Type 20: Data Link Management Message\n";
  print "  Type 21: Aid-to-Navigation Report\n";
  print "  Type 22: Channel Management\n";
  print "  Type 23: Group Assignment Command\n";
  print "  Type 24: Static Data Report\n";
  print "  Type 25: Single Slot Binary Message\n";
  print "  Type 26: Multiple Slot Binary Message With Communications State\n";
  print "  Type 27: Long Range AIS Broadcast Message\n";
  print "\n";
  print "  X. Exit\n\n";
  print "Enter message type (1-27) or 'X' to halt: ";
  chomp ($type = <STDIN>);
  $type = uc ($type);
  }
  until (($type >= 1 && $type <= 27) || $type eq "X");

# If type = "X", exit the program

if ($type eq "X")
  {
  print "Goodbye!\n\n";
  exit;
  };

$command = "perl AIS_ping.pl --type=" . int($type) . " ";

#####
# STEP 2: Determine the talker ID and AIS encapsulation formatter to form the sentence type.
# Default is "!AIVDM"

@talker_id = ("","!AB","!AD","!AI","!AN","!AR","!AS","!AT","!AX","!SA");
@encap_fmt = ("","VDM","VDO");

do
  {
  print "\nEnter talker ID (1-9) from the following list:\n";
  print "  1. !AB - NMEA 4.0 AIS Base Station           2. !AD - NMEA 4.0 Dependent AIS Base Station\n";
  print "  3. !AI - Mobile AIS Station [default]        4. !AN - NMEA 4.0 Aid to Navigation AIS Station\n";
  print "  5. !AR - NMEA 4.0 AIS Receiving Station      6. !AS - NMEA 4.0 Limited Base Station\n";
  print "  7. !AT - NMEA 4.0 AIS Transmitting Station   8. !AX - NMEA 4.0 Repeater AIS Station\n";
  print "  9. !SA - NMEA 4.0 Physical Shore AIS Station\n";
  print " Default = 3: ";
  chomp ($talker = <STDIN>);
  }
  until ((looks_like_number ($talker) && $talker >= 1 && $talker <= 9) || $talker eq ""); 
if ($talker eq "") { $talker = 3; }

do
  {
  print "\nEnter AIS encapsulation formatter (1-2) from the following list:\n";
  print "  1. VDM - AIS VHF Data-Link Message [default]\n";
  print "  2. VDO - AIS VHF Data-Link Own-Vessel Report\n";
  print "Default = 1: ";
  chomp ($encap = <STDIN>);
  }
  until ((looks_like_number ($encap) && $encap >= 1 && $encap <= 2) || $encap eq ""); 
if ($encap eq "") { $encap = 1; }

$ais_sent = $talker_id [$talker] . $encap_fmt [$encap];
if ($ais_sent ne "!AIVDM") { $command .= "--ais_sent='" . $ais_sent . "' "; }

#####
# STEP 3: All messages contain the sender's Maritime Mobile Service Identity (MMSI)

do
  {
  print "\nThe sender's 9-digit Maritime Mobile Service Identity (MMSI) is required.\n In general, a valid MMSI begins with the digits 201-775. See\n";
  print " https://www.navcen.uscg.gov/?pageName=mtmmsi for more details.\n\nEnter MMSI (9 decimal digits): ";
  chomp ($mmsi = <STDIN>);
  }
  until (verify_numeric_string ($mmsi,9));

$command .= "--mmsi=" . $mmsi . " ";

#####
# STEP 4: Step through each message type and get the necessary parameters, building the $command command string
# Each routine will ask for values of those parameters used by the requested message type and will
# provide the default value. If default values are selected, they are not added to $command

# PROGRAMMING NOTE: In the routines below, there is a lot of use of a variable named $x. In most of the program,
# $x represents a value that is used one time in a localized section of the code. Rather than name every instance
# with a more descriptive name, I save a lot of memory (yeah, I know, who cares nowadays?) by calling everything $x.
# The use of the value can be found in the context of the value being requested and the $command string switch label.

print "\nSee The 'AIVDM/AIVDO protocol decoding' page at https://gpsd.gitlab.io/gpsd/AIVDM.html\n";
print "and relevant standards for additional information on selecting parameter values.\n";

BUILD_MSG:
  {
  $type >= 1 && $type <= 3 && do
    {
    p_navstat ();
    p_rot ();
    p_cog ();
    p_hdg ();
    p_sog ();
    p_position ();
    p_accuracy ();
    p_timestamp ();
    p_raim ();
    last BUILD_MSG;
    };
  ($type == 4 || $type == 11) && do
    {
    p_date (4);
    p_position ();
    p_accuracy ();
    p_epfd ();
    p_raim ();
    last BUILD_MSG;
    };
  $type == 5 && do
    {
    p_epfd ();
    p_type05 ();
    p_dte ();
    p_vname ();
    p_callsign ();
    p_vtype ();
    p_dimensions ();
    last BUILD_MSG;
    };
  ($type == 6 || $type == 8) && do
    {
    type0608_handler ();
    last BUILD_MSG;
    };
  ($type == 7 || $type == 13) && do
    {
    p_mmsi_seq4 ();
    last BUILD_MSG;
    };
  $type == 9 && do
    {
    p_position ();
    p_accuracy ();
    p_cog ();
    p_alt_speed ();
    p_dte ();
    p_timestamp ();
    p_raim ();
   last BUILD_MSG;
    };
  $type == 10 && do
    {
    p_mmsi_seq (0);
    last BUILD_MSG;
    };
  $type == 12 && do
    {
    p_mmsi_seq (1);
    p_sart_msg ();
    last BUILD_MSG;
    };
  $type == 14 && do
    {
    p_sart_msg ();
    last BUILD_MSG;
    };
  $type == 15 && do
    {
    type15_handler ();
    last BUILD_MSG;
    };
  $type == 16 && do
    {
    type16_handler ();
    last BUILD_MSG;
    };
  $type == 17 && do
    {
    p_position ();
    p_dgnss_msg ();
    last BUILD_MSG;
    };
  $type == 18 && do
    {
    p_cog ();
    p_hdg ();
    p_sog ();
    p_position ();
    p_accuracy ();
    p_timestamp ();
    p_raim ();
    last BUILD_MSG;
    };
  $type == 19 && do
    {
    p_cog ();
    p_hdg ();
    p_sog ();
    p_position ();
    p_accuracy ();
    p_timestamp ();
    p_vname ();
    p_vtype ();
    p_dimensions ();
    p_epfd ();
    p_raim ();
    p_dte ();
    last BUILD_MSG;
    };
  $type == 20 && do
    {
    p_fatdma ();
    last BUILD_MSG;
    };
  $type == 21 && do
    {
    p_aton ();
    p_accuracy ();
    p_position ();
    p_dimensions ();
    p_epfd ();
    p_timestamp ();
    p_offposition ();
    p_aid_status ();
    p_raim ();
    last BUILD_MSG;
    };
  $type == 22 && do
    {
    p_channel ();
    p_ais_region ();
    last BUILD_MSG;
    };
  $type == 23 && do
    {
    p_vtype ();
    p_ais_region ();
    p_interval ();
    last BUILD_MSG;
    };
  $type == 24 && do
    {
    type24_handler ();
    last BUILD_MSG;
    };
  ($type == 25 || $type == 26) && do
    {
    type2526_handler ();
    last BUILD_MSG;
    };
  $type == 27 && do
    {
    p_navstat ();
    p_position ();
    p_accuracy ();
    p_raim ();
    p_type27 ();
    last BUILD_MSG;
    };
  };

#####
# STEP 5: Get parameters for the final part of the command string to output. First determine
# if we want to produce output as an NMEA string (--nmea) or for AiS_TX (--tx)

do
  {
  print "\nSet command for NMEA output (N) or AiS_TX output (T) [default = N]: ";
  chomp ($out = <STDIN>);
  $out = uc ($out);
  }
  until ($out eq "N" || $out eq "T" || $out eq "");
if ($out eq "") { $out = "N"; }

# Select AIS channel A or B

do
  {
  print "\nEnter AIS channel (A/B) [default = A]: ";
  chomp ($channel = <STDIN>);
  $channel = uc ($channel);
  }
  until ($channel eq "A" || $channel eq "B" || $channel eq "");
if ($channel eq "") { $channel = "A"; }

# If creating NMEA input for AIS_ping, check to see if we're also in auto-mode
# auto_mode means that AIS_menu is running as part of a shell script, so output is directed to a temp file

if ($out eq "N")
    {
    $command .= "--nmea=" . $channel;
    if ($auto)
      {
      open (TMPFILE, ">", "ais_tmp_file.sh");
      print TMPFILE "$command --auto\n";
      close TMPFILE;
      }
    }
  else
    { $command .= "--tx=" . $channel;}

print "\n$command\n\n";

# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# type_0608_22_23
# ********************************************************

# Routine to get message-specific information for:
#   Type 6, DAC 1, FI 23 and DAC 367, FI 22
#   Type 8, DAC 1, FI 22 and DAC 367, FI 22

sub type_0608_22_23
  {
  my ($dac) = @_;
  my ($i, $sa, $sa_num, $x);

# There are a few differences between the DAC=001 (IMO) version and DAC=367 (USCG) version

  my ($max_distance, $num_chars, $num_subareas);
  
  if ($dac == 1)
      { $max_distance = 1023; $num_chars = 14; $num_subareas = 10; }
    else
      { $max_distance = 2047; $num_chars = 15; $num_subareas = 9; }

  do
    {
    print "\nEnter Notice Description (0-127):\n";
    print "   0- 21: Caution Area messages          23- 30: Environmental Caution Area messages\n";
    print "  32- 38: Restricted Area messages       40- 45: Anchorage Area messages\n";
    print "  56- 58: Security Alert messages        64- 76: Distress Area messages\n";
    print "  80- 85: Instruction messages           88- 95: Information messages\n";
    print "  96-108: Chart Feature messages        112-114: Report From Ship messages\n";
    print " 120-122: Route messages                125-126: Other messages\n";
    print "Default = 127 (Undefined): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 127) || $x eq ""); 
  if ($x ne "")
    { $command .= "--notice=" . int($x) . " "; }

  do
    {
    print "\nEnter Duration (1 to 262142 minutes [182.04 days]; 0 = cancel,\ndefault = 262143 [undefined]): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 262143) || $x eq "");
  if ($x ne "")
    { $command .= "--duration=" . int($x) . " "; }

# This message can include up to 10 (DAC 1) or 9 sub-areas (DAC 367)
# In general, the sub-areas have different formats and have between 2-10 fields per sub-area
# The parameters for each sub-area are in a comma-delimited string in switch subNN

  for ($sa_num=1; $sa_num<=$num_subareas; $sa_num++)
    {
    do
      {
      printf "\nEnter sub-area %d type (0-5): \n", $sa_num;
      print " 0. Circle/Point   1. Rectangle   2. Sector\n";
      print " 3. Polyline       4. Polygon     5. Associated text\n";
      print "Hit <ENTER> when done: ";
      chomp ($sa = <STDIN>);
      }
      until ((looks_like_number ($sa) && $sa >= 0 && $sa <= 5) || $sa eq ""); 

    if ($sa ne "")
        { $command .= "--sa" . sprintf ("%02d", $sa_num) . "=" . int($sa) ; }
      else
        { last; }

# For sub-area types 0-4, the next parameter is the Scale Factor

  if ($sa <= 4)
    {
    do
      {
      print "\nEnter Scale Factor (n=0-3) where factor is 10^n. Values 0 (default), 1, 2, and 3\n";
      print " correspond to scale factors 1, 10, 100, and 1000, respectively: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 3) || $x eq "");

    if ($x eq "")
      { $x = 0; }
    $command .= "," . int($x) ;
    }

# For sub-area types 0-2, the next three parameters are Longitude, Latitude, and Precision

  if ($sa <= 2)
    {
    print "\nLatitude and longitude are of the center point of a circle or sector, and the\n";
    print " southwest corner of a box/rectangle\n";
    do
      {
      print "\nEnter longitude (-180 to 180; default = 181 [not available]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= -180 && $x <= 181) || $x eq ""); 

    if ($x eq "")
      { $x = 181; }
    $command .= "," . $x ;

    do
      {
      print "\nEnter latitude (-90 to 90; default = 91 [not available]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= -90 && $x <= 91) || $x eq ""); 

    if ($x eq "")
      { $x = 91; }
    $command .= "," . $x ;

    do
      {
      print "\nEnter precision (0-4), number of decimal places for lat/long [default = 4]: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 4) || $x eq "");

    if ($x eq "")
      { $x = 4; }
    $command .= "," . int($x) ;
    }

# Finish up the rest of the parameters for the rest of the sub-areas

    SUBAREA_ENCODE:
      {

# Circle or point

      $sa == 0 && do
        {
        do
          {
          print "\nEnter Radius (0-4095 meter steps; default = 0 [single point]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 4095) || $x eq "");
        if ($x eq "")
          { $x = 0; }
        $command .= "," . int($x) ;

        last SUBAREA_ENCODE;
        };

# Rectangle

      $sa == 1 && do
        {
        do
          {
          print "\nEnter box dimension east from the corner point (1-255 meter steps;\n default = 0 [north-south line]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 255) || $x eq "");
        if ($x eq "") 
          { $x = 0; }
        $command .= "," . int($x) ;

        do
          {
          print "\nEnter box dimension north from the corner point (1-255 meter steps;\n default = 0 [east-west line]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 255) || $x eq "");
        if ($x eq "")
          { $x = 0; }
        $command .= "," . int($x) ;

        do
          {
          print "\nEnter orientation of box around the southwest corner (1-359 degree steps;\n default = 0 [no rotation]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 359) || $x eq "");
        if ($x eq "")
          { $x = 0; }
        $command .= "," . int($x) ;

        last SUBAREA_ENCODE;
        };

# Sector

      $sa == 2 && do
        {
        do
          {
          print "\nEnter radius (0-4095 meter steps; default = 0 [single point]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 4095) || $x eq "");
        if ($x eq "")
          { $x = 0; }
        $command .= "," . int($x) ;

        do
          {
          print "\nEnter orientation of left boundary edge (1-359 degree steps;\ndefault = 0 [no rotation]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 359) || $x eq "");
        if ($x eq "")
          { $x = 0; }
        $command .= "," . int($x) ;

        do
          {
          print "\nEnter orientation of right boundary edge (1-359 degree steps;\ndefault = 0 [no rotation]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 359) || $x eq "");
        if ($x eq "")
          { $x = 0; }
        $command .= "," . int($x) ;

        last SUBAREA_ENCODE;
        };

# Polyline and Polygon

      ($sa == 3 || $sa == 4) && do
        {
        for ($i=1; $i<=4; $i++)
          {
          do
            {
            printf "\nEnter true bearing from Point %d to Point %d (0.0-359.5 degrees;\ndefault = 360 [not available]): ", $i-1, $i;
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq "");
          if ($x eq "")
              { $x = 360; }
          $command .= "," . int ($x*2) ;  # The value is translated to half-degree units

          do
            {
            printf "\nEnter distance from Point %d to Point %d (1-%d meter steps;\ndefault = 0 [no point]): ", $i-1, $i, $max_distance;
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= $max_distance) || $x eq "");
          if ($x eq "")
              { $x = 0; }
          $command .= "," . int ($x) ;
          }

        last SUBAREA_ENCODE;
        };

# Associated Text

      $sa == 5 && do
        {
        print "\nEnter up to $num_chars characters of text: ";
        chomp ($x = <STDIN>);

        if (length ($x) > $num_chars)
          { $x = substr ($x,0,$num_chars); }
        if (length ($x) < $num_chars)
          { $x .= "@" x ($num_chars-length($x)); }
        $command .= ",'" . uc($x) . "'" ;

        last SUBAREA_ENCODE;
        };
      }

    $command .= " ";
    }
  }

# ********************************************************
# type_06_000_00
# ********************************************************

# Routine to get message-specific information for Type 6 "Monitoring aids to navigation"
# messages (DAC=0, FI=0)

# All information (except off-position status) goes into a --info68 comma-delimited string

sub type_06_000_00
  {
  my ($x);

  do
    {
    print "\nEnter sub-application identfier (1-65535); Default = 1: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 1 && $x <= 65535) || $x eq ""); 

  if ($x eq "") { $x = 1; }
  $command .= "--info68=" . int($x) . ",";
  
  do
    {
    print "\nEnter lantern supply voltage, in 0.1 V resolution (0-409.5): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 409.5); 
    $command .= $x . ",";
  
  do
    {
    print "\nEnter lantern drain current, in 0.1 A resolution (0-102.3): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 102.3); 
    $command .= $x . ",";

  do
    {
    print "\nEnter power supply type (0 = AC, 1 = DC): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && ($x == 0 || $x == 1)); 
  
  do
    {
    print "\nEnter light status (0 = light off, 1 = light on): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && ($x == 0 || $x == 1)); 
    $command .= $x . ",";
  
  do
    {
    print "\nEnter battery status (0 = good, 1 = low voltage): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && ($x == 0 || $x == 1)); 
    $command .= $x . " ";
  
  return;
  }

# ********************************************************
# type_0608_001_00
# ********************************************************

# Routine to get Type 6 and 8 6-bit ASCII text messages, using DAC/FI = 000/01

sub type_0608_001_00
  {
  my ($msg_max,$x);
  my $msg = "@";

# Get the 11-bit text sequence number and the text message. The two values will be "packed" into the --info68 parameter

  do
    {
    print "\nEnter text sequence number (0-2047): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 2047);
    
  if ($type == 6) { $sm_max = 151; }
    else { $sm_max = 156; }
  do
    {
    print "\nEnter text message (1-$sm_max characters); Encoder default = '\@':\n";
    chomp ($msg = <STDIN>);
    }
    until (length ($msg) <= $sm_max); 

  if ($msg ne "")
    { $command .= "--info68=" . $x . ",'" . uc($msg) . "' "; }

  return;
  }

# ********************************************************
# type_06_001_02_03
# ********************************************************

# Routine to get message-specific information for Type 6 "Interrogation for a specific functional message"
# messages (DAC=1, FI=2) and "Capability interrogation " (DAC=1, FI=3)

# These messages are very similar; both need a requested DAC but only 6/1/2 requires a requested FI

sub type_06_001_02_03
  {
  my ($fi) = @_;
  my ($x,$code);
  my $y = 0;
  
  do
    {
    print "\nEnter requested DAC code to interrogate (0-1023): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 1023); 

  if ($fi == 2)
    {
    do
      {
      print "\nEnter requested FI code to interrogate (0-63): ";
      chomp ($y = <STDIN>);
      }
      until (looks_like_number ($y) && $y >= 0 && $y <= 63);
    }
    
  $code = (int($x)*100) + int($y);
  $command .= "--info68=" . $code . " ";

  return;
  }

# ********************************************************
# type_06_001_04
# ********************************************************

# Routine to get message-specific information for Type 6 "Capability reply " (DAC=1, FI=4)

sub type_06_001_04
  {
  my ($x,$code);
  my $y = 0;
  
  do
    {
    print "\nEnter requested DAC code to interrogate (0-1023): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 1023); 
  $command .= "--info68=" . int($x);

  print "\nEnter list of supported functional identifier (FI) code, one per line (0-63, <null> to end):\n";
  do
    {
    do
      {
      print "--> ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >=0 && $x <= 63) || $x eq "");
      if ($x ne "") { $command .= "," . $x; }
    }
    until ($x eq "");
  $command .= " ";
  
  return;
  }

# ********************************************************
# type_06_001_16_40
# ********************************************************

# Routine to get message-specific information for Type 6 "Number of Persons On Board - Addressed"
# messages (DAC=1, FI=16) and "Number of Person on Board" (DAC=1, FI=40)

sub type_06_001_16_40
  {
  my ($x);

  do
    {
    print "\nEnter number of people on board (0-8191, use 0 if not available\n or 8191 for 8191 or higher; Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 8191) || $x eq ""); 

  if ($x ne "")
    { $command .= "--pob=" . int($x) . " "; }

  return;
  }

# ********************************************************
# type_08_001_16
# ********************************************************

# Routine to get message-specific information for Type 8 "VTS Targets" messages (DAC=1, FI=16)

sub type_08_001_16
  {
  my ($id,$x);
  my $i = 0;

# We are allowed up to seven target reports

  while ($i < 7)
    {
    $i++;
    print "\nEnter type of target identifier (0-3) for VTS target #$i or X to stop --";
    do
      {
      print "\n  0 = MMSI, 1 == IMO number, 2 = callsign, 3 = other/unknown [default]: ";
      chomp ($id = <STDIN>);
      }
      until (($id >= 0 && $id <= 4) || uc($id) eq "X" || $id eq ""); 
    if (uc($id) eq "X") { last; }

    if ($id eq "") { $id = 3; }
    $command .= "--sa" . sprintf ("%02d", $i) . "=" . $id . ",";

    ID_TYPE:
      {
      ($id == 0) && do
        {
        do
          {
          print "\nEnter MMSI for VTS target (9 decimal digits): ";
          chomp ($x = <STDIN>);
          }
          until (verify_numeric_string ($x,9));
        last ID_TYPE;
        };
      ($id == 1) && do
        {
        do
          {
          print "\nEnter IMO Number of VTS target (7 decimal digits): ";
          chomp ($x = <STDIN>);
          }
          until (verify_numeric_string ($x,7));
        last ID_TYPE;
        };
      ($id == 2) && do
        {
        do
          {
          print "\nEnter VTS target vessel's call sign (1-7 characters):\n";
          chomp ($x = <STDIN>);
          }
          until (length ($x) <= 7);
        $x = uc ($x);
        last ID_TYPE;
        };
      ($id == 3) && do
        {
        $x = 0;
        last ID_TYPE;
        };
      };
    $command .= $x . ",";

    do
      {
      print "\nEnter latitude (-90 to 90; default = 91 [not available]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= -90 && $x <= 91) || $x eq ""); 
    if ($x eq "") { $x = 91; }
    $command .= $x . "," ;

    do
      {
      print "\nEnter longitude (-180 to 180; default = 181 [not available]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= -180 && $x <= 181) || $x eq ""); 
    if ($x eq "") { $x = 181; }
    $command .= $x . ",";

    do
      {
      print "\nEnter course over ground, in degrees (0 to 359; use 360 if data not available [default]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 
    if ($x eq "") { $x = 360; }
    $command .= $x . ",";

    do
      {
      print "\nEnter UTC timestamp, in seconds (0-59; 60 = time stamp not available [default]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 63) || $x eq ""); 
    if ($x eq "") { $x = 60; }
    $command .= $x . ",";

    do
      {
      print "\nEnter speed, in knots (0-254; 255 = speed not available [default]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 255) || $x eq ""); 
    if ($x eq "") { $x = 255; }
    $command .= $x . " ";
    };

  return;
  }

# ********************************************************
# type_06_001_18
# ********************************************************

# Routine to get message-specific information for Type 6 "Clearance Time to Enter Port - Addressed"
# messages (DAC=1, FI=18)

sub type_06_001_18
  {
  my ($x);

  do
    {
    print "\nEnter name of port and berth (1-20 characters):\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 20); 

  if ($x ne "")
    { $command .= "--info68='" . uc($x) . "' "; }

  do
    {
    print "\nEnter Destination (2-5 characters, see\nhttps://unece.org/trade/cefact/unlocode-code-list-country-and-territory for UN/LOCODE list): ";
    chomp ($x = <STDIN>);
    }
    until (length ($x) >= 2 && length ($x) <= 5); 

  if ($x ne "") 
    { $command .= "--dest='" . uc($x) . "' "; }

  return;
  }

# ********************************************************
# type_08_001_18
# ********************************************************

# Routine to get message-specific information for Type 8 "Advice of Waypoints/VTS Route Plan" messages (DAC=1, FI=18)

sub type_08_001_18
  {
  my ($awp,$i,$x);

  do
    {
    print "\nEnter number of waypoints (1-12) or 0 if none [default]: ";
    chomp ($awp = <STDIN>);
    }
    until ((looks_like_number ($awp) && $awp >= 0 && $awp <= 12) || $awp eq ""); 
  if ($awp eq "") { $awp = 0; }
  $command .= "--info68=" . $awp . ",";

  for ($i=1; $i<=$awp; $i++)
    {
    print "\n--Waypoint #$i information --\n";
    do
     {
     print "\nEnter longitude (-180 to 180): ";
     chomp ($x = <STDIN>);
     }
     until ((looks_like_number ($x) && $x >= -180 && $x <= 181)); 
    $command .= $x . ",";

    do
      {
      print "\nEnter latitude (-90 to 90): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= -90 && $x <= 91)); 
    $command .= $x . ",";

    do
      {
      print "\nEnter turning radius, in meters (1-4095) or 0 if not available [default]: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 4095) || $x eq "");
    if ($x eq "") { $x = 0; }
    $command .= $x . ",";
    };

  do
    {
    print "\nEnter route description (1-20 characters) or null if not available:\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 20); 
  $command .= "'" . uc($x) . "' ";

  return;
  }
# ********************************************************
# type_06_001_19
# ********************************************************

# Routine to get message-specific information for Type 6 "Extended ship static and voyage related data"
# messages (DAC=1, FI=19)

sub type_06_001_19
  {
  my ($x);

  do
    {
    print "\nEnter height over keel, in meters, in 0.1 m resolution (0-204.6 m;\nuse  204.7 for heights of 204.7 m or higher; use 0 if height not available; Default = 0: ";

    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 204.7) || $x eq ""); 

  if ($x eq "") { $x = 0; }
  $command .= "--info68=" . int($x*10) . " ";

  return;
  }

# ********************************************************
# type_06_001_20
# ********************************************************

# Routine to get message-specific information for Type 6 "Berthing Data - Addressed"
# messages (DAC=1, FI=20)

sub type_06_001_20
  {
  my ($i,$x);
  my @service_list = (
   "", # List in standard starts at 1
   "Agent", #1
   "Bunker/fuel", #2
   "Chandler", #3
   "Stevedore", #4
   "Electrical", #5
   "Potable water", #6
   "Customs house", #7
   "Cartage", #8
   "Crane(s)", #9
   "Lift(s)", #10
   "Medical facilities", #11
   "Navigation repair", #12
   "Provisions", #13
   "Ship repair", #14
   "Surveyor", #15
   "Steam", #16
   "Tugs", #17
   "Waste disposal (solid)", #18
   "Waste disposal (liquid)", #19
   "Waste disposal (hazardous)", #20
   "Reserved ballast exchange", #21
   "Additional services are also available", #22
   "(Reserved for regional use)", #23
   "(Reserved for regional use)", #24
   "(Reserved for future use)", #25
   "(Reserved for future use)", #26
   );

  do
    {
    print "\nEnter berth length, in meters (1-510; use 511 for lengths >=511m\n and 0 if info not available) [Default = 0]: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 511) || $x eq ""); 

  if ($x ne "")
    { $command .= "--berth_len=" . int($x) . " "; }

  do
    {
    print "\nEnter water depth at berth, in meters, in 0.1 m resolution (0.1-25.4;\nuse 25.5 for depths >=25.5 m and 0 if info not available)\nDefault = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 25.5) || $x eq ""); 

  if ($x ne "")
    { $command .= "--berth_dep=" . $x . " "; }

  do
    {
    print "\nEnter mooring position (0-5):\n";
    print " 0. Undefined      2. Starboard-side to       4. Mooring buoy\n";
    print " 1. Port-side to   3. Mediterranean mooring   5. Anchorage\n";
    print "Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 5) || $x eq ""); 

  if ($x ne "")
    { $command .= "--mooring=" . int($x) . " "; }

  do
    {
    print "\nEnter 0 if service types unknown [default] or 1 to enter types of services: ";
    chomp ($svc = <STDIN>);
    }
    until ((looks_like_number ($svc) && $svc >= 0 && $svc <= 1) || $svc eq ""); 

  if ($svc ne "")
    { $command .= "--svc=" . $svc . " "; }

  if ($svc == 1)
    {
    print "Enter status of available services. For each service listed, enter:\n";
    print "  0: Not available/requested [default]   1: Available   2: No data or unknown\n\n";

    $tos = "";

    for ($i=1; $i<=22; $i++)
      {
      printf "Service: %s...\n", $service_list [$i];
      do
        {
        print " Status [0-2]: ";
        chomp ($x = <STDIN>);
        }
        until ((looks_like_number ($x) && $x >= 0 && $x <= 2) || $x eq "");

      if ($x eq "") { $x = 0; }
      $tos .= $x;
      };

    $command .= "--tos=" . $tos . "0000 ";
    }

  do
    {
    print "\nEnter name of berth (1-20 characters):\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 20); 

  $command .= "--info68='" . uc($x) . "' ";

  return;
  }

# ********************************************************
# type_0608_001_27_28
# ********************************************************

# Routine to get message-specific information for Type 6 "Route Information - Addressed" messgaes
# (DAC=1, FI=28) and Type 8 "Route Information - Broadcast" messages (DAC=1, FI=27)

sub type_0608_001_27_28
  {
  my ($i,$lat,$long,$numpoints,$x);

  do
    {
    print "\nEnter sender classification (0 = ship [default], 1 = authority):\n";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 1) || $x eq ""); 

  if ($x ne "")
    { $command .= "--sndrclass=" . int($x) . " "; }

  do
    {
    print "\nEnter route type (0-5 or 31):\n";
    print " 0. Not available       3. Alternative route               31. Cancellation\n";
    print " 1. Mandatory route     4. Recommended route through ice\n";
    print " 2. Recommended route   5. Ship route plan\n";
    print "Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && (($x >= 0 && $x <= 5) || $x == 31)) || $x eq ""); 

  if ($x ne "")
    { $command .= "--routetype=" . int($x) . " "; }

  do
    {
    print "\nEnter the duration of the validity of the route, in minutes (1-262142; use 262143 if info not available\nor 0 to cancel route [Default = 262143]: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 262143) || $x eq ""); 

  if ($x ne "")
    { $command .= "--duration=" . int($x) . " "; }

# Assemble the info68 string, in format "N,long1,lat1...,longN,latN"

  do
    {
    print "\nEnter the number of waypoints (0-16 [Default=0]): ";
    chomp ($numpoints = <STDIN>);
    }
    until ((looks_like_number ($numpoints) && $numpoints >= 0 && $numpoints <= 16) || $numpoints eq ""); 

  if ($numpoints eq "") { $numpoints = 0; }
  $command .= "--info68=" . int($numpoints);

# I am reusing the p_position code to get latitude and longitude, because we just need the value...

  if ($numpoints > 0)
    {
    for ($i=1; $i<=$numpoints; $i++)
      {
      print "\n--Waypoint #$i latitude and longitude --\n";
      do
        {
        print "Enter latitude (-90 to 90) or 91 if not available; Default = 91: ";
        chomp ($lat = <STDIN>);
        }
        until ((looks_like_number ($lat) && $lat >= -90 && $lat <= 91) || $lat eq ""); 

      if ($lat eq "") { $lat = 91; }

      do
       {
       print "\nEnter longitude (-180 to 180) or 181 if not available; Default = 181: ";
       chomp ($long = <STDIN>);
       }
       until ((looks_like_number ($long) && $long >= -180 && $long <= 181) || $long eq ""); 

      if ($long eq "") { $x = 181; }
 
      $command .= "," . $long . "," . $lat;
      }
    
    $command .= " ";
    }
    
  return;
  }

# ********************************************************
# type_08_001_31
# ********************************************************

# Routine to get message-specific information for Type 8 "Meteorological and Hydrological Data"
# messages (DAC=1, FI=31)

sub type_08_001_31
  {
  my $x;

  do
    {
    print "\nEnter horizontal visibility, in increments of 0.1 nautical miles (0-12.5;\nuse 12.6 for visibility >12.5 nm and 12.7 if info not available)\nDefault = 12.7: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 12.7) || $x eq ""); 

  if ($x ne "")
    { $command .= "--vis=" . $x . " "; }

  do
    {
    print "\nEnter wave height, in meters, in 0.1 m resolution (0-25;\nuse 25.1 for heights of >25 m and 25.5 if info not available)\nDefault = 25.5: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && (($x >= 0 && $x <= 25.1) || $x == 25.5)) || $x eq ""); 

  if ($x ne "")
    { $command .= "--waveheight=" . $x . " "; }

  do
    {
    print "\nEnter wave period, in seconds (0-60; 63 if info not available)\nDefault = 63: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && (($x >= 0 && $x <= 60) || $x == 63)) || $x eq ""); 

  if ($x ne "")
    { $command .= "--waveperiod=" . int($x) . " "; }

  do
    {
    print "\nEnter wave direction, in degrees (0 to 359; 360 if data not\navailable); Default = 360: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 

  if ($x ne "")
    { $command .= "--wavedir=" . int($x) . " "; }

  do
    {
    print "\nEnter swell height, in meters, in 0.1 m resolution (0-25;\nuse 25.1 for heights of >25 m and 25.5 if info not available)\nDefault = 25.5: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && (($x >= 0 && $x <= 25.1) || $x == 25.5)) || $x eq ""); 

  if ($x ne "")
    { $command .= "--swellheight=" . $x . " "; }

  do
    {
    print "\nEnter swell period, in seconds (0-60; 63 if info not available)\nDefault = 63: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && (($x >= 0 && $x <= 60) || $x == 63)) || $x eq ""); 

  if ($x ne "")
    { $command .= "--swellperiod=" . int($x) . " "; }

  do
    {
    print "\nEnter swell direction, in degrees (0 to 359; 360 if data not\navailable); Default = 360: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 

  if ($x ne "")
    { $command .= "--swelldir=" . int($x) . " "; }

  do
    {
    print "\nEnter sea state per Beaufort scale (0-13):\n";
    print "  0. Calm: <1 kn (Sea like a mirror)\n";
    print "  1. Light air: 1-3 kn (Ripples without crests)\n";
    print "  2. Light breeze: 4-6 kn (Small wavelets; crests not breaking)\n";
    print "  3. Gentle breeze: 7-10 kn (Large wavelets; crests begin to break)\n";
    print "  4. Moderate breeze: 11-16 kn (Small waves)\n";
    print "  5. Fresh breeze: 17-21 kn (Moderate waves)\n";
    print "  6. Strong breeze: 22-27 kn (Large waves with foam crests and some spray)\n";
    print "  7. High wind: 28-33 kn (Sea heaps up; foam begins to streak)\n";
    print "  8. Gale: 34-40 kn (Moderately high waves with breaking crests)\n";
    print "  9. Strong gale: 41-47 kn (High waves)\n";
    print " 10. Storm: 48-55 kn (Very high waves; reduced visibility)\n";
    print " 11. Violent storm: 56-63 kn (Exceptionally high waves)\n";
    print " 12. Hurricane force: >63 kn (Huge waves)\n";
    print " 13. Information not available\n";
    print "Default = 13: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 13) || $x eq ""); 

  if ($x ne "")
    { $command .= "--seastate=" . int($x) . " "; }

  do
    {
    print "\nEnter precipitation type (1-5, 7):\n";
    print " 1. Rain         2. Thunderstorm     3. Freezing rain\n";
    print " 4. Mixed/ice    5. Snow             7. (Not available)\n";
    print "Default = 7: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 1 && $x <= 7 && $x != 6) || $x eq ""); 

  if ($x ne "")
    { $command .= "--precip=" . int($x) . " "; }

  do
    {
    print "\nEnter ice status (0, 1, 3):\n";
    print " 0. None     1. Present     3. (Not available)\n";
    print "Default = 3: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 3 && $x != 2) || $x eq ""); 

  if ($x ne "")
    { $command .= "--ice=" . int($x) . " "; }

  return;
  }

# ********************************************************
# type_06_235_250_10
# ********************************************************

# Routine to get message-specific information for Type 6 "Aids to Navigation Monitoring Data"
# messages (DAC=235/250, FI=10)

sub type_06_235_250_10
  {
  my ($i,$status,$x);
  my @analog_list = (
    "internal supply voltage to AIS unit", #0
    "voltage from hardware analogue input no. 1", #1
    "voltage from hardware analogue input no. 2", #2
    );

  $command .= "--info68=";
  for ($i=0; $i<3; $i++)
    {
    do
      {
      printf "\nEnter %s in 0.05 V steps (0-36): ", $analog_list [$i];
      chomp ($x = <STDIN>);
      }
      until (looks_like_number ($x) && $x >= 0 && $x <= 36); 
    $command .= int($x*20) . ",";
    }

  $status = 0;
  for ($i=0; $i<8; $i++)
    {
    do
      {
      print "\nEnter status of digital input #$i (0 = off, 1 = on): ";
      chomp ($x = <STDIN>);
      }
      until ($x == 0 || $x == 1);
    $status += $x * (2**$i);
    }
  $command .= $status . " ";
  
  return;
  }

# ********************************************************
# type_06_235_20
# ********************************************************

# Routine to get message-specific information for Type 6 "Buoy Position Monitoring"
# messages (DAC=235, FI=20)

sub type_06_235_20
  {
  my ($i,$status,$x);

  do
    {
    printf "\nEnter AtoN name (1-34 characters); Encoder default = '%s':\n", "@"x34;
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 34);

  if ($x ne "")
    { $command .= "--aid_name='" . uc($x) . "' "; }
  
  return;
  }

# ********************************************************
# type_08_366_20
# ********************************************************

# Routine to get message-specific information for Type 8 "Dam Status - Broadcast"
# messages (DAC=366, FI=20)

sub type_08_366_20
  {
  my ($i,$n,$x);
  my @discharge = ("over the dam", "through the associated hydroelectric plant");

  do
    {
    print "\nEnter last three characters of lock/dam UN-LOCODE (0-3 characters): ";
    chomp ($x = <STDIN>);
    }
    until (length ($x) >= 0 && length($x) <= 3); 
    $command .= "--lockinfo='" . uc($x) . "',";

  do
    {
    print "\nEnter 4-character Navigation Unit Identifier (GUID) code: ";
    chomp ($x = <STDIN>);
    }
    until (length ($x) == 4); 
    $command .= "'" . uc($x) . "',";

  do
    {
    print "\nEnter the total amount of dam opening (all gates), in 0.1 m units\n(0-810.0 m or 810.1 if not available [default]): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && ($x >= 0 && $x <= 810.1)) || $x eq ""); 
  if ($x eq "") { $x = 810.1; }
  $command .= $x;

  for ($i = 0; $i <= 1; $i++)
    {
    do
      {
      printf "\nEnter the total discharge %s in cubic meters/sec\n(0-409.0; use 409.1 if not applicable or 409.2 if not available [default]): ", $discharge [$i];
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && ($x >= 0 && $x <= 409.2)) || $x eq ""); 
    if ($x eq "") { $x = 409.2; }
    $command .= "," . $x;
    
    do
      {
      printf "\nEnter the source of the discharge information\n(0 = measured, 1 = calculated, 2 = observed, 3 = unknown [default]: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && ($x >= 0 && $x <= 3)) || $x eq ""); 
    if ($x eq "") { $x = 3; }
    $command .= "," . int($x);
    }
  $command .= " ";

  do
    {
    print "\nEnter the total amount of gates to report settings (0-30 or 31 if more than 30 gates); Default = 0: ";
    chomp ($n = <STDIN>);
    }
    until ((looks_like_number ($n) && ($n >= 0 && $n <= 31)) || $n eq ""); 
  if ($n eq "") { $n = 0; }
  $command .= "--info68=" . int($n);

  if ($n > 0)
    {
    $n = min ($n,30);
    for ($i = 1; $i <= $n; $i++)
      {
      do
        {
        printf "\nEnter the opening for gate $i in meters\n(0-25.0; use 25.1 if greater than 25 m or 25.2 if not available [default]): ";
        chomp ($x = <STDIN>);
        }
        until ((looks_like_number ($x) && ($x >= 0 && $x <= 25.2)) || $x eq ""); 
      if ($x eq "") { $x = 25.2; }
      $command .= "," . $x;
      }
    }
  $command .= " ";

  return;
  }

# ********************************************************
# type_06_367_16
# ********************************************************

# Routine to get message-specific information for Type 6 "Passenger and Crew Count – Addressed"
# messages (DAC=367, FI=16)

sub type_06_367_16
  {
  my ($x);

  do
    {
    print "\nEnter number of crew members on board (0-8191, use 0 if not available\n or 8191 for 8,191 crew or more; Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 8191) || $x eq ""); 

  if ($x eq "") { $x = 0; }
  $command .= "--pob=" . int($x);
  
  do
    {
    print "\nEnter number of passengers on board (0-32500 passengers, 32501 for 32,501 passengers or more,\n; or 32502 if unknown; Default = 32502: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 32502) || $x eq ""); 

  if ($x eq "") { $x = 32502; }
  $command .= "," . int($x) . " ";

  do
    {
    print "\nEnter additional vessel type information:\n";
    print " 0. Unknown or not in list     1. Ferry vessel (people and vehicles)     2. Ferry vessel (people only)\n";
    print " 3. Cruise ship                4. Excursion vessel                       5-15. Reserved for future use\n";
    print "Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 4) || $x eq ""); 

  if ($x ne "") { $command .= "--info68=" . int($x) . " "; }
  return;
  }
  
# ********************************************************
# type_08_367_26
# ********************************************************

# Routine to get message-specific information for Type 8 "GPS Jamming and Spoofing Report"
# messages (DAC=367, FI=26)

sub type_08_367_26
  {
  my ($x);

  do
    {
    print "\nEnter type of event (0 = jamming, 1 = spoofing): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && ($x == 0 || $x == 1));
  $command .= "--info68=" . int($x) . ",";
  
  do
    {
    print "\nEnter event status (0 = start of event, 1 = end of event): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && ($x == 0 || $x == 1));
  $command .= int($x) . " ";

  return;
  }
  
# ********************************************************
# type_08_367_33
# ********************************************************

# Routine to get information for Type 8 "Environmental Message"
# messages (DAC=367, FI=33)

# The EM message is composed of 1-8 Sensor Reports, each of which is 85 bits. There are a dozen different sensor report formats which is why this
# routine is so large. Many of the sensor rpeorts have overlapping data. Several of the parameters match those in other AIS messages but this
# routine can't use the ordinary switches for the AIS_ping program, so there's a lot of duplication

sub type_08_367_33
  {
  my ($i,$j,$k,$num_reports,$rep_type,$x);
  my @vector = ("North","East","Up");

  do
    {
    print "\nEnter number of sensor reports (1-8): ";
    chomp ($num_reports = <STDIN>);
    }
    until (looks_like_number ($num_reports) && $num_reports >= 1 && $num_reports <= 8);

  for ($i=1; $i<=$num_reports; $i++)
    {
    print "\nWhat type of sensor data for sensor report #$i (0-11)?\n";
    print " 0: Site Location                 1: Station ID                     2: Wind\n";
    print " 3: Tide/River Water Level        4: 2D Vertical Current Profile    5: 3D Vertical Current Profile\n";
    print " 6: Horizontal Current Profile    7: Sea State                      8: Salinity\n";
    print " 9: Weather                      10: Air Gap                       11: Wind v2\n";
    do
      {
      print "--->  ";
      chomp ($rep_type = <STDIN>);
      }
      until (looks_like_number ($rep_type) && $rep_type >= 0 && $rep_type <= 11);
      
    $command .= "--sa" . sprintf ("%02d", $i) . "=" . $rep_type . ",";
    $command .= em_date ("time of data",1) . ",";

    do
      {
      print "\nEnter the Binary identifier of the sensor site (0-127): ";
      chomp ($x = <STDIN>);
      }
      until (looks_like_number ($x) && $x >= 0 && $x <= 127);
    $command .= $x . ",";
    
    SENSOR_REPORT:
      {
      $rep_type == 0 && do
        {
        $command .= "3,";
        do
          {
          print "\nEnter longitude (-180 to 180; default = 181 [not available]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= -180 && $x <= 181) || $x eq ""); 
        if ($x eq "") { $x = 181; }
        $command .= $x . "," ;
        do
          {
          print "\nEnter latitude (-90 to 90; default = 91 [not available]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= -90 && $x <= 91) || $x eq ""); 
        if ($x eq "") { $x = 91; }
        $command .= $x . ",";
        do
          {
          print "\nEnter number of decimal places for lat/long (0-4; default = 5 [unknown]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 5) || $x eq "");
        if ($x eq "") { $x = 5; }
        $command .= int($x) . ",";
        do
          {
          print "\nEnter altitude of the sensor relative to MSL in 0.1 m steps (-200 to 200);\nuse 200.1 if the altitude is > 200 m or 200.2 if data unavailable [default]: ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= -200 && $x <= 200.2) || $x eq "");
        if ($x eq "") { $x = 200.2; }
        $command .= $x . ",";

        print "\nEnter owner of the sensor/responsible for the sensor data (0-14) -- \n";
        print " 0. Unknown (default)                     1. U.S. NOAA National Ocean Service\n";
        print " 2. U.S. Army Corps of Engineers          3. Coastal directorate\n";
        print " 4. NOAA National Weather Service         5. State or Port Authority\n";
        print " 6. U.S. Coast Guard                      7. U.S. Navy\n";
        print " 8. Energy or Environmental Agency        9. Transportation Agency\n";
        print "10. Academic or Research Institution     11. Private or Commercial Service\n";
        print "12. Marine Exchange or Port Service      13. US Geologic Survey\n";
        print "14. Other                                15. (not to be used)\n";
        do
          {
          print "--->  ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 14) || $x eq "");
        if ($x eq "") { $x = 0; }
        $command .= $x . ",";

        print "\nEnter length of time that data is valid (0-6) -- \n";
        print " 0. No timeout period (default)       1. 10 min.        2. 1 hour\n";
        print " 3. 6 hours        4. 12 hours        5. 24 hours       6. 5 min.\n";
        do
          {
          print "--->  ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 6) || $x eq "");
        if ($x eq "") { $x = 0; }
        $command .= $x;
        last SENSOR_REPORT;
        };

      $rep_type == 1 && do
        {
        do
          {
          print "\nEnter station identifier (1-14 characters): ";
          chomp ($x = <STDIN>);
          }
          until (length ($x) >= 1 && length ($x) <= 14);
        $command .= "'" . uc($x) . "'" ;  
        last SENSOR_REPORT;
        };

      $rep_type == 2 && do
        {
        $command .= em_wind ("previous 10-minute",1);
        $command .= em_sensor_data_description ("wind") . ",";
        $command .= em_wind ("predicted",0);
        $command .= em_date ("valid time of forecast",1) . ",";
        $command .= em_forecast_duration ();
        last SENSOR_REPORT;
        };
  
      $rep_type == 3 && do
        {
        $command .= em_water ("current") . ",";
        
        do
          {
          print "\nEnter water level trend (0 = increasing, 1 = decreasing, 2 = steady,\n3 = unknown/unavailable [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 3) || $x eq "");
        if ($x eq "") { $x = 3; }
        $command .= $x . ",";
        
        print "\nEnter vertical reference datum (0-14) -- \n";
        print " 0. Mean Lower Low Water (MLLW)                   1. International Great Lakes Datum (IGLD-85)\n";
        print " 2. Local river datum                             3. Station Datum (STND)\n";
        print " 4. Mean Higher High Water (MHHW)                 5. Mean High Water (MHW)\n";
        print " 6. Mean Sea Level (MSL)                          7. Mean Low Water (MLW)\n";
        print " 8. National Geodetic Vertical Datum (NGVD-29)    9. North American Vertical Datum (NAVD-88)\n";
        print "10. World Geodetic System (WGS-84)               11. Lowest Astronomical Tide (LAT)\n";
        print "12. Pool                                         13. Gauge\n";
        print "14. Unknown/unavailable [default\n";
        do
          {
          print "--->  ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 14) || $x eq "");
        if ($x eq "") { $x = 14; }
        $command .= $x . ",";

        $command .= em_sensor_data_description ("water level") . ",";
        $command .= em_water ("forecast") . ",";
        $command .= em_date ("valid time of forecast",1) . ",";
        $command .= em_forecast_duration ();
        last SENSOR_REPORT;
        };

      $rep_type == 4 && do
        {
        for ($j=1; $j<=3; $j++)
          {
          do
            {
            print "\nEnter speed of current $j, in 0.1 knot increments (0-24.5; use 24.6 for speeds 24.6 or greater\n";
            print "or 24.7 if data is unavailable [default]): ";
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 24.7) || $x eq ""); 
          if ($x eq "") { $x = 24.7; }
          $command .= $x . ",";

          do
            {
            print "\nEnter direction of current $j, in 1-degree increments (0-359; 360 = data not available [default]): ";
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 
          if ($x eq "") { $x = 360; }
          $command .= $x . ",";

          do
            {
            print "\nEnter measuring level below water surface of current $j, in meter increments (0-360;\n 361 = 361 m or more and 362 = data unavaulable [default]): ";
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 362) || $x eq ""); 
          if ($x eq "") { $x = 362; }
          $command .= $x . ",";
          };

        $command .= em_sensor_data_description ("current");
        last SENSOR_REPORT;
        };

      $rep_type == 5 && do
        {
        for ($j=1; $j<=2; $j++)
          {
          for ($k=0; $k<3; $k++)
            {
            do
              {
              printf "\nEnter speed of %s component of current %d, in 0.1 knot increments (-25.0 to 25.0,\n", $vector[$k], $j;
              print "use -25.1 for speeds -25.1 or less, 25.1 for speeds 25.1 or more, or -25.6 if data unavailable [default]): ";
              chomp ($x = <STDIN>);
              }
              until ((looks_like_number ($x) && ($x >= -25.1 && $x <= 25.1) || $x == -25.6) || $x eq ""); 
            if ($x eq "") { $x = -25.6; }
            $command .= $x . ",";
            };

          do
            {
            print "\nEnter measuring level below water surface of current $j, in meter increments (0-360;\n 361 = 361 m or more and 362 = data unavaulable [default]): ";
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 362) || $x eq ""); 
          if ($x eq "") { $x = 362; }
          $command .= $x . ",";
          };

        $command .= em_sensor_data_description ("current");
        last SENSOR_REPORT;
        };

      $rep_type == 6 && do
        {
        do
          {
          print "\nEnter bearing of current readings from the sensor position, in 1 degree increments\n(0-359; 360 = data not available [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 
        if ($x eq "") { $x = 360; }
        $command .= $x . ",";

        for ($j=1; $j<=2; $j++)
          {
          do
            {
            print "\nEnter distance of current $j reading from sensor position, in 1 m steps\n";
            print "(0-360; 361 = 361 m or more, or 362 = data unavailable [default]): ";
            chomp ($x = <STDIN>);
            }
            until (looks_like_number ($x) &&($x >= 0 && $x <= 362) || $x eq ""); 
          if ($x eq "") { $x = 362; }
          $command .= $x . ",";

          do
            {
            print "\nEnter speed of current $j, in 0.1 knot increments (0-24.5; 24.6 = 24.6 kn or greater,\n24.7 = data unavailable [default]): ";
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 24.7) || $x eq ""); 
          if ($x eq "") { $x = 24.7; }
          $command .= $x . ",";

          do
            {
            print "\nEnter direction of current $j, in 1 degree increments (0-359; 360 = data not available [default]): ";
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 
          if ($x eq "") { $x = 360; }
          $command .= $x . ",";
          
          do
            {
            print "\nEnter measurement level below water surface of current $j, in meter increments (0-360;\n 361 = 361 m or more and 362 = data unavaulable [default]): ";
            chomp ($x = <STDIN>);
            }
            until ((looks_like_number ($x) && $x >= 0 && $x <= 362) || $x eq ""); 
          if ($x eq "") { $x = 362; }
          $command .= $x . ",";
          };

        $command .= em_sensor_data_description ("current");
        last SENSOR_REPORT;
        };
        
      $rep_type == 7 && do
        {
        do
          {
          print "\nEnter swell height, in meters, in 0.1 m resolution (0-24.5;\n24.6 = 24.6 m or greater, 24.7 = info not available [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && ($x >= 0 && $x <= 24.7)) || $x eq ""); 
        if ($x eq "") { $x = 24.7; }
        $command .= $x . ","; 

        do
          {
          print "\nEnter swell period, in seconds (0-60; 61 = info not available [default]): ";
         chomp ($x = <STDIN>);
          }
          until (looks_like_number ($x) && (($x >= 0 && $x <= 61)) || $x eq ""); 
        if ($x eq "") { $x = 61; }
        $command .= $x . ","; 

        do
          {
          print "\nEnter swell direction, in degrees (0 to 359; 360 = data not\navailable [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 
        if ($x eq "") { $x = 360; }
        $command .= $x . ","; 

        print "\nEnter sea state per Beaufort scale (0-13):\n";
        print "  0. Calm: <1 kn (Sea like a mirror)\n";
        print "  1. Light air: 1-3 kn (Ripples without crests)\n";
        print "  2. Light breeze: 4-6 kn (Small wavelets; crests not breaking)\n";
        print "  3. Gentle breeze: 7-10 kn (Large wavelets; crests begin to break)\n";
        print "  4. Moderate breeze: 11-16 kn (Small waves)\n";
        print "  5. Fresh breeze: 17-21 kn (Moderate waves)\n";
        print "  6. Strong breeze: 22-27 kn (Large waves with foam crests and some spray)\n";
        print "  7. High wind: 28-33 kn (Sea heaps up; foam begins to streak)\n";
        print "  8. Gale: 34-40 kn (Moderately high waves with breaking crests)\n";
        print "  9. Strong gale: 41-47 kn (High waves)\n";
        print " 10. Storm: 48-55 kn (Very high waves; reduced visibility)\n";
        print " 11. Violent storm: 56-63 kn (Exceptionally high waves)\n";
        print " 12. Hurricane force: >63 kn (Huge waves)\n";
        print " 13. Information not available [Default]\n";
        do
          {
          print "--> ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 13) || $x eq ""); 
        if ($x eq "") { $x = 13; }
        $command .= $x . ","; 

        $command .= em_sensor_data_description ("swell") . ",";
        $command .= em_water_temp () . ",";

        do
          {
          print "\nEnter depth of water temperature sensor, in 0.1 m steps (0-12.0;\n12.1 = 12.1 m or more, 12.2 = data not available [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 12.2) || $x eq ""); 
        if ($x eq "") { $x = 12.2; }
        $command .= $x . ",";

        $command .= em_sensor_data_description ("water temperature") . ",";
        
        do
          {
          print "\nEnter wave height, in 0.1 m steps (0-24.5; 24.6 = 24.6 m or greater;\n24.7 = data unavailable [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 24.7) || $x eq ""); 
        if ($x eq "") { $x = 24.7; }
        $command .= $x . ",";

        do
          {
          print "\nEnter wave period, in seconds (0-60; 61 = data unavailable [default]): ";
          chomp ($x = <STDIN>);
          }
          until (looks_like_number ($x) &&($x >= 0 && $x <= 61) || $x eq ""); 
        if ($x eq "") { $x = 61; }
        $command .= $x . ",";

        do
          {
          print "\nEnter wave direction, in degrees (0-359; 360 = data unavailable [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 
        if ($x eq "") { $x = 360; }
        $command .= $x . ",";

        $command .= em_sensor_data_description ("wave") . ",";
        $command .= em_salinity ();
        last SENSOR_REPORT;
        };
        
      $rep_type == 8 && do
        {
        $command .= em_water_temp () . ",";

        do
          {
          print "\nEnter water conductivity, in steps of 0.01 Siemens/meter (0.00-7.00; 7.01 = 7.01 S/m or more,\n";
          print "7.02 = data not available, 7.03 = sensor not available [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 7.03) || $x eq ""); 
        if ($x eq "") { $x = 7.03; }
        $command .= $x . ",";

        do
          {
          print "\nEnter water pressure, in steps of 0.1 decibars (0.0-6000.0; 6000.1 = 6000.1 decibars or more,\n";
          print "6000.2 = data not available, 6000.3 = sensor not available [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 6000.3) || $x eq ""); 
        if ($x eq "") { $x = 6000.3; }
        $command .= $x . ",";

        $command .= em_salinity () . ",";
        do
          {
          print "\nEnter salinity type (0 = measured, 1 = calculated using PSS-78, 2 = measured using another method): ";
          chomp ($x = <STDIN>);
          }
          until (looks_like_number ($x) && $x >= 0 && $x <= 2);
        $command .= $x . ",";

        $command .= em_sensor_data_description ("salinity");
        last SENSOR_REPORT;
        };

      $rep_type == 9 && do
        {
        do
          {
          print "\nEnter dry bulb air temperature, in units of 0.1\xB0C Celsius (-60. to 60.0)\nor -102.4 if not available; Default = -102.4: ";
          chomp ($x = <STDIN>);
          }
          until (((looks_like_number ($x) && ($x >= -60 && $x <= 60)) || $x == -102.4) || $x eq ""); 
        if ($x eq "") { $x = -102.4; }
        $command .= $x . ",";
        $command .= em_sensor_data_description ("air temperature") . ",";

        do
          {
          print "\nEnter WMO precipitation type (0-3):\n";
          print " 0. Rain     1. Snow     2. Rain & snow     3. None (default): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 3) || $x eq ""); 
        if ($x eq "") { $x = 3; }
        $command .= int($x) . ",";

        do
          {
          print "\nEnter visibility, in increments of 0.1 nautical miles (0-24.0),\n24.1 for visibility 24.1 or greater, 24.2 if data unavailable\n";
          print "or 24.3 if sensor not available; Default = 24.3: ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 24.3) || $x eq "");
        if ($x eq "") { $x = 24.3; }
        $command .= $x . ",";

        do
          {
          print "\nEnter dew point, in increments of 0.1 \xB0C (-20.0-50.0), 50.1 for data unavailable,\n";
          print "50.2 for <-20\xB0C, or 50.3 for >50\xB0C); Default = 50.1: ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= -20 && $x <= 50.3) || $x eq "");
        if ($x eq "") { $x = 50.1; }
        $command .= $x . ",";
        $command .= em_sensor_data_description ("dew point") . ",";

        do
          {
          print "\nEnter air pressure, in hPa/millibars (800-1200), 799 for <800, 1202 for >1200,\n";
          print "or 1203 if data not available; Default = 1203: ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && ($x >= 799 && $x <= 1203)) || $x eq ""); 
        if ($x eq "") { $x = 1203; }
        $command .= $x . ",";
  
        do
          {
          print "\nEnter pressure tendency (0-3):\n";
          print " 0. Steady     1. Decreasing     2. Increasing     3. Not available [default]: ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 3) || $x eq "");
        if ($x eq "") { $x = 3; }
        $command .= int($x) . ",";
  
        $command .= em_sensor_data_description ("air pressure") . ",";
        $command .= em_salinity () . " ";
        last SENSOR_REPORT;
        };

      $rep_type == 10 && do
        {
        $command .= em_air_gap (0) . "," . em_air_gap (1) . ",";

        do
          {
          print "\nEnter trend of the air gap measurement (0-3):\n";
          print " 0. Steady     1. Decreasing     2. Increasing     3. Not available [default]: ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 3) || $x eq "");
        if ($x eq "") { $x = 3; }
        $command .= int($x) . ",";

        $command .= em_air_gap (2) . ",";
        $command .= em_date ("time of predicted air gap",1) . ",";
        $command .= em_sensor_data_description ("air gap");
        last SENSOR_REPORT;
        };

      $rep_type == 11 && do
        {
        $command .= em_wind ("previous N-minute",0);

        do
          {
          print "\nEnter number of minutes (N) for averaging (1-60; 61 = 61 or more; 0 = unknown [default]): ";
          chomp ($x = <STDIN>);
          }
          until ((looks_like_number ($x) && $x >= 0 && $x <= 61) || $x eq "");
        if ($x eq "") { $x = 0; }
        $command .= $x . ",";
        $command .= em_sensor_data_description ("wind") . ",";
        $command .= em_wind ("predicted",0);
        $command .= em_date ("valid time of forecast",0) . ",";
        $command .= em_forecast_duration ();
        last SENSOR_REPORT;
        };
      };

    $command .= " ";
    }

  return;
  }

# ********************************************************
# em_air_gap
# ********************************************************

# Get air draught/gap information for Environmental Messages (Type 8, 367/33)

sub em_air_gap
  {
  my ($opt) = @_;
  my $x;
  my @label = (
    "air draught (vertical distance from the ship's waterline to the highest point on the ship)", #0
    "air gap (vertical distance from the surface of the water to the sensor)", #1
    "predicted air gap (predicted vertical distance from the surface of the water to the sensor)", #2
    );

  do
    {
    printf "\nEnter the %s,\n", $label [$opt];
    print " in 1-cm increments (1-8190; 0 = data unavailable [default], or 8191 = 8191 or greater): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 8191) || $x eq ""); 
  if ($x eq "") { $x = 0; }

  return $x;
  };

# ********************************************************
# em_date
# ********************************************************

# Get date information for Environmental Messages (Type 8, 367/33)

sub em_date
  {
  my ($label,$opt) = @_;
  my ($x);
  my $date_string = "";
  
  print "\nProvide the $label --\n";
  
  if ($opt == 1)
    {
    do
      {
      print "\nEnter UTC day (1-31 or 0 if not available); Default = 0: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 31) || $x eq ""); 
    if ($x eq "") { $x = 0; }
    $date_string .= $x . ",";
    }

  do
    {
    print "\nEnter UTC hour (0-23 or 24 if not available); Default = 24: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 24) || $x eq ""); 
  if ($x eq "") { $x = 24; }
  $date_string .= $x . ",";

  do
    {
    print "\nEnter UTC minute (0-59 or 60 if not available); Default = 60: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 60) || $x eq ""); 
  if ($x eq "") { $x = 60; }
  $date_string .= $x;

  return $date_string;
  }

# ********************************************************
# em_forecast_duration
# ********************************************************

# Get forecast duration for Environmental Messages (Type 8, 367/33)

sub em_forecast_duration
  {
  my ($x);
  
  do
    {
    print "\nEnter duration of the validity of the forecast, in minutes\n(1-255, or 0 to cancel the forecast [default]): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 255) || $x eq ""); 
  if ($x eq "") { $x = 0; }
 
  return $x;
  }

# ********************************************************
# em_salinity
# ********************************************************

# Get salinity for Environmental Messages (Type 8, 367/33)

sub em_salinity
  {
  my ($x);
  
  do
    {
    print "\nEnter salinity, in 0.1 per mille steps (0.0-50.0; 50.1 = 50.1\x{2030} or greater,\n";
    print " 50.2 = data unavailable [default], or 50.3 = sensor not available): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 50.3) || $x eq ""); 
  if ($x eq "") { $x = 50.2; }
 
  return $x;
  }

# ********************************************************
# em_sensor_data_description
# ********************************************************

sub em_sensor_data_description
  {
  my ($label) = @_;
  my ($x);

  do
    {
    print "\nEnter $label sensor data description (0-7) --\n";
    print " 0. No data [default]                  1. Raw real time    2. Real time with Quality Control\n";
    print " 3. Predicted [using historical statistics]      4. Forecast [predicted, plus real-time information]\n";
    print " 5. Nowcast [a continuous forecast]    6. Reserved         7. Sensor not available\n";
    print "--> ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 7) || $x eq ""); 
    if ($x eq "") { $x = 0; }
  
  return $x;
  }

# ********************************************************
# em_water
# ********************************************************

# Get water level information for Environmental Messages (Type 8, 367/33)

sub em_water
  {
  my ($label) = @_;
  my ($x);
  my $water_string = "";

    do
      {
      printf "\nEnter $label type of water level (0 = relative to reference datum, 1 = water depth): ";
      chomp ($x = <STDIN>);
      }
      until (looks_like_number ($x) && ($x == 0 || $x == 1));
    $water_string .= $x . ",";

    do
      {
      print "\nEnter $label water level in cm (-327.66 to 327.66 m; use -327.67 for -327.67 m or less,\n";
      print "327.67 for 327.67 m or greater, and -327.68 for data unavailable [default]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= -327.68 && $x <= 327.67) || $x eq ""); 
    if ($x eq "") { $x = -327.68; }
    $water_string .= $x;
  
  return $water_string;
  }

# ********************************************************
# em_water_temp
# ********************************************************

# Get water temperature information for Environmental Messages (Type 8, 367/33)

sub em_water_temp
  {
  my ($temp, $x);

    do
      {
      print "\nEnter water temperature in degrees Celsius (-10 to 50; 50.1 = data unavailable [default]): ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= -10 && $x <= 50.1) || $x eq "");
    if ($x eq "") { $x = 50.1; }
    $temp = $x;
  
  return $temp;
  }

# ********************************************************
# em_wind
# ********************************************************

# Get wind information for Environmental Messages (Type 8, 367/33)
# OPT = 0 provides windspeed, gust, and speed direction
# OPT = 1 provides speed, gust, speed direction, and gust direction

sub em_wind
  {
  my ($label,$opt) = @_;
  my ($i,$x);
  my @wind_label = ("average", "maximum gust");
  my $wind_string = "";

  for ($i = 0; $i <= 1; $i++)
    {
    do
      {
      printf "\nEnter %s %s wind speed in 1-knot increments (0-120;\n121 = 120 or greater; 122 = data not available [default]): ", $label, $wind_label [$i];
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 122) || $x eq ""); 
    if ($x eq "") { $x = 122; }
    $wind_string .= $x . ",";
    };

  for ($i = 0; $i <= $opt; $i++)
    {
    do
      {
      printf "\nEnter %s %s wind direction in 1-degree increments (0-359\xB0;\n360 = data not available [default]): ", $label, $wind_label [$i];
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 
    if ($x eq "") { $x = 360; }
    $wind_string .= $x . ",";
    };
  
  return $wind_string;
  }


# ********************************************************
# type_0608_367_35
# ********************************************************

# Routine to get Type 6 and 8 Waterways Management messages, using DAC/FI = 367/35.
# There are three different message subtypes...

sub type_0608_367_35
  {
  my ($i,$subtype,$x);

# Start with common information... 
  
  print "\nEnter the waterways message type (0-13)...\n";
  print "  0. Lock                            7. Traffic Advisory\n";
  print "  1. Gate                            8. Cleared to Enter/Proceed\n";
  print "  2. Narrows                         9. Not Cleared to Enter/Do Not Proceed\n";
  print "  3. Bridge                         10. Proceed to Berth\n";
  print "  4. Restricted channel - one       11. Proceed to _____ (defined in Linked\n";
  print "     vessel at a time - could be        Text message\n";
  print "     alternating directions - no    12. Vessels Awaiting Lockage\n";
  print "     passing or overtaking          13. Ice Convoy (specify speed and spacing\n";
  print "  5. Estimated Arrival Time             in Linked Text message\n";
  print "  6. Assigned Arrival Time          14-15. (undefined)\n";
  do
    {
    print "--> ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 15);
  $command .= "--wmtype=" . $x;
  
  print "\nEnter the waterways message subtype (0-2)\n";
  print "  0. Vessel list    1. Place identification   2. Vessel name\n";
  do
    {
    print "--> ";
    chomp ($subtype = <STDIN>);
    }
    until (looks_like_number ($subtype) && $subtype >= 0 && $subtype <= 2);
  $command .= $subtype . " ";

# This message has three subtypes, all with different formats
  
  WM_SUBTYPE:
    {
    $subtype == 0 && do
      {
      do
        {
        print "\nHow many vessels are on the vessel list? (1-7): ";
        chomp ($x = <STDIN>);
        }
        until (looks_like_number ($x) && $x >= 1 && $x <= 7);
      for ($i = 1; $i <= $x; $i++)
        { $command .= "--sa" . sprintf ("%02d", $i) . "=" . p_vessel_info ($i) . " "; }
      last WM_SUBTYPE;
      };

    $subtype == 1 && do
      {
      do
        {
        print "\nEnter longitude (-180 to 180; default = 181 [not available]): ";
        chomp ($x = <STDIN>);
        }
        until ((looks_like_number ($x) && $x >= -180 && $x <= 181) || $x eq ""); 
      if ($x eq "") { $x = 181; }
      $command .= "--info68=" . $x ;

      do
        {
        print "\nEnter latitude (-90 to 90; default = 91 [not available]): ";
        chomp ($x = <STDIN>);
        }
        until ((looks_like_number ($x) && $x >= -90 && $x <= 91) || $x eq ""); 
      if ($x eq "") { $x = 91; }
      $command .= "," . $x ;

      do
        {
        print "\nEnter precision (0-4), number of decimal places for lat/long [default = 4]: ";
        chomp ($x = <STDIN>);
        }
        until ((looks_like_number ($x) && $x >= 0 && $x <= 4) || $x eq "");
      if ($x eq "") { $x = 4; }
      $command .= "," . int($x) ;
        
      do
        {
        print "\nEnter orientation of area, in 1 deg. increments (0-359); 360 = data unavailable [default]: ";
        chomp ($x = <STDIN>);
        }
        until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq "");
      if ($x eq "") { $x = 360; }
      $command .= "," . int($x) ;

      do
        {
        print "\nEnter name/description of the item (1-35 characters): ";
        chomp ($x = <STDIN>);
        }
        until (length ($x) >= 1 && length ($x) <= 35);
      $command .= ",'" . uc($x) . "' " ;
      last WM_SUBTYPE;
      };

    $subtype == 2 && do
      {
      print "\nEnter vessel number (3-1023, to match the last four digits\nof the MMSI supplied in the subtype 0 message)\n";
      do
        {
        print "--> ";
        chomp ($x = <STDIN>);
        }
        until (looks_like_number ($x) && $x >= 3 && $x <= 1023);
      $command .= "--info68=" . $x ;

      print "\nEnter Descriptor type (describes the text field to follow)\n";
      print "   0. Name (default)               1. IMO Number\n";
      print "   2. State numer                  3-6. (reserved for future use)\n";
      print "   7. Other\n";
      do
        {
        print "--> ";
        chomp ($x = <STDIN>);
        }
        until ((looks_like_number ($x) && $x >= 0 && $x <= 7) || $x eq ""); 
      if ($x eq "") { $x = 0; }
      $command .= "," . $x ;
      
      do
        {
        print "\nEnter vessel name (1-35 characters): ";
        chomp ($x = <STDIN>);
        }
        until (length ($x) >= 1 && length ($x) <= 35);
      $command .= ",'" . uc($x) . "' " ;
      last WM_SUBTYPE;
      };

    $subtype == 3 && do
      {
      print "\nInvalid subtype ($subtype)\n";    
      last WM_SUBTYPE;
      };
    }

  return;
  }

# ********************************************************
# type0608_handler
# ********************************************************

# There are many types of Binary Addressed and Broadcast Messages (Types 6 and 8).
# This routine supports the messages and DAC/FI values as shown below.

sub type0608_handler
  {
  my ($dac,$fi,$i,$x);

# These arrays list the supported Type 6 and 8 message types and DAC/FI pairs
  
  my @type6_list = ( ["",-1,-1], #0 - unused
    ["Monitoring aids to navigation",0,0], #1
    ["Text using 6-bit ASCII",1,0], #2
    ["Interrogation for a specific functional message",1,2], #3
    ["Capability interrogation",1,3], #4
    ["Capability reply",1,4], #5
    ["Number of Persons On Board – Addressed",1,16], #6
    ["Clearance Time to Enter Port – Addressed",1,18], #7
    ["Extended ship static and voyage related data",1,19], #8
    ["Berthing Data – Addressed",1,20], #9
    ["Area Notice – Addressed",1,23], #10
    ["Route Information – Addressed",1,28], #11
    ["Text Description – Addressed",1,30], #12
    ["Number of Persons On Board",1,40], #13
    ["Aids to Navigation Monitoring Data",235,10], #14
    ["Buoy Position Monitoring",235,20], #15
    ["Aids to Navigation Monitoring Data",250,10], #16
    ["Passenger and Crew Count – Addressed",367,16], #17
    ["Geographic Notice – Addressed",367,22], #18
    ["Linked Text Message - Addressed",367,29], #19
    ["Waterways Management Message (v2) - Addressed",367,35], #20
    );
my $type6_list_len = scalar @type6_list - 1;

my @type8_list = ( ["",0-1,-1], #0 - unused
   ["Text using 6-bit ASCII",1,0], #1
   ["VTS Targets",1,16], #2
   ["Advice of Waypoints/VTS Route Plan",1,18], #3
   ["Area Notice – Broadcast",1,22], #4
   ["Route Information – Broadcast",1,27], #5
   ["Text Description – Broadcast",1,29], #6
   ["Meteorological and Hydrological Data",1,31], #7
   ["Dam Status - Broadcast",366,20], #8
   ["Geographic Notice – Broadcast",367,22], #9
   ["Satellite Ship Weather – Broadcast",367,23], #10
   ["Satellite Ship Weather Small – Broadcast",367,24], #11
   ["Satellite Ship Weather Tiny – Broadcast",367,25], #12
   ["GPS Jamming and Spoofing Report",367,26], #13
   ["Linked Text Message - Broadcast",367,29], #14
   ["Environmental Management Message",367,33], #15
   ["Waterways Management Message (v2) - Broadcast",367,35], #16
    );
my $type8_list_len = scalar @type8_list - 1;

# Start by determining which kind of Type 6 or 8 message we have

  if ($type == 6)
      {
      do
        {
        print "\nSpecify the Type 6 message:\n";
        for ($i=1; $i<=$type6_list_len; $i++) { printf "  %2d. %s (DAC=%d, FI=%d)\n", $i, $type6_list[$i][0], $type6_list[$i][1], $type6_list[$i][2]; }
        print "--> ";
        chomp ($x = <STDIN>);
        }
        until (looks_like_number ($x) && $x >= 1 && $x <= $type6_list_len);

        $dac = $type6_list [$x][1];
        $fi = $type6_list [$x][2];

# All Type 6 messages require a sequence number and destination MMSI

      p_mmsi_seq (1);
      }
    else
      {
      do
        {
        print "\nSpecify the Type 8 message:\n";
        for ($i=1; $i<=$type8_list_len; $i++) { printf "  %2d. %s (DAC=%d, FI=%d)\n", $i, $type8_list[$i][0], $type8_list[$i][1], $type8_list[$i][2]; }
        print "--> ";
        chomp ($x = <STDIN>);
        }
        until (looks_like_number ($x) && $x >= 1 && $x <= $type8_list_len);

        $dac = $type8_list [$x][1];
        $fi = $type8_list [$x][2];
      }

# We are here if a valid Type 6 or 8 DAC/FI pair was selected. Now add to the $command string and
# build the messages. (Specify the type below only when needed to avoid ambiguity.)

  $command .= "--dac=" . $dac . " --fi=" . $fi . " ";

  TYPE0608_ENCODE:
    {
# This option handles Type 6, 0/0

    $type == 6 && $dac == 0 && $fi == 0 && do
      {
      type_06_000_00 ();
      p_offposition ();
      last TYPE0608_ENCODE;
      };

# This option handles Types 6 and 8, 1/0

    $dac == 1 && $fi == 0 && do
      {
      type_0608_001_00 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/2 and 1/3

    $type == 6 && $dac == 1 && ($fi == 2 || $fi == 3) && do
      {
      type_06_001_02_03 ($fi);
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/4

    $type == 6 && $dac == 1 && $fi == 4 && do
      {
      type_06_001_04 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/16 and 1/40

    $type == 6 && $dac == 1 && ($fi == 16 || $fi == 40) && do
      {
      type_06_001_16_40 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 1/16

    $type == 8 && $dac == 1 && $fi == 16 && do
      {
      type_08_001_16 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/18

    $type == 6 && $dac == 1 && $fi == 18 && do
      {
      p_linkage_id ();
      p_date (2);
      type_06_001_18 ();
      p_position ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 1/18

    $type == 8 && $dac == 1 && $fi == 18 && do
      {
      type_08_001_18 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/19

    $type == 6 && $dac == 1 && $fi == 19 && do
      {
      type_06_001_19 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/20

    $type == 6 && $dac == 1 && $fi == 20 && do
      {
      p_linkage_id ();
      p_date (2);
      type_06_001_20 ();
      p_position ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/23 & 367/22 *and* Type 8, 1/22 & 367/22

    (($type == 6  && (($dac == 1 && $fi == 23) || ($dac == 367 && $fi == 22))) ||
    ($type == 8 && (($dac == 1 && $fi == 22) || ($dac == 367 && $fi == 22)))) && do
      {
      p_linkage_id ();
      p_date (2);
      type_0608_22_23 ($dac);
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/28 and Type 8, 1/27

    $dac == 1 && ($fi == 27 || $fi == 28) && do
      {
      p_linkage_id ();
      p_date (2);
      type_0608_001_27_28 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 1/30 and Type 8, 1/29

    $dac == 1 && ($fi == 29 || $fi == 30) && do
      {
      p_linkage_id ();
      p_text68 ($type,155,161);
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 1/31

    $type == 8 && $dac == 1 && $fi == 31 && do
      {
      p_date (1);
      type_08_001_31 ();
      p_air_pressure (1,511);
      p_wind (2);
      p_air_temperature ();
      p_position ();
      p_accuracy ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 235/10 and 250/10

    $type == 6 && ($dac == 235 || $dac == 250) && $fi == 10 && do
      {
      type_06_235_250_10 ();
      p_aid_status ();
      p_offposition ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 235/20

    $type == 6 && $dac == 235 && $fi == 20 && do
      {
      type_06_235_20 ();
      p_date (1);
      p_position ();
      p_offposition ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 6, 367/16

    $type == 6 && $dac == 367 && $fi == 16 && do
      {
      p_linkage_id ();
      p_date (1);
      type_06_367_16 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 366/20

    $type == 8 && $dac == 366 && $fi == 20 && do
      {
      p_linkage_id ();
      p_date (0);
      type_08_366_20 ();
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 367/23

    $type == 8 && $dac == 367 && $fi == 23 && do
      {
      p_date (1);
      p_position ();
      p_air_pressure (0,403);
      p_air_temperature ();
      p_wind (1);
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 367/24

    $type == 8 && $dac == 367 && $fi == 24 && do
      {
      p_date (0);
      p_position ();
      p_air_pressure (0,403);
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 367/25

    $type == 8 && $dac == 367 && $fi == 25 && do
      {
      p_date (0);
      p_air_pressure (0,403);
      p_wind (0);
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 367/26

    $type == 8 && $dac == 367 && $fi == 26 && do
      {
      p_date (1);
      p_position ();
      type_08_367_26 ();
      last TYPE0608_ENCODE;
      };

# This option handles Types 6 and 8, 367/29

    $dac == 367 && $fi == 29 && do
      {
      p_linkage_id ();
      p_text68 ($type,154,160);
      last TYPE0608_ENCODE;
      };

# This option handles Type 8, 367/33

    $type == 8 && $dac == 367 && $fi == 33 && do
      {
      type_08_367_33 ();
      last TYPE0608_ENCODE;
      };

# This option handles Types 6 and 8, 367/35

    $dac == 367 && $fi == 35 && do
      {
      p_linkage_id ();
      type_0608_367_35 ();
      last TYPE0608_ENCODE;
      };
    }

  return;
  }

# ********************************************************
# type15_handler
# ********************************************************

# Message Type 15 can have four states and, thus, four formats:

# 1) One MMSI and one interrogation ["MMSI1,query1.1"]
# 2) One MMSI and two interrogations ["MMSI1,query1.1,query1.2"]
# 3) Two MMSIs, each with one interrogation ["MMSI1,query1.1,MMSI2,query2"]
# 4) Two MMSIs, where the first has two interrogations and the second has one ["MMSI1,query1.1,query1.2,MMSI2,query2"]

sub type15_handler
  {
  my $x;

# The command line output is a comma-delimited string with the general format shown above.
# A "query" includes the message type being requested and a TDMA slot offset.

  $command .= "--target=";

# Determine which format we have and then call routines to obtain the information.

  do
    {
    print "\nEnter the message format from the following list:\n";
    print "  1. One MMSI and one interrogation\n";
    print "  2. One MMSI and two interrogations\n";
    print "  3. Two MMSIs, each with one interrogation\n";
    print "  4. One MMSI with two interrogations and a second MMSI with one interrogation\n";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 1 && $x <= 4);

# All formats require one MMSI and one interrogration
# Formats 2 and 4 require a second interrogation associated with the first MMSI
# Formats 3 and 4 require a second MMSI and associated interrogation
  
  p_mmsi15 (1);
  if ($x ==2 || $x == 4) { p_mmsi15 (0); }
  if ($x >= 3) { $command .= ","; p_mmsi15 (1); }
  $command .= " ";

  return;
  }

# ********************************************************
# type16_handler
# ********************************************************

# Message Type 16 can have one or two target MMSI/offset/increment values.

sub type16_handler
  {
  my ($i,$n,$x,$y);

# Ask how many MMSIs we are assigning

  do
    {
    print "\nHow many MMSIs are being assigned (1-2): \n";
    chomp ($n = <STDIN>);
    }
    until (looks_like_number ($n) && $n >= 1 && $n <= 2);

# The command line output is a comma-delimited string with the general format "MMSI,offset,increment,MMSI,offset,increment"

  $command .= "--target=";

  for ($i=1; $i<=$n; $i++)
    {
    do
      {
      print "\nEnter MMSI for interrogated device (9 decimal digits): ";
      chomp ($x = <STDIN>);
      }
      until (verify_numeric_string ($x,9));

    if ($i > 1) { $command .= ","; }
    $command .=  $x . ",";

# The next two values that we need are offset ($x) and increment ($y), but the interpretation of offset depends upon the value of increment. So,
# we get increment first....

    do
      {
      print "\nEnter increment to next slot value (0-7):\n";
      print "  0 = Num. reports/10-min intervals   1 = 1,125 slots   2 = 375 slots   3 = 225 slots\n";
      print "  4 = 125 slots                       5 = 75 slots      6 = 45 slots    7 = disregard assignment\n";
      chomp ($y = <STDIN>);
      }
      until (looks_like_number ($y) && $y >= 0 && $y <= 7);

    if ($y == 0)
        {
        do
          {
          print "\nEnter number of reports per 10-minute intervals (20-600 in intervals of 20): ";
          chomp ($x = <STDIN>);
          }
          until (looks_like_number ($x) && $x >= 20 && $x <= 600 && ($x%20) == 0);
        }
      else
        {
        do
          {
          print "\nEnter offset from current slot to first assigned slot (1-4095): ";
          chomp ($x = <STDIN>);
          }
          until (looks_like_number ($x) && $x >= 1 && $x <= 4095);
        }

    $command .= $x . "," . $y;
    }

  $command .= " ";
  
  return;
  }

# ********************************************************
# type24_handler
# ********************************************************

# Message Type 24 has two parts and three formats! So, we handle that here...

sub type24_handler
  {
  my ($part,$part24,$x);

# First, figure out whether we are Part A or Part B...

  do
    {
    print "\nEnter Static Data Report part number (A or B); Default = A: ";
    chomp ($part = <STDIN>);
    $part = uc($part);
    }
    until ($part eq "A" || $part eq "B" || $part eq ""); 

  $part24 = $part;
  if ($part eq "")
      { $part24 = "A"; }
    else
      { $command .= "--part=" . $part . " "; }

  if ($part24 eq "A")

# Type 24, Part A

      {
      p_vname ();
      }
    else

# Type 24, Part B

      {
      p_vendorinfo ();
      p_callsign ();
      p_vtype ();

# If MMSI starts with "98", get the mothership's MMSI; if not, get the dimensions

      if (substr ($mmsi,0,2) eq "98")
          { 
          do
            {
            print "\nMMSI for this vessel indicates that is an auxiliary craft.\n";
            print "Enter MMSI for mothership (9 decimal digits): ";
            chomp ($x = <STDIN>);
            }
            until (verify_numeric_string ($x,9));

          $command .= "--mothership=" . $x . " ";
          }
        else
          { p_dimensions (); }
      }

  return;
  }

# ********************************************************
# type2526_handler
# ********************************************************

# There are many similarities in the format of the Single Slot (Type 25) and Multiple Slot (Type 26) Binary Messages.
# This routine supports the messages and DAC/FI values as shown below. Note that DAC/FI = 1/0 is an ITU standard,
# while Type 26, DAC/FI 1/x are experimental, used for AIS R-mode messages
# NOTE: While these message types support "binary data," all input is in 6-bit ASCII characters -- except as noted above for R-mode messages

sub type2526_handler
  {
  my ($dac,$fi,$bd_flag,$dest_ind,$max_data,$x);
  my $t26type=0;
 
 # AIS R-mode messages (Type = 26, DAC = 1, FI = 10/11/25) has a fixed binary payload (converted here to hex) plus the required spare bits
 # NOTE: The payload can be changed in AIS_ping...
 
  my @r_mode_payload = (
    "956773DEC8679A84763D0", #0 with four spare bits (0x0)
    "5992A2D96F56E6E41206C41502919257AAACA019811DE2F78426271D546C627323CD8E0F31B00", #1 with eight spare bits (0x00)
    "066B01ECB763CE804996314ACECD3DBCC20796C45451B26E244C068B49D00CCBB2C9D2A6C193552A24A157A4194402B650C551C36E11991363A70E6C0E5CB8BD6AE4BDCF092BC700032AC3D164801C70AE47669916157AC395A555CE869CF60E9BDC9AD35B5DEA2AC66275B9AAC8CA338E84E7980F045DEB5F000", #2 with 12 spare bits (0x000)
    );
  
# These arrays list the supported Types 25 and 26 DAC/FI pairs

  my @type25_dac_fi_list = ([1,0]);
  my @type26_dac_fi_list = ([1,0], [1,10], [1,11], [1,14]);
  my $type25_dac_fi_len = scalar @type25_dac_fi_list;
  my $type26_dac_fi_len = scalar @type26_dac_fi_list;

# max_data is the maximum number of binary message bits, which is dependent upon the Message Type,
# Destination Indicator, and the binary data flag. Set max_data here.

  if ($type == 25)
      { $max_data = 128; }
    else
      { $max_data = 1000; }

# Determine whether this is structured or unstructured binary data. If structured, there will be a DAC/FI pair, otherwise there is none.
# Although the Destination Indicator bit comes first in the message, check the structured bit first because the R-mode DAC/FI pairs will skip
# the Destination Indicator query that follows. If structured, adjust max_data.

  do
    {
    print "\nIs this binary data unstructured (0) or structured (1); Default = 0: ";
    chomp ($bd_flag = <STDIN>);
    }
    until ($bd_flag == 0 || $bd_flag == 1 || $bd_flag eq ""); 
  if ($bd_flag eq "") { $bd_flag = 0; }
    
# If bd_flag indicates structured binary data, determine the DAC/FI pair and adjust max_data

  if ($bd_flag == 1)
    {
    $max_data -= 16;
    if ($type == 25)
        {
        do
          {
          print "\nSpecify the Type 25 message:\n";
          print "  1. Binary Data Message - Single Slot (DAC=001, FI=0)\n";
          chomp ($x = <STDIN>);
          }
          until (looks_like_number ($x) && $x >= 1 && $x <= $type25_dac_fi_len);

          $dac = $type25_dac_fi_list [$x-1][0];
          $fi = $type25_dac_fi_list [$x-1][1];
        }
      else
        {
        do
          {
          print "\nSpecify the Type 26 message:\n";
          print "  1. Binary Data Message - Multiple Slot with Communication State (DAC=001, FI=0)\n";
          print "  2. R-Mode Navigation Message - 1 slot  (DAC=001, FI=10)\n";
          print "  3. R-Mode Navigation Message - 2 slots  (DAC=001, FI=11)\n";
          print "  4. R-Mode Navigation Message - 5 slots  (DAC=001, FI=14)\n";
          chomp ($x = <STDIN>);
          }
          until (looks_like_number ($x) && $x >= 1 && $x <= $type26_dac_fi_len);

          $dac = $type26_dac_fi_list [$x-1][0];
          $fi = $type26_dac_fi_list [$x-1][1];
          $t26type = $x;
        }

    $command .= "--dac=" . $dac . " --fi=" . $fi . " ";
    }

# Determine whether this message is addressed or broadcast (default)
# NOTE: This is not an option for R-mode messages (i.e., DAC = 1, FI = 10, 11, and 14 (aka t26type = 2-4)

  if ($type == 25 || $t26type == 1)
    {
    do
      {
      print "\nIs this message broadcast (0) or addressed to a specific MMSI (1); Default = 0: ";
      chomp ($dest_ind = <STDIN>);
      }
      until ($dest_ind == 0 || $dest_ind == 1 || $dest_ind eq "");
    }
  
  if ($dest_ind eq "" || $t26type > 1) { $dest_ind = 0; }

# If dest_ind indicates an addressed message, get the target MMSI. Adjust max binary data field for MMSI plus spare bits

  if ($dest_ind == 1)
    {
    p_mmsi_seq (0);
    $max_data -= 32;    # Subtract 30 bits for the MMSI plus 2 spare bits for octet-alignment
    }

# Now we need to get the binary data! The data is fixed for R-mode messages. Binary data will be stored here in hex format and octet-aligned

# If this is an AIS R-mode message (t26type > 1), use standard R-mode payloads (R); otherwise, get binary data from the user as text (T)

  if ($t26type > 1)
      { $command .= "--info68=R," . $r_mode_payload [$t26type-2] . " "; }
    else
      {
      do
        {
        print "\nEnter binary message using valid 6-bit ASCII characters (1-$max_data characters):\n";
        chomp ($x = <STDIN>);
        }
        until (length ($x) >= 1 && length ($x) <= $max_data); 

      $command .= "--info68=T,'" . uc($x) . "' ";
      }
      
  return;
  }

# ********************************************************
# verify_numeric_string
# ********************************************************

# Verify that a numeric string has all of the required digits

sub verify_numeric_string # ($num_str, $str_len)
  {
  my ($num_str, $str_len) = @_;

  if ($num_str !~ m/^\d{$str_len}$/)
      { return 0; }
    else
      { return 1; }
  }

########################################################################
# The p_* routines get a variety of parameters which are the building
# blocks for the various messages.
########################################################################

# ********************************************************
# p_accuracy
# ********************************************************

# Position accuracy is used in message types 1-3, 4, 8/1/31, 9, 11, 18, 19, 21, & 27

sub p_accuracy
  {
  my $x;

  do
    {
    print "\nEnter position accuracy (0 = Unaugmented GNSS/>10m [default], 1 = DGPS-quality/<10m): ";
    chomp ($accuracy = <STDIN>);
    }
    until ($accuracy eq "0" || $accuracy eq "1" || $accuracy eq ""); 

  if ($accuracy ne "")
    { $command .= "--accuracy=" . $accuracy . " "; }

  return;
  }

# ********************************************************
# p_aid_status
# ********************************************************

# Set AtonN status word (Type 21 and some Type 6 messages)
# aid_status is a packed five-bit value 000rrllu where:
#    u (bit 0)    = unit status
#   ll (bits 1-2) = light status
#   rr (bits 3-4) = RACON status

sub p_aid_status
  {
  my ($status,$x);

  $status = 0;
  do
    {
    print "\nEnter AtonN unit status (0 = good health, 1 = alarm): ";
    chomp ($x = <STDIN>);
    }
    until ($x == 0 || $x == 1);
  $status += $x;
  
  do
    {
    print "\nEnter light status (0 = no light, 1 = light on, 2 = light off, 3 = light error): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 3); 
  $status += int($x) * 2;
  
  do
    {
    print "\nEnter RACON status (0 = no RACON, 1 = not monitored, 2 = operational, 3 = error): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 3); 
  $status += ($x) * 8;

  $command .= "--aid_status=" . $status . " ";
  return;
  }

# ********************************************************
# p_air_pressure
# ********************************************************

# Routine to get air pressure (and, optionally), air pressure tendency ($opt=1). Used in some weather-related messages

sub p_air_pressure
  {
  my ($opt,$unknown) = @_;
  my ($x);

  do
    {
    print "\nEnter air pressure, in hPa/millibars (500-1500 or 9999 if data not available); Default = 9999: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && ($x >= 500 && $x <= 1500) || $x == 9999) || $x eq ""); 

  if ($x eq "") { $x = 9999; }
  print "     ***** Pressure value $x converted to ";
  AIR_PRESS:
    {
    (int($x) == 9999) && do
      { $x = $unknown; last AIR_PRESS; };
    (int($x) <= 799) && do
      { $x = 0; last AIR_PRESS; };
    (int($x) >= 1201) && do
      { $x = 402; last AIR_PRESS; };
    do
      { $x -= 799; };
    };
  print "$x *****\n";

  $command .= "--pressure=" . int($x) . " ";
  
  if ($opt == 1)
    {
    do
      {
      print "\nEnter pressure tendency (0-3):\n";
      print " 0. Steady     1. Decreasing     2. Increasing     3. Not available\n";
      print "Default = 3: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 3) || $x eq ""); 

    if ($x ne "")
      { $command .= "--presstend=" . int($x) . " "; }
    }
  
  return;
  }
 
# ********************************************************
# p_air_temperature
# ********************************************************

# Air temperature is used in some weather-related messages, e.g., 8 (DAC=1, FI=31) and (DAC=367, FI=23)

sub p_air_temperature
  {
  my $x;

  do
    {
    print "\nEnter dry bulb air temperature, in units of 0.1\xB0C Celsius (-60 to 60)\nor -102.4 if not available; Default = -102.4: ";
    chomp ($x = <STDIN>);
    }
    until (((looks_like_number ($x) && ($x >= -60 && $x <= 60)) || $x == -102.4) || $x eq ""); 

  if ($x ne "")
    { $command .= "--airtemp=" . $x . " "; }

  return;
  }

# ********************************************************
# p_ais_region
# ********************************************************

# AIS coverage region parameters are used in message types 22 & 23
# These parameters define the rectagular AIS jurisdication area by specifying the northeast corner lat and long,
# and the southwest corner lat and long.

sub p_ais_region
  {
  my $x;

  do
    {
    print "\nEnter northeast corner's latitude of AIS rectangular jurisdiction (-90 to 90);\nUse 91 if not available: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= -90 && $x <= 91) || $x eq ""); 
  if ($x ne "")
    { $command .= "--ne_lat=" . $x . " "; }

  do
    {
    print "\nEnter northeast corner's longitude (-180 to 180); Use 181 if not available: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= -180 && $x <= 181) || $x eq ""); 
  if ($x ne "")
    { $command .= "--ne_lon=" . $x . " "; }

  do
    {
    print "\nEnter southwest corner's latitude of AIS rectangular jurisdiction (-90 to 90);\nUse 91 if not available: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= -90 && $x <= 91) || $x eq ""); 
  if ($x ne "")
    { $command .= "--sw_lat=" . $x . " "; }

  do
    {
    print "\nEnter southwest corner's longitude (-180 to 180); Use 181 if not available: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= -180 && $x <= 181) || $x eq ""); 
  if ($x ne "")
    { $command .= "--sw_lon=" . $x . " "; }

  return;
  }

# ********************************************************
# p_alt_speed
# ********************************************************

# Altitude and (air)speed parameters are used in message type 9

sub p_alt_speed
  {
  my $x;

  do
    {
    print "\nEnter airspeed, in knots (0-1021; use 1022 for speeds of 1022 kn or higher;\nuse 1023 if speed not available); Default = 1023: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 1023) || $x eq ""); 

  if ($x ne "")
    { $command .= "--airspeed=" . int($x) . " "; }

  do
    {
    print "\nEnter altitude, in meters (0 to 4093; use 4094 for altitude of 4094 m or higher;\nuse 4095 if altitude not available); Default = 4095: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && ($x >= 0 && $x <= 4095)) || $x eq ""); 

  if ($x ne "")
    { $command .= "--altitude=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_aton
# ********************************************************

# Aids-to-Navigation (AtoN) parameters are used in message 21.

sub p_aton
  {
  my $x;

  do
    {
    print "\nEnter 0 if this is a real AtoN at indicated position; 1 if this is a\nvirtual AtoN simulated by nearby AIS station; Encoder default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ($x eq "0" || $x eq "1" || $x eq ""); 

  if ($x eq "1")
    { $command .= "--v_AtoN" . " "; }

  do
    {
    print "\nEnter type of navigational aid (0-31) from the following list:\n";
    print "  0. AIS Default; type of AtoN not specified\n";
    print "  1. Reference point\n";
    print "  2. RACON (radar transponder marking a navigation hazard)\n";
    print "  3. Fixed structure off shore, such as oil platforms, wind farms, rigs.\n";
    print "  4. Spare, Reserved for future use     5. Light, without sectors\n";
    print "  6. Light, with sectors                7. Leading Light Front\n";
    print "  8. Leading Light Rear                 9. Beacon, Cardinal N\n";
    print " 10. Beacon, Cardinal E                11. Beacon, Cardinal S\n";
    print " 12. Beacon, Cardinal W                13. Beacon, Port hand\n";
    print " 14. Beacon, Starboard hand\n";
    print " 15. Beacon, Preferred Channel port hand\n";
    print " 16. Beacon, Preferred Channel starboard hand\n";
    print " 17. Beacon, Isolated danger           18. Beacon, Safe water\n";
    print " 19. Beacon, Special mark              20. Cardinal Mark N\n";
    print " 21. Cardinal Mark E                   22. Cardinal Mark S\n";
    print " 23. Cardinal Mark W                   24. Port hand Mark\n";
    print " 25. Starboard hand Mark               26. Preferred Channel Port hand\n";
    print " 27. Preferred Channel Starboard hand\n";
    print " 28. Isolated danger                   29. Safe Water\n";
    print " 30. Special Mark                      31. Light Vessel / LANBY / Rigs\n";
    print " Encoder default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 31) || $x eq ""); 

  if ($x ne "")
    { $command .= "--aid_type=" . int($x) . " "; }

  do
    {
    print "\nEnter AtoN name (1-20 characters); Encoder default = '@@@@@@@@@@@@@@@@@@@@':\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 20); 

  if ($x ne "")
    { $command .= "--aid_name='" . uc($x) . "' "; }

  return;
  }

# ********************************************************
# p_callsign
# ********************************************************

# Vessel callsign is part of messages type 5 or type 24 (Part B format)

sub p_callsign
  {
  my $x;

  do
    {
    print "\nEnter vessel call sign (1-7 characters):\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 7); 

  if ($x ne "")
    { $command .= "--callsign='" . uc($x) . "' "; }

  return;
  }

# ********************************************************
# p_channel
# ********************************************************

# Channel management parameters apply only to message type 22
# These parameters specify the frequency for AIS channels A and B based upon the
# ITU frequency designators

sub p_channel
  {
  my $x;

  do
    {
    print "\nEnter AIS Channel A channel number, from ITU-R Rec. M.1084;\nEncoder default = 2087 (87B = 161.975 MHz): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0) || $x eq ""); 

  if ($x ne "")
    { $command .= "--channel_a=" . int($x) . " "; }

  do
    {
    print "\nEnter AIS Channel B channel number, from ITU-R Rec. M.1084;\nEncoder default = 2088 (88B = 162.025 MHz): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0) || $x eq ""); 

  if ($x ne "")
    { $command .= "--channel_b=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_cog
# ********************************************************

# Course over ground is in message types 1-3, 9, 18, 19 (see also 27)

sub p_cog
  {
  my $x;

  do
    {
    print "\nEnter course over ground, in degrees (0 to 359.9; use 360 if data not available)\nDefault = 360 (N/A): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 

  if ($x ne "")
    { $command .= "--course=" . $x . " "; }


  return;
  }

# ********************************************************
# p_date
# ********************************************************

# Date information is in Types 4, 6, 8, and 11
# Types 4 & 11: year, month, day/hour/minute, second (opt = 4)
# Type 6 and Type 8 need day/hour/minute (opt = 1) and, sometimes, month (opt = 2)
# Some Type 6/8 messages only need hour and minute (opt = 0)

# OPTION Year Month Day Hour Min Sec
#   0                    X    X
#   1    2001   1    X   X    X
#   2    2001   X    X   X    X
#   4     X     X    X   X    X   X


sub p_date
  {
  my ($opt) = @_;
  my ($month,$x,$year);
  my @mdays = (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

# Dates are checked for validity but that is difficult with options 1 and 2! In these cases, set default values...

  if ($opt == 1) { $month = 1; }
  if ($opt == 1 || $opt == 2) { $year = 2001; } # Assume a non-leap year if no year is given
  
  if ($opt == 4)
    {
    do
      {
      print "\nEnter UTC year (1-9999 or 0 if not available); Default = 0: ";
      chomp ($year = <STDIN>);
      }
      until ((looks_like_number ($year) && $year >= 0 && $year <= 9999) || $year eq ""); 

    if ($year ne "")
      { $command .= "--year=" . int($year) . " "; }
      
# If leap year, change number of days in February

    if ((($year % 400) == 0) || (($year % 4) == 0 && ($year % 100) != 0)) { $mdays [2] = 29; }
    }

  if ($opt >= 2)
    {
    do
      {
      print "\nEnter UTC month (1-12 or 0 if not available); Default = 0: ";
      chomp ($month = <STDIN>);
      }
      until ((looks_like_number ($month) && $month >= 0 && $month <= 12) || $month eq ""); 

    if ($month ne "")
      { $command .= "--month=" . int($month) . " "; }
    }

  if ($opt >= 1)
    {
    do
      {
      printf "\nEnter UTC day (1-%d or 0 if not available); Default = 0: ", $mdays [$month];
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= $mdays [$month]) || $x eq ""); 

    if ($x ne "")
      { $command .= "--day=" . int($x) . " "; }
    }

  do
    {
    print "\nEnter UTC hour (0-23 or 24 if not available); Default = 24: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 24) || $x eq ""); 

  if ($x ne "")
    { $command .= "--hour=" . int($x) . " "; }

  do
    {
    print "\nEnter UTC minute (0-59 or 60 if not available); Default = 60: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 60) || $x eq ""); 

  if ($x ne "")
    { $command .= "--minute=" . int($x) . " "; }

  if ($opt == 4)
    {
    do
      {
      print "\nEnter UTC second (0-59 or 60 if not available); Default = 60: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 60) || $x eq ""); 

    if ($x ne "")
      { $command .= "--second=" . int($x) . " "; }
    }

  return;
  }

# ********************************************************
# p_dgnss_msg
# ********************************************************

# Message Type 17 contains an ITU-R Rec. M.823 GNSS message. Get here the GNSS message common information and then a
# binary/hex string representing the message data (0-736 bits/0-92 bytes)

sub p_dgnss_msg
  {
  my ($i,$in_type,$n,$x);

  # Get all of the M.823 DGNSS message common information
  
  do
    {
    print "\nEnter the M.823 Message Type:\n";
    print "   1. GPS Differential GNSS corrections (full set of satellites)\n";
    print "   3. GPS Reference station parameters\n";
    print "   4. GPS/GLONASS Reference Station Datum\n";
    print "   5. GPS Constellation health\n";
    print "   6. GPS Null frame\n";
    print "   7. GPS Radio beacon almanacs\n";
    print "   9. GPS Subset differential GNSS corrections\n";
    print "  16. GPS Special message\n";
    print "  27. GPS/GLONASS Extended radio beacon almanac\n";
    print "  31. GLONASS Differential GNSS corrections (full set of satellites)\n";
    print "  32. GLONASS Reference station parameters\n";
    print "  33. GLONASS Constellation health\n";
    print "  34. GLONASS Null frame (0 or 1 data words)\n";
    print "  34. GLONASS Subset differential GNSS corrections (>1 data word)\n";
    print "  35. GLONASS Radio beacon almanacs\n";
    print "  36. GLONASS Special message\n";
    print "--> ";
    chomp ($x = <STDIN>);
    }
    until ($x == 1 || ($x >= 3 && $x <= 7) || $x == 9 || $x == 16 || $x == 27 || ($x >= 31 && $x <= 36));
  $command .= "--dgnss_msg=" . $x;
    
  do
    {
    print "\nEnter the M.823 station identifier (0-1023): ";
    chomp ($x = <STDIN>);
    }
    until ($x >= 0 && $x <= 1023);
  $command .= "," . $x;

  do
    {
    print "\nEnter the number of 0.6 s increments in Z count (0-5999 where 5999 = 3,599.4 s): ";
    chomp ($x = <STDIN>);
    }
    until ($x >= 0 && $x <= 5999);
  $command .= "," . $x;

  do
    {
    print "\nEnter the message sequence number (0-7): ";
    chomp ($x = <STDIN>);
    }
    until ($x >= 0 && $x <= 7);
  $command .= "," . $x;

  do
    {
    print "\nEnter the number of DGNSS data words (0-29): ";
    chomp ($n = <STDIN>);
    }
    until ($n >= 0 && $n <= 29);
  $command .= "," . $n;

  do
    {
    print "\nEnter the reference station health (0-5 = healthy, 6 = transmission is unmonitored,\n";
    print "   7 = station not working properly): ";
    chomp ($x = <STDIN>);
    }
    until ($x >= 0 && $x <= 7);
  $command .= "," . $x;

  if ($n > 0)
    {
    do
      {
      print "\nEnter data words in binary (B) or hex (H): ";
      chomp ($in_type=<STDIN>);
      $in_type = uc ($in_type);
      }
      until ($in_type eq "B" || $in_type eq "H");

    for ($i=1; $i<=$n; $i++)
      {
      if ($in_type eq "B")
          {
          do
            {
            print "\nEnter data word $i as a 24-bit binary string: ";
            chomp ($x=<STDIN>);
            }
            until (length ($x) == 24 && ($x =~ /[0-1]{24}/))
          }
        else
          {
          do
            {
            print "\nEnter data word $i as six hex digits: ";
            chomp ($x=<STDIN>);
            $x = uc ($x);
            }
            until (length ($x) == 6 && ($x =~ /[0-9A-F]{6}/))
          }

      $command .= "," . $x;
      }    
    }
    
  $command .= " ";
  return;
  }

# ********************************************************
# p_dimensions
# ********************************************************

# Vessel dimensions are part of message types 5, 19, 21, and type 24, Part B format (with MMSI not starting with "98")

sub p_dimensions
  {
  my $x;

  do
    {
    print "\nEnter distance from AIS antenna to bow (Dimension A), in meters (0-511): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 511) || $x eq "");

  if ($x ne "")
    { $command .= "--vsize_a=" . int($x) . " "; }

  do
    {
    print "\nEnter distance from AIS antenna to stern (Dimension B), in meters (0-511): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 511) || $x eq "");

  if ($x ne "")
    { $command .= "--vsize_b=" . int($x) . " "; }

  do
    {
    print "\nEnter distance from AIS antenna to port (Dimension C), in meters (0-63): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 63) || $x eq "");

  if ($x ne "")
    { $command .= "--vsize_c=" . int($x) . " "; }

  do
    {
    print "\nEnter distance from AIS antenna to starboard (Dimension D), in meters (0-63): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 63) || $x eq "");

  if ($x ne "")
    { $command .= "--vsize_d=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_dte
# ********************************************************

# Data Terminal Equipment (DTE) status is used in message types 5, 9, & 19

sub p_dte
  {
  my $x;

  do
    {
    print "\nEnter data terminal status (0 = ready, 1 = not ready [default]): ";
    chomp ($dte = <STDIN>);
    }
    until ($dte eq "0" || $dte eq "1" || $dte eq ""); 

  if ($dte ne "")
    { $command .= "--dte=" . $dte . " "; }

  return;
  }

# ********************************************************
# p_epfd
# ********************************************************

# The Electronic Position Fixing Device parameter is used in message types 4, 5, 11, 19, and 21

sub p_epfd
  {
  my $x;

  do
    {
    print "\nEnter Electronic Position Fixing Device (0-15), encoded as follows:\n";
    print " 0. Undefined                      1. GPS\n";
    print " 2. GLONASS                        3. Combined GPS/GLONASS\n";
    print " 4. Loran-C                        5. Chayka\n";
    print " 6. Integrated navigation system   7. Surveyed\n";
    print " 8. Galileo                        9-15. Unused\n";
    print "Encoder default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 15) || $x eq ""); 

  if ($x ne "")
    { $command .= "--epfd=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_fatdma
# ********************************************************

# Fixed Access Time Division Multiple Access (FATDMA) parameters are used in message type 20
# It has been difficult to get exact meaning and values, but I am allowing values based upon the field sizes.
# See also https://fccid.io/UYW-4230002/User-Manual/User-manual-2545256.html
# There can be from 1-4 blocks of information so there will be 4, 8, 12, or 16 items of information here

sub p_fatdma
  {
  my ($fatdmaincrement,$fatdmaoffset,$fatdmaslots,$fatdmatimeout,$i,$x);

  $command .= "--fatdma=";
  for ($i=1; $i<=4; $i++)
    {
    do
      {
      print "\nEnter FATDMA reserved offset number (0-4095); -1 when done, Encoder default = 0: ";
      chomp ($fatdmaoffset = <STDIN>);
      }
      until ((looks_like_number ($fatdmaoffset) && $fatdmaoffset >= -1 && $fatdmaoffset <= 4095) || $fatdmaoffset eq ""); 
    if ($fatdmaoffset == -1)
        { last; }
      else
        {
        if ($fatdmaoffset == "") { $fatdmaoffset = 0; }
        }

    do
      {
      print "\nEnter FATDMA number of reserved slots (0-15); Encoder default = 0: ";
      chomp ($fatdmaslots = <STDIN>);
      }
      until ((looks_like_number ($fatdmaslots) && $fatdmaslots >= 0 && $fatdmaslots <= 15) || $fatdmaslots eq ""); 
      if ($fatdmaslots == "") { $fatdmaslots = 0; }

    do
      {
      print "\nEnter FATDMA timeout, in minutes (0-7); Encoder default = 0: ";
      chomp ($fatdmatimeout = <STDIN>);
      }
      until ((looks_like_number ($fatdmatimeout) && $fatdmatimeout >= 0 && $fatdmatimeout <= 7) || $fatdmatimeout eq ""); 
      if ($fatdmatimeout == "") { $fatdmatimeout = 0; }

    do
      {
      print "\nEnter FATDMA repeat increment (0-2047); Encoder default = 0: ";
      chomp ($fatdmaincrement = <STDIN>);
      }
      until ((looks_like_number ($fatdmaincrement) && $fatdmaincrement >= 0 && $fatdmaincrement <= 2047) || $fatdmaincrement eq ""); 
      if ($fatdmaincrement == "") { $fatdmaincrement = 0; }

    if ($i>1) { $command .= ","; }
    $command .= int($fatdmaoffset) . "," . int($fatdmaslots) . "," . int($fatdmatimeout) . "," . int($fatdmaincrement);
    }
  $command .= " ";

  return;
  }

# ********************************************************
# p_hdg
# ********************************************************

# Heading is used in message types 1-3, 18 & 19

sub p_hdg
  {
  my $x;

  do
    {
    print "\nEnter true heading, in degrees (0 to 359; use 511 if data not available)\nDefault = 511 (N/A): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && ($x >= 0 && $x <= 359) || $x == 511) || $x eq ""); 

  if ($x ne "")
    { $command .= "--heading=" . $x . " "; }

  return;
  }

# ********************************************************
# p_interval
# ********************************************************

# The report interval and quiet time parameters are only used in message type 23

sub p_interval
  {
  my $x;

  do
    {
    print "\nEnter station reporting interval (0-15) from the following list:\n";
    print "  0. As given by the autonomous mode     1. 10 Minutes\n";
    print "  2. 6 Minutes      3. 3 Minutes         4. 1 minute\n";
    print "  5. 30 seconds     6. 15 Seconds        7. 10 Seconds\n";
    print "  8. 5 Seconds      9. Next Shorter Reporting Interval\n";
    print " 10. Next Longer Reporting Interval\n";
    print " 11-15. Reserved for future use\n";
    print " Encoder default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 15) || $x eq ""); 

  if ($x ne "")
    { $command .= "--interval=" . int($x) . " "; }

do
    {
    print "\nEnter station quiet time, in minutes  (1-15; 0 = none/default):\n";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 15) || $x eq ""); 

  if ($x ne "")
    { $command .= "--quiet=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_linkage_id
# ********************************************************

# Some Binary Messages (Types 6 and 8) need a Linkage Identifier.

sub p_linkage_id
  {
  my $x;

  do
    {
    print "\nEnter Message Linkage Identifier (1 to 1023; default = 0 [not available]): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 1023) || $x eq "");

  if ($x ne "")
    { $command .= "--msglink=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_mmsi_seq
# ********************************************************

# Types 10, 25, and 26 messages ask for a target MMSI ($opt = 0)
# Types 6 and 12 messages ask for a target MMSI and 2-bit sequence number ($opt = 1)

sub p_mmsi_seq
  {
  my ($opt) = @_;
  my $x;

  do
    {
    print "\nEnter MMSI for target device (9 decimal digits): ";
    chomp ($x = <STDIN>);
    }
    until (verify_numeric_string ($x,9));

  $command .= "--target=" . $x;

  if ($opt > 0)
    {
    do
      {
      print "\nEnter sequence number (0-3): ";
      chomp ($x = <STDIN>);
      }
      until (looks_like_number ($x) && $x >= 0 && $x <= 3);

    $command .= "," . $x;
    };

  $command .= " ";

  return;
  }

# ********************************************************
# p_mmsi_seq4
# ********************************************************

# The Binary Acknowledge (type 7) and Safety-Related Acknowledgement (type 13) messages
# ask for up to four target MMSI/sequence number pairs...

sub p_mmsi_seq4
  {
  my ($j,$num,$target);

  $command .= "--target=";

  for ($j=1; $j<=4; $j++)
    {
    do
      {
      print "\nEnter target MMSI #$j (9 decimal digits or <ENTER> when done): ";
      chomp ($target = <STDIN>);
      }
      until (verify_numeric_string ($target,9) || $target eq "");

    if ($target eq "")
        { last; }

    do
      {
      print "\nEnter sequence number (0-3): ";
      chomp ($num = <STDIN>);
      }
      until (looks_like_number ($num) && $num >= 0 && $num <= 3);

      if ($j>1)
        { $command .= ","; }
      $command .= $target . "," . $num;
    };

  $command .= " ";

  return;
  }

# ********************************************************
# p_mmsi15
# ********************************************************

# Obtain a MMSI and interrogation information for message type 15
# If the mmsi_flag = 1, ask for a MMSI value and interrogation information; if mmsi_flag = 0, we don't need a MMSI

sub p_mmsi15
  {
  my ($mmsi_flag) = @_;
  my $x;

  if ($mmsi_flag == 1)
    {
    do
      {
      print "\nEnter MMSI for interrogated device (9 decimal digits): ";
      chomp ($x = <STDIN>);
      }
      until (verify_numeric_string ($x,9));

    $command .= "$x";
    }
    
  do
    {
    print "\nEnter message type being interrogated (3-5, 9, 18, 21, 24): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && (($x >= 3 && $x <= 5) || $x == 9 || $x == 18 || $x == 21 || $x == 24));

  $command .= "," . $x;
    
  do
    {
    print "\nEnter TDMA slot offset (0-4095): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && $x >= 0 && $x <= 4095);

  $command .= "," . $x;

  return;
  }

# ********************************************************
# p_navstat
# ********************************************************

# Navigation Status parameter is used in messages 1-3 & 27.

sub p_navstat
  {
  my $x;

  do
    {
    print "\nEnter navigation status (0-15) from the following list:\n";
    print "  0. Underway using engine\n";
    print "  1. At anchor\n";
    print "  2. Not under command\n";
    print "  3. Restricted ability to maneuver\n";
    print "  4. Constrained by draft\n";
    print "  5. Moored\n";
    print "  6. Anchored\n";
    print "  7. Engaged in fishing\n";
    print "  8. Underway sailing\n";
    print "  9-13. Reserved for future use\n";
    print " 14. AIS-SART is active\n";
    print " 15. Not defined\n";
    print "Default = 15: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 15) || $x eq "");  

  if ($x ne "")
    { $command .= "--navstat=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_offposition
# ********************************************************

# The Off-Position Indicator flag is used in message type 21 and several type 6 messages

sub p_offposition
  {
  my $x;

  do
    {
    print "\nEnter off_position flag (0 = on position [default], 1 = off position): ";
    chomp ($x = <STDIN>);
    }
    until ($x eq "0" || $x eq "1" || $x eq ""); 

  if ($x ne "")
    { $command .= "--offpos=" . $x . " "; }

  return;
  }

# ********************************************************
# p_position
# ********************************************************

# Position parameters (latitude, longitude) are used in message types 1-4, 6 (FI=18/20), 8 (FI=31), 9, 11, 18, 19, 21, & 27

sub p_position
  {
  my $x;

  do
    {
    print "\nEnter latitude (-90 to 90) or 91 if not available; Default = 91: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= -90 && $x <= 91) || $x eq ""); 

  if ($x ne "")
    { $command .= "--lat=" . $x . " "; }

  do
    {
    print "\nEnter longitude (-180 to 180) or 181 if not available; Default = 181: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= -180 && $x <= 181) || $x eq ""); 

  if ($x ne "")
    { $command .= "--long=" . $x . " "; }

  return;
  }

# ********************************************************
# p_raim
# ********************************************************

# The Receiver Autonomous Integrity Monitoring (RAIM) flag is used in message types 1-3, 4, 9, 11, 18, 19, 21, & 27

sub p_raim
  {
  my $x;

  do
    {
    print "\nEnter RAIM flag (0 = RAIM not in use [default], 1 = RAIM in use): ";
    chomp ($x = <STDIN>);
    }
    until ($x eq "0" || $x eq "1" || $x eq ""); 

  if ($x ne "")
    { $command .= "--raim=" . $x . " "; }

  return;
  }

# ********************************************************
# p_rot
# ********************************************************

# Rate of turn (ROT) parameter is used in messages 1-3.

sub p_rot
  {
  my $x;

  do
    {
    print "\nEnter rate of turn (-127 - 128), encoded as follows:\n";
    print " -127 = turning left at more than 5 deg/30 sec (No turn info available)\n";
    print " -1...-126 = turning left at up to 708 degrees per minute or higher\n";
    print " 0 = not turning\n";
    print " 1...126 = turning right at up to 708 degrees per minute or higher\n";
    print " 127 = turning right at more than 5 deg/30 s (No turn info available)\n";
    print " 128 = no turn information available\n";
    print "Default = 128: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= -127 && $x <= 128) || $x eq ""); 

  if ($x ne "")
    { $command .= "--rot=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_sart_msg
# ********************************************************

# The sart_msg parameter is used in message types 12 and 14
# The Search and Rescue Transponder (SART) message can be up to 156 (type 12) or 161 (type 14) characters in length

sub p_sart_msg
  {
  my ($sart_msg,$sm_max,$x);

  if ($type == 12)
      { $sm_max = 156; }
    else
      { $sm_max = 161; }
  do
    {
    print "\nEnter SART message (1-$sm_max characters); Encoder default = 'SART ACTIVE':\n";
    chomp ($sart_msg = <STDIN>);
    }
    until (length ($sart_msg) <= $sm_max); 

  if ($sart_msg ne "")
    { $command .= "--sart_msg='" . uc($sart_msg) . "' "; }

  return;
  }

# ********************************************************
# p_sog
# ********************************************************

# Speed is used in message types 1-3, 18, & 19 (see also 27)

sub p_sog
  {
  my $x;

  do
    {
    print "\nEnter speed, in knots, in 0.1 kn resolution (0-102.1;\nuse 102.2 for speeds of 102.2 kn or higher;\nuse 102.3 if speed not available); Default = 102.3: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 102.3) || $x eq ""); 

  if ($x ne "")
    { $command .= "--speed=" . $x . " "; }

  return;
  }

# ********************************************************
# p_text68
# ********************************************************

# Several Type 6 and 8 messages need large blocks of text. This routine gets such text messages. The first parameters is the message type
# and the next two parameters are the maximumm length of the text for Type 6 and 8, respectively.

sub p_text68
  {
  my ($type,$max_6,$max_8) = @_;
  my ($max_chars,$x);

  if ($type == 6) { $max_chars = $max_6; }
    else { $max_chars = $max_8; }
  do
    {
    printf "\nEnter up to $max_chars free 6-bit text characters:\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) >= 1 && length ($x) <= $max_chars);
    
  $command .= "--info68='" . uc ($x) . "' ";
  return;
  }

# ********************************************************
# p_timestamp
# ********************************************************

# Timestamp is used in message types 1-3, 9, 18, 19, & 21

sub p_timestamp
  {
  my $x;

  do
    {
    print "\nEnter UTC timestamp, in seconds (0-59; 60 = time stamp not available;\n";
    print "61 = positioning system is in manual input mode;\n62 = positioning system operates in estimated [dead reckoning] mode,\n";
    print "63 = positioning system is inoperative); Standard default = 60: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 63) || $x eq ""); 

  if ($x ne "")
    { $command .= "--ts=" . int($x+0.5) . " "; }

  return;
  }

# ********************************************************
# p_type05
# ********************************************************

# These parameters are only used for Type 5 messages:
# The International Maritime Organization (IMO) Number is a seven-digit
# number; other parameters are ETA (mon/day/hour/min), draft, and destination

sub p_type05
  {
  my $x;

  do
    {
    print "\nEnter IMO Number (7 decimal digits): ";
    chomp ($x = <STDIN>);
    }
    until (verify_numeric_string ($x,7) || $x eq "");

  if ($x ne "")
    { $command .= "--imo=" . $x . " "; }

  do
    {
    print "\nEnter UTC month of ETA (1-12 or 0 if not available); Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 12) || $x eq ""); 

  if ($x ne "")
    { $command .= "--month=" . int($x) . " "; }

  do
    {
    print "\nEnter UTC day of ETA (1-31 or 0 if not available); Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 31) || $x eq ""); 

  if ($x ne "")
    { $command .= "--day=" . int($x) . " "; }

  do
    {
    print "\nEnter UTC hour of ETA (0-23 or 24 if not available); Default = 24: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 24) || $x eq ""); 

  if ($x ne "")
    { $command .= "--hour=" . int($x) . " "; }

  do
    {
    print "\nEnter UTC minute of ETA (0-59 or 60 if not available); Default = 60: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 60) || $x eq ""); 

  if ($x ne "")
    { $command .= "--minute=" . int($x) . " "; }

  do
    {
    print "\nEnter draft, in meters, in 0.1 m increments (0-25.5); Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 25.5) || $x eq ""); 

  if ($x ne "")
    { $command .= "--draft=" . $x . " "; }

  do
    {
    print "\nEnter Destination (1-20 characters; see\nhttps://unece.org/trade/cefact/unlocode-code-list-country-and-territory for UN/LOCODE list): ";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 20); 

  if ($x ne "")
    { $command .= "--dest='" . uc($x) . "' "; }

  return;
  }

# ********************************************************
# p_type27
# ********************************************************

# These parameters are only used for Type 27 messages:
# Course and speed, albeit differently from other messages, and GNSS Position status

sub p_type27
  {
  my $x;

  do
    {
    print "\nEnter speed, in knots (0-62; use 63 [default] if speed not available): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 63) || $x eq ""); 

  if ($x ne "")
    { $command .= "--speed27=" . int($x) . " "; }

  do
    {
    print "\nEnter course over ground, in degrees (0 to 359; use 511 if data not available)\nDefault = 511 (N/A): ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && ($x >= 0 && $x <= 359) || $x == 511) || $x eq ""); 

  if ($x ne "")
    { $command .= "--course27=" . int($x) . " "; }

  do
    {
    print "\nEnter GNSS Position status (0 = current GNSS position, 1 = not GNSS position [default]): ";
    chomp ($x = <STDIN>);
    }
    until ($x eq "0" || $x eq "1" || $x eq ""); 

  if ($x ne "")
    { $command .= "--gnss_pos=" . $x . " "; }

  return;
  }

# ********************************************************
# p_vendorinfo
# ********************************************************

# Vendor information is used by Message 24 (Part B)

sub p_vendorinfo
  {
  my $x;

  do
    {
    print "\nEnter vendor ID (1-3 characters):\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 3); 

  if ($x ne "")
    { $command .= "--vendor='" . uc($x) . "' "; }

  do
    {
    print "\nEnter unit model number (0-15); default - 0:\n";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 15) || $x eq ""); 

  if ($x ne "")
    { $command .= "--model=" . int($x) . " "; }

  do
    {
    print "\nEnter serial number (0-1048575); default - 0:\n";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 1048575) || $x eq ""); 

  if ($x ne "")
    { $command .= "--serial=" . int($x) . " "; }

  return;
  }

# ********************************************************
# p_vessel_info
# ********************************************************

# Obtain vessel information for Waterways Management messages

sub p_vessel_info
  {
  my ($num) = @_;
  my ($info_string,$x);

# Get MMSI
  
  print "\n\n ********** Enter vessel information for vessel #$num **********\n\n";
  
  print "Enter vessel's MMSI (9 decimal digits). The following non-valid 10-digit MMSIs have special meaning:\n";
  print "0000000000 = No vessels in queue [default]\n";
  print "1000000000 = additional lockage for MMSI previous in sequence\n";
  print "1000000001 = non-AIS vessel (e.g. pleasure boats)\n";
  print "1000000002 = ice/debris\n";
  print "1000000003-1000001023 = unknown MMSI\n";
  do
    {
    print "--> ";
    chomp ($x = <STDIN>);
    }
    until (verify_numeric_string ($x,9) || verify_numeric_string ($x,10) ||$x eq "");
  if ($x eq "") { $x = "0000000000"; }
  $info_string .= $x . ",";
  
# Get day, hour, and minute  

  do
    {
    print "\nEnter UTC day (1-31 or 0 if not available); Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 31) || $x eq ""); 
  if ($x eq "") { $x = 0; }
  $info_string .= int($x) . ",";

  do
    {
    print "\nEnter UTC hour (0-23 or 24 if not available); Default = 24: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 24) || $x eq ""); 
  if ($x eq "") { $x = 24; }
  $info_string .= int($x) . ",";

  do
    {
    print "\nEnter UTC minute (0-59 or 60 if not available); Default = 60: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 60) || $x eq ""); 
  if ($x eq "") { $x = 60; }
  $info_string .= int($x) . ",";

# Now get vessel status, chamber, direction, and lockage type
  
  do
    {
    print "\nEnter vessel status (0 = in the lock, 1 = vessel arrived at check-in point): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && ($x == 0 || $x == 1));
  $info_string .= int($x) . ",";
  
  do
    {
    print "\nEnter chamber number (1-7 or 0 if unknown/unapplicable); Default = 0: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 7) || $x eq ""); 
  if ($x eq "") { $x = 0; }
  $info_string .= int($x) . ",";
  
  do
    {
    print "\nEnter direction of the ship as defined by the river flow or channel markings\n(0 = up/in, 1 = down/out): ";
    chomp ($x = <STDIN>);
    }
    until (looks_like_number ($x) && ($x == 0 || $x == 1));
  $info_string .= int($x) . ",";

  print "\nEnter type of lockage per USACE system [default = 0]:\n";
  print "0. Undefined/not applicable    1. Open pass (O)         2. Fast double (F)    3. Jackknife (J)\n";
  print "4. Knockout (K)                5. Navigable pass (N)    6. Straight (S)      7. Barge transfer (T)\n";
  print "8. Setove (V)                  9. Other (Z)\n";
  do
    {
    print "--> ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 9) || $x eq ""); 
  if ($x eq "") { $x = 0; }
  $info_string .= int($x);
 
  return $info_string;
  }

# ********************************************************
# p_vname
# ********************************************************

# Vessel name is used by Messages 5, 19, and 24 (Part A)

sub p_vname
  {
  my $x;

  do
    {
    print "\nEnter vessel name (1-20 characters); Encoder default = NaN:\n";
    chomp ($x = <STDIN>);
    }
    until (length ($x) <= 20); 

  if ($x ne "")
    { $command .= "--vname='" . uc($x) . "' "; }

  return;
  }

# ********************************************************
# p_vtype
# ********************************************************

# Vessel type is part of messages types 5, 19, 23, and 24 (Part B format)

sub p_vtype
  {
  my $x;

  do
    {
    print "\nEnter vessel type (0-99) from the following list:\n";
    print "  0. Not available, AIS default    1-19. Reserved\n";
    print " 20-29. Wing in ground (WIG)\n";
    print " 30. Fishing                       31. Towing ahead or alongside\n";
    print " 32. Towing astern, length exceeds 200m or breadth exceeds 25m\n";
    print " 33. Dredging or underwater ops    34. Diving operationss\n";
    print " 35. Military operations           36. Sailing vessels\n";
    print " 37. Pleasure Craft                38-39. reserved\n";
    print " 40-49. High speed craft (HSC)     50. Pilot Vessel\n";
    print " 51. Search and Rescue vessel      52. Tug\n";
    print " 53. Port Tender                   54. Anti-pollution equipment\n";
    print " 55. Law Enforcement               56-57. spare\n";
    print " 58. Medical Transport             59. Noncombatant ship\n";
    print " 60-69. Passenger ship             70-79. Cargo\n";
    print " 80-89. Tanker                     90-99. Other ship type\n";
    print " >>> ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 99) || $x eq ""); 

  if ($x ne "")
    { $command .= "--vtype=" . int($x) . " "; }

  return;
  }
  
# ********************************************************
# p_wind
# ********************************************************

# Various messages need a variety of wind information. If $opt = 0, just get the wind speed and direction. For $opt = 1,
# also get the gust speed. If $opt = 2, get all four wind parameters. 

sub p_wind
  {
  my ($opt) = @_;
  my $x;
  print "***** opt $opt\n";
  do
    {
    print "\nEnter 10-minute average wind speed, in knots (0 to 126; 127 if data not\navailable); Default = 127: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 127) || $x eq ""); 

  if ($x ne "")
    { $command .= "--wspeed=" . int($x) . " "; }

  do
    {
    print "\nEnter wind direction, in degrees (0 to 359; 360 if data not available);\nDefault = 360: ";
    chomp ($x = <STDIN>);
    }
    until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 

  if ($x ne "")
    { $command .= "--wdir=" . int($x) . " "; }

  if ($opt >= 1)
    {
    do
      {
      print "\nEnter 10-minute maximum wind speed, in knots (0 to 126; 127 if data not\navailable); Default = 127: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 127) || $x eq ""); 

    if ($x ne "")
      { $command .= "--wgust=" . int($x) . " "; }
    }

  if ($opt == 2)
    {
    do
      {
      print "\nEnter wind gust direction, in degrees (0 to 359; 360 if data not\navailable); Default = 360: ";
      chomp ($x = <STDIN>);
      }
      until ((looks_like_number ($x) && $x >= 0 && $x <= 360) || $x eq ""); 

    if ($x ne "")
      { $command .= "--wgustdir=" . int($x) . " "; }
    }

  return;
  }
