#!/usr/bin/perl -w
use Getopt::Long qw(GetOptions);
use File::Spec;
use Scalar::Util qw(looks_like_number);

# Program AIS_ping
# (c) Gary C. Kessler, 2018-2023

## See README.TXT for background commentary and version information. ##

$build_date = "03/04/2024";
$version = "7.3.7";

# Set default values for all switches so that unsupplied values don't cause "uninitialized value" errors

$aid_name = $callsign = $dest = $dgnss_msg = $vendor = $vname = "";
$aid_type = $aid_status = $draft = $epfd = $imo = $v_aton = $vsize_a = $vsize_b = $vsize_c = $vsize_d = $vtype = 0;
$interval = $quiet = $year = $month = $day = $mothership = $model = $serial = $offpos = 0;
$accuracy = $berth_len = $berth_dep = $mooring = $pob = $raim = $svc = $tos = $routetype = $sndrclass = 0;
$fatdma = $info68 = $lockinfo = $target = "";
$msglink = $wmtype = 0;
$airspeed = 1023;
$airtemp = -102.4;
$altitude = 4095;
$channel_a = 2087;
$channel_b = 2088;
$course = $wdir = $wgustdir = $swelldir = $wavedir = 360;
$dac = $fi = 0;
$dte = $gnss_pos = 1;
$duration = 262143;
$hour = 24;
$ice = $presstend = 3;
$lat = $ne_lat = $sw_lat = 91;
$long = $ne_lon = $sw_lon = 181;
$minute = 60;
$navstat = 15;
$part = "A";
$precip = 7;
$heading = $pressure = $course27 = 511;
$rot = 128;
$sart_msg = "SART ACTIVE";
$seastate = 13;
$speed = 102.3;
$speed27 = 63;
for ($i=1; $i<=10; $i++) { $subarea [$i] = ""; }
$swellheight = $waveheight = 25.5;
$swellperiod = $waveperiod = 63;
$vis = 12.7;
$notice = $wspeed = $wgust = 127;

$ais_sent = "!AIVDM";
$nmea = "";
$tx = "";
$auto = $batch = 0;

# Parse the command line to get message parameters

GetOptions
  (
  "help" => \$help,
  "type=s" => \$type,
  "mmsi=s" => \$mmsi,

  "accuracy=s" => \$accuracy,
  "aid_name=s" => \$aid_name,
  "aid_status=s" => \$aid_status,
  "aid_type=s" => \$aid_type,
  "airspeed=s" => \$airspeed,
  "airtemp=s" => \$airtemp,
  "altitude=s" => \$altitude,
  "berth_dep=s" => \$berth_dep,
  "berth_len=s" => \$berth_len,
  "callsign=s" => \$callsign,
  "channel_a=s" => \$channel_a,
  "channel_b=s" => \$channel_b,
  "course=s" => \$course,
  "course27=s" => \$course27,
  "dac=s" => \$dac,
  "day=s" => \$day,
  "dest=s" => \$dest,
  "dgnss_msg=s" => \$dgnss_msg,
  "draft=s" => \$draft,
  "dte=s" => \$dte,
  "duration=i" => \$duration,
  "epfd=s" => \$epfd,
  "fid=s" => \$fi, # Keep this switch name for backward compatibility. Accepts both --fi and --fid
  "fatdma=s" => \$fatdma,
  "gnss_pos=s" => \$gnss_pos,
  "heading=s" => \$heading,
  "hour=s" => \$hour,
  "ice=s" => \$ice,
  "imo=s" => \$imo,
  "interval=s" => \$interval,
  "lat=s" => \$lat,
  "lockinfo=s" => \$lockinfo,
  "long=s" => \$long,
  "minute=s" => \$minute,
  "model=s" => \$model,
  "month=s" => \$month,
  "mooring=i" => \$mooring,
  "mothership=s" => \$mothership,
  "msglink=i" => \$msglink,
  "navstat=s" => \$navstat,
  "ne_lat=s" => \$ne_lat,
  "ne_lon=s" => \$ne_lon,
  "notice=i" => \$notice,
  "offpos=s" => \$offpos,
  "part=s" => \$part,
  "pob=s" => \$pob,
  "precip=s" => \$precip,
  "presstend=s" => \$presstend,
  "pressure=s" => \$pressure,
  "quiet=s" => \$quiet,
  "raim=s" => \$raim,
  "rot=s" => \$rot,
  "routetype=s" => \$routetype,
  "sa01=s" => \$subarea [1],
  "sa02=s" => \$subarea [2],
  "sa03=s" => \$subarea [3],
  "sa04=s" => \$subarea [4],
  "sa05=s" => \$subarea [5],
  "sa06=s" => \$subarea [6],
  "sa07=s" => \$subarea [7],
  "sa08=s" => \$subarea [8],
  "sa09=s" => \$subarea [9],
  "sa10=s" => \$subarea [10],
  "sart_msg=s" => \$sart_msg,
  "seastate=s" => \$seastate,
  "second=s" => \$second,
  "serial=s" => \$serial,
  "sndrclass=s" => \$sndrclass,
  "speed=s" => \$speed,
  "speed27=s" => \$speed27,
  "svc=s" => \$svc,
  "sw_lat=s" => \$sw_lat,
  "sw_lon=s" => \$sw_lon,
  "swelldir=s" => \$swelldir,
  "swellheight=s" => \$swellheight,
  "swellperiod=s" => \$swellperiod,
  "target=s" => \$target,
  "tos=s" => \$tos,
  "ts=s" => \$ts,
  "v_aton" => \$v_aton,
  "vendor=s" => \$vendor,
  "vis=s" => \$vis,
  "vname=s" => \$vname,
  "vsize_a=s" => \$vsize_a,
  "vsize_b=s" => \$vsize_b,
  "vsize_c=s" => \$vsize_c,
  "vsize_d=s" => \$vsize_d,
  "vtype=s" => \$vtype,
  "wavedir=s" => \$wavedir,
  "waveheight=s" => \$waveheight,
  "waveperiod=s" => \$waveperiod,
  "wspeed=s" => \$wspeed,
  "wgust=s" => \$wgust,
  "wdir=s" => \$wdir,
  "wgustdir=s" => \$wgustdir,
  "wmtype=s" => \$wmtype,
  "year=s" => \$year,
  "info68=s" => \$info68,

  "ais_sent=s" => \$ais_sent,
  "nmea=s" => \$nmea,
  "tx=s" => \$tx,
  "batch" => \$batch,
  "auto" => \$auto,
  )
  or die "Usage: $0 --help for command line assistance\n";

if ($help)
  {
  help ();
  exit;
  }

# Set the output based upon the --batch switch. If in batch mode, only direct command to print

if ($batch == 0)
    {
    *OUTDEV = STDOUT;
    open (OUTDEV, ">");
    }
  else
    {
    open (OUTDEV, ">", File::Spec->devnull);
    }

print OUTDEV "\n***** AIS_ping (Version: $version, Build date: $build_date) *****\n\n";

# If type and MMSI not provided, display error message and help instructions...

if (not $type or not $mmsi)
  {
  print OUTDEV "\n***** You need to provide a message type and MMSI!! *****\n";
  help ();
  exit;
  }

## Case statement -- encode message based upon type

# Common code. All messages start with the same 38 bits....

################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI

$payload = sprintf ("%06b", int($type)) . "00" . sprintf ("%030b", int($mmsi));

ENCODE:
  {
  $type >= 1 && $type <= 3 && do
    {
    $payload = $payload . encode_010203 ($type,$navstat,$rot,$speed,$accuracy,$long,$lat,$course,$heading,$ts,$raim);
    last ENCODE;
    };
  ($type == 4 || $type == 11) && do
    {
    $payload = $payload . encode_0411 ($year,$month,$day,$hour,$minute,$second,$accuracy,$long,$lat,$epfd,$raim);
    last ENCODE;
    };
  $type == 5 && do
    {
    $payload = $payload . encode_05 ($imo,$callsign,$vname,$vtype,$vsize_a,$vsize_b,$vsize_c,$vsize_d,$epfd,$month,$day,$hour,$minute,$draft,$dest,$dte);
    last ENCODE;
    };

# Type 6 and Type 8 messages need special handling, as there are many -- sometimes nearly overlapping -- formats!
# The $type68 variable "packs" the type, DAC, and FI to eliminate a lot of ANDs and ORs, and to make it a bit easier
# to add options going forward. ($type68 format = TDDDFF)

  ($type == 6 || $type == 8) && do
    {
    $type68 = $type*100000 + $dac*100 + $fi;
    TYPE0608:
      {
      ($type68 == 600000) && do
        {
        # MONITORING AIDS TO NAVIGATION
        $payload = $payload . encode_06_000_00 ($dac,$fi,$target,$info68,$offpos);
        last TYPE0608;
        };
      ($type68 == 600100 || $type68 == 800100) && do
        {
        # TEXT USING 8-BIT ASCII
        $payload = $payload . encode_0608_001_00 ($type,$dac,$fi,$target,$info68);
        last TYPE0608;
        };
      ($type68 == 600102 || $type68 == 600103) && do
        {
        # INTERROGATION FOR A SPECIFIC FUNCTIONAL MESSAGE (02) and CAPABILITY INTERROGATION (03)
        $payload = $payload . encode_06_001_02_03 ($dac,$fi,$target,$info68);
        last TYPE0608;
        };
      $type68 == 600104 && do
        {
        # CAPABILITY REPLY (04)
        $payload = $payload . encode_06_001_04 ($dac,$fi,$target,$info68);
        last TYPE0608;
        };
      ($type68 == 600116 || $type68 == 600140) && do
        {
        # NUMBER OF PERSONS ON BOARD - ADDRESSED (16) and NUMBER OF PERSONS ON BOARD (40)
        $payload = $payload . encode_06_001_16_40 ($dac,$fi,$target,$pob);
        last TYPE0608;
        };
      ($type68 == 800116) && do
        {
        # VTS TARGETS
        $payload = $payload . encode_08_001_16 ($dac,$fi,\@subarea); # Pass array by reference
        last TYPE0608;
        };
      ($type68 == 600118) && do
        {
        # CLEARANCE TIME TO ENTER PORT - ADDRESSED
        $payload = $payload . encode_06_001_18 ($dac,$fi,$target,$msglink,$month,$day,$hour,$minute,$info68,$dest,$long,$lat);
        last TYPE0608;
        };
      ($type68 == 800118) && do
        {
        # ADVICE OF WAYPOINTS/VTS ROUTE PLAN
        $payload = $payload . encode_08_001_18 ($dac,$fi,$info68);
        last TYPE0608;
        };
      ($type68 == 600119) && do
        {
        # EXTENDED SHIP STATIC AND VOYAGE RELATED DATA
        $payload = $payload . encode_06_001_19 ($dac,$fi,$target,$info68);
        last TYPE0608;
        };
      ($type68 == 600120) && do
        {
        # BERTHING DATA - ADDRESSED
        $payload = $payload . encode_06_001_20 ($dac,$fi,$target,$msglink,$berth_len,$berth_dep,$mooring,$month,$day,$hour,$minute,$svc,$tos,
        $info68,$long,$lat);
        last TYPE0608;
        };
      ($type68 == 600123 || $type68 == 800122) && do
        {
        # AREA NOTICE - ADDRESSED (6) and BROADCAST (8)
        $payload = $payload . encode_0608_001_23_22 ($type,$dac,$fi,$target,$msglink,$notice,$month,$day,$hour,$minute,$duration,\@subarea); # Pass array by reference
        last TYPE0608;
        };
      ($type68 == 600128 || $type68 == 800127) && do
        {
        # ROUTE INFORMATION - ADDRESSED (06) and BROADCAST (08)
        $payload = $payload . encode_0608_001_28_27 ($type,$dac,$fi,$target,$msglink,$sndrclass,$routetype,$month,$day,$hour,$minute,$duration,$info68);
        last TYPE0608;
        };
      ($type68 == 600130 || $type68 == 800129) && do
        {
        # TEXT DESCRIPTION - ADDRESSED (6) and BROADCAST (8)
        $payload = $payload . encode_0608_001_29_30__367_29 ($type,$dac,$fi,$target,$msglink,$info68);
        last TYPE0608;
        };
      ($type68 == 800131) && do
        {
        # METEOROLOGICAL AND HYDROGRAPHIC DATA
        $payload = $payload . encode_08_001_31 ($dac,$fi,$long,$lat,$accuracy,$day,$hour,$minute,$wspeed,$wgust,$wdir,$wgustdir,$pressure,$presstend,$vis,
        $waveheight,$waveperiod,$wavedir,$swellheight,$swellperiod,$swelldir,$seastate,$precip,$ice);
        last TYPE0608;
        };
      ($type68 == 623510 || $type68 == 625010) && do
        {
        # AIDS TO NAVIGATION MONITORING DATA
        $payload = $payload . encode_06_235_250_10 ($dac,$fi,$target,$info68,$offpos);
        last TYPE0608;
        };
      ($type68 == 623520) && do
        {
        # BUOY POSITION MONITORING
        $payload = $payload . encode_06_235_20 ($dac,$fi,$target,$aid_name,$day,$hour,$minute,$long,$lat,$offpos);
        last TYPE0608;
        };
      ($type68 == 636716) && do
        {
        # PASSENGER AND CREW COUNT - ADDRESSED
        $payload = $payload . encode_06_367_16 ($dac,$fi,$target,$msglink,$day,$hour,$minute,$pob,$info68);
        last TYPE0608;
        };
      ($type68 == 836620) && do
        {
        # DAM STATUS - BROADAST
        $payload = $payload . encode_08_366_20 ($dac,$fi,$msglink,$hour,$minute,$lockinfo,$info68);
        last TYPE0608;
        };
      ($type68 == 636722 || $type68 == 836722) && do
        {
        # GEOGRAPHIC NOTICE - ADDRESSED (6) and BROADCAST (8)
        $payload = $payload . encode_0608_367_22 ($type,$dac,$fi,$target,$msglink,$notice,$month,$day,$hour,$minute,$duration,\@subarea); # Pass array by reference
        last TYPE0608;
        };
      ($type68 == 836723) && do
        {
        # SATELLITE SHIP WEATHER - BROADCAST
        $payload = $payload . encode_08_367_23 ($dac,$fi,$day,$hour,$minute,$long,$lat,$pressure,$airtemp,$wspeed,$wgust,$wdir);
        last TYPE0608;
        };
      ($type68 == 836724) && do
        {
        # SATELLITE SHIP WEATHER SMALL - BROADCAST
        $payload = $payload . encode_08_367_24 ($dac,$fi,$hour,$minute,$long,$lat,$pressure);
        last TYPE0608;
        };
      ($type68 == 836725) && do
        {
        # SATELLITE SHIP WEATHER SMALL - BROADCAST
        $payload = $payload . encode_08_367_25 ($dac,$fi,$hour,$minute,$pressure,$wspeed,$wdir);
        last TYPE0608;
        };
      ($type68 == 836726) && do
        {
        # GLS JAMMING AND SPOOFING REPORT
        $payload = $payload . encode_08_367_26 ($dac,$fi,$day,$hour,$minute,$long,$lat,$info68);
        last TYPE0608;
        };
      ($type68 == 636729 || $type68 == 836729) && do
        {
        # LINKED TEXT - ADDRESSED (6) and BROADCAST (8)
        $payload = $payload . encode_0608_001_29_30__367_29 ($type,$dac,$fi,$target,$msglink,$info68);
        last TYPE0608;
        };
      ($type68 == 836733) && do
        {
        # ENVIRONMENTAL MANAGMEMNT
        $payload = $payload . encode_08_367_33 ($type,$dac,$fi,\@subarea); # Pass array by reference
        last TYPE0608;
        };
      ($type68 == 636735 || $type68 == 836735) && do
        {
        # WATERWAYS MANAGMEMNT - ADDRESSED (6) and BROADCAST (8)
        $payload = $payload . encode_0608_367_35 ($type,$dac,$fi,$target,$msglink,$wmtype,$info68,\@subarea); # Pass array by reference
        last TYPE0608;
        };
      }
    last ENCODE;
    };
  ($type == 7 || $type == 13) && do
    {
    $payload = $payload . encode_0713 ($target);
    last ENCODE;
    };
  $type == 9 && do
    {
    $payload = $payload . encode_09 ($altitude,$airspeed,$accuracy,$long,$lat,$course,$ts,$dte,$raim);
    last ENCODE;
    };
  $type == 10 && do
    {
    $payload = $payload . encode_10 ($target);
    last ENCODE;
    };
  ($type == 12 || $type == 14) && do
    {
    $payload = $payload . encode_1214 ($type,$target,$sart_msg);
    last ENCODE;
    };
  $type == 15 && do
    {
    $payload = $payload . encode_15 ($target);
    last ENCODE;
    };
  $type == 16 && do
    {
    $payload = $payload . encode_16 ($target);
    last ENCODE;
    };
  $type == 17 && do
    {
    $payload = $payload . encode_17 ($long,$lat,$dgnss_msg);
    last ENCODE;
    };
  $type == 18 && do
    {
    $payload = $payload . encode_18 ($speed,$accuracy,$long,$lat,$course,$heading,$ts,$raim);
    last ENCODE;
    };
  $type == 19 && do
    {
    $payload = $payload . encode_19 ($speed,$accuracy,$long,$lat,$course,$heading,$ts,$vname,$vtype,$vsize_a,$vsize_b,$vsize_c,$vsize_d,$epfd,$raim,$dte);
    last ENCODE;
    };
  $type == 20 && do
    {
    $payload = $payload . encode_20 ($fatdma);
    last ENCODE;
    };
  $type == 21 && do
    {
    $payload = $payload . encode_21 ($aid_type,$aid_name,$accuracy,$long,$lat,$vsize_a,$vsize_b,$vsize_c,$vsize_d,$epfd,$ts,$offpos,$aid_status,$raim,$v_aton);
    last ENCODE;
    };
  $type == 22 && do
    {
    $payload = $payload . encode_22 ($channel_a,$channel_b,$ne_lon,$ne_lat,$sw_lon,$sw_lat);
    last ENCODE;
    };
  $type == 23 && do
    {
    $payload = $payload . encode_23 ($ne_lon,$ne_lat,$sw_lon,$sw_lat,$vtype,$interval,$quiet);
    last ENCODE;
    };
  $type == 24 && do
    {
    $payload = $payload . encode_24 ($mmsi,$part,$vname,$vtype,$vendor,$model,$serial,$callsign,$vsize_a,$vsize_b,$vsize_c,$vsize_d,$mothership);
    last ENCODE;
    };
  ($type == 25 || $type == 26) && do
    {
    $payload = $payload . encode_2526 ($type,$target,$dac,$fi,$info68);
    last ENCODE;
    };
  $type == 27 && do
    {
    $payload = $payload . encode_27 ($accuracy,$raim,$navstat,$long,$lat,$speed27,$course27,$gnss_pos);
    last ENCODE;
    };

# If we get here, we have an invalid message type...

  do
    {
    print OUTDEV "\n***** Unsupported message type: $type *****\n";
    help ();
    exit;
    };
  }

# According to ITU-R Rec. M-1371, messages must be byte-aligned. If not, pad the message.

$payload_len = length ($payload);
$pad = 8 - ($payload_len % 8);
if ($pad != 8)
  {
  print OUTDEV "Payload ($payload_len bits) is not byte-aligned; $pad padding bits added\n\n";
  $payload .= "0" x $pad;
  }

printf OUTDEV "Payload length = %d bits (%d 8-bit bytes)\n\n", length ($payload), length ($payload) / 8;

# Now, display the correct command based upon the --nmea/--tx switches.
# It is possible for the user to supply both switches, in which case we will display two commands.
# Both commands need to direct action to the same AIS channel.
# Default is NMEA sentences for channel A; if both NMEA and TX switches are used, use the NMEA channel.

# If AIS_ping is called in batch mode, assume that the NMEA program should also use batch mode...
# Assume that we will create a !AIVDM sentence unless otherwise told

if ($nmea ne "")
    { $ais_channel = $nmea; }
  else
    { $ais_channel = $tx; }

if ($ais_channel ne "A" && $ais_channel ne "B")
  { $ais_channel = "A"; }

print OUTDEV "Use the following command(s) next:\n\n";
if ($nmea ne "")
  {
  $nmea_command = "perl AIS_NMEA.pl";

  if ($ais_sent ne "!AIVDM")
    { $nmea_command .= " --ais_sent='" . $ais_sent . "'"; }
  $nmea_command .= " --payload=" . $payload . " --channel=". $ais_channel;
  if ($batch)
    { $nmea_command .= " --batch"; }
  print "$nmea_command\n";
  print OUTDEV "\n";
  if ($auto)
    {
    open (TMPFILE, ">", "ais_tmp_file2.sh");
    print TMPFILE "$nmea_command\n";
    close TMPFILE;
    }
  }
if ($tx ne "")
  { print OUTDEV "./AiS_TX.py --payload=$payload --channel=$ais_channel\n\n"; }

close OUTDEV;

## END OF MAIN()


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# encode_string
# ********************************************************

# Accept an ASCII character string and convert to six-bit ASCII character set.

sub encode_string
  {
# Parameter list
  my ($message) = @_;
  my ($i,$n,$len);
  my $vocabulary = '@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_ !"#$%&\'()*+,-./0123456789:;<=>?';
  my $string = "";

  $len = length ($message);

  for ($i=0; $i<$len; $i++)
    {
    $n = index ($vocabulary, uc (substr($message,$i,1)));
    if ($n >= 0)
      { $string .= sprintf ("%06b", $n); }
    }
  return $string;
  }

# ********************************************************
# pad_string
# ********************************************************

# Accept an ASCII character string and pad it to PAD_LEN using "@" symbols

sub pad_string
  {
# Parameter list
  my ($string,$pad_len) = @_;

  if (length($string) < $pad_len)
    { $string = $string . "@" x ($pad_len - length($string)); }

  return $string;
  }

# ********************************************************
# encode -- AIS MESSAGE ENCODING ROUTINES
# ********************************************************

# The encode_X routines encode Type X messages, where X is the type value.

# *** Encode_010203 builds Types 1-3, Position Report Class A messages ***
# Expected to be 168 bits, occupying one AIS sentence.

sub encode_010203
  {
# Parameter list
  my ($type,$navstat,$rot,$speed,$accuracy,$long,$lat,$course,$heading,$ts,$raim) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-41      4   Navigation Status
#  42-49      8   Rate of Turn (ROT)
#  50-59     10   Speed Over Ground (SOG)
#  60         1   Position Accuracy
#  61-88     28   Longitude
#  89-115    27   Latitude
#  116-127   12   Course Over Ground (COG)
#  128-136    9   True Heading (HDG)
#  137-142    6   Time Stamp
#  143-144    2   Maneuver Indicator            ** Not currently supported; set to 0
#  145-147    3   Spare
#  148        1   RAIM flag
#  149-167   19   Radio Status                  ** Not currently supported; set to 0

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  $encode = sprintf ("%04b", int($navstat)) . sprintf ("%08b", int($rot) & (2**8-1)) . sprintf ("%010b", int($speed*10+0.5)) . sprintf ("%01b", int($accuracy))
  . sprintf ("%028b", (int($long*600000) & (2**28-1))) . sprintf ("%027b", (int($lat*600000) & (2**27-1))) . sprintf ("%012b", int($course*10+0.5))
  . sprintf ("%09b", int($heading)) . sprintf ("%06b", int($ts)) . "0"x5 . sprintf ("%01b", int($raim)) . "0"x19 ;
  return $encode;
  }

# *** Encode_0411 builds a Type 4, Base Station Report or Type 11, UTC/Date Response message ***
# Expected to be 168 bits, occupying one AIS sentence.

sub encode_0411
  {
# Parameter list
  my ($year,$month,$day,$hour,$minute,$second,$accuracy,$long,$lat,$epfd,$raim) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-51     14   UTC year
#  52-55      4   UTC month
#  56-60      5   UTC day
#  61-65      5   UTC hour
#  66-71      6   UTC minute
#  72-77      6   UTC second
#  78         1   Position Accuracy
#  79-106    28   Longitude
#  107-133   27   Latitude
#  134-137    4   Type of EPFD
#  138-147   10   Spare
#  148        1   RAIM flag
#  149-167   19   Radio Status                  ** Not currently supported; set to 0

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  $encode = sprintf ("%014b", $year) . sprintf ("%04b", $month) . sprintf ("%05b", $day) . sprintf ("%05b", $hour) . sprintf ("%06b", $minute)
  . sprintf ("%06b", $second) . sprintf ("%01b", int($accuracy)) . sprintf ("%028b", (int($long*600000) & (2**28-1)))
  . sprintf ("%027b", (int($lat*600000) & (2**27-1))) . sprintf ("%04b", int ($epfd)) . "0"x10 . sprintf ("%01b", int($raim)) . "0"x19;
  return $encode;
  }

# *** Encode_05 builds a Type 5, Static and Voyage Related Data message ***
# Expected to be 424 bits, occupying two AIS sentences.

sub encode_05
  {
# Parameter list
  my ($imo, $callsign, $vname, $vtype, $vsize_a, $vsize_b, $vsize_c, $vsize_d, $epfd, $month, $day, $hour, $minute, $draft, $dest, $dte) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   AIS Version                   ** Not currently supported; set to 2
#  40-69     30   IMO Number
#  70-111    42   Call Sign
# 112-231   120   Vessel Name
# 232-239     8   Ship Type
# 240-248     9   Dimension to Bow
# 249-257     9   Dimension to Stern
# 258-263     6   Dimension to Port
# 264-269     6   Dimension to Starboard
# 270-273     4   EPFD Type
# 274-277     4   ETA - UTC month
# 278-282     5   ETA - UTC day
# 283-287     5   ETA - UTC hour
# 288-293     6   ETA - UTC minute
# 294-301     8   Draught
# 302-421   120   Destination
# 422         1   DTE
# 423         1   Spare

  $encode = "10" . sprintf ("%030b", int($imo)) . encode_string (pad_string ($callsign,7)) . encode_string (pad_string ($vname, 20))
  . sprintf ("%08b", int ($vtype)) . sprintf ("%09b", int ($vsize_a)) . sprintf ("%09b", int ($vsize_b)) . sprintf ("%06b", int ($vsize_c))
  . sprintf ("%06b", int ($vsize_d)) . sprintf ("%04b", int ($epfd)) . sprintf ("%04b", int ($month)) . sprintf ("%05b", int ($day))
  . sprintf ("%05b", int ($hour)) . sprintf ("%06b", int ($minute)) . sprintf ("%08b", int($draft*10+0.5)) . encode_string (pad_string ($dest, 20))
  . sprintf ("%1b", int ($dte)) . "0";
  return $encode;
  }

# *** The encode_06, encode_0608, and encode_08 functions build Type 6, Binary Addressed and Type 8, Binary Broadcast Messages ***
# *** See https://www.iala-aism.org/asm/ for the best list of known Types 6 and 8 message formats                              ***

# 06_000_00 -- Build Type 6, DAC 0, FI 00 (Monitoring aids to navigation) [Zeni Lite Buoy Co., Ltd.]
# This message is 136 bits in length.

sub encode_06_000_00
  {
# Parameter list
  my ($dac,$fi,$target,$info68,$offpos) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description                    FIELDS
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-103    16   Sub-application ID (set to 1)     0
# 104-115    12   Voltage data                      1
# 116-125    10   Current data                      2
# 126         1   Power supply type                 3
# 127         1   Light status                      4
# 128         1   Battery status                    5
# 129         1   Off-position status
# 130-135     6   Spare (set to 0)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi);
  
# $info68 is a comma-delimited string containing the next six message parameters

  @fields = split (",", $info68);
  $encode .= sprintf ("%016b", $fields [0]) . sprintf ("%012b", int($fields [1]*10+0.5)) . sprintf ("%010b", int($fields [2]*10+0.5)) . sprintf ("%1b", $fields [3])
  . sprintf ("%1b", $fields [4]) . sprintf ("%1b", $fields [5]) . sprintf ("%1b", $offpos) . "000000";
  return $encode;
  }

# *** Encode_0608_001_00 -- Build Types 6 and 8, DAC = 1, FI = 0 (Text using 6-bit ASCII - Addressed/Broadcast) [ITU Rec. M.1371] messages
# This message can have a variable number of text characters, up to 151 (6) or 156 (8) characters in length

sub encode_0608_001_00
  {
# Parameter list
  my ($type,$dac,$fi,$target,$info68) = @_;
  my ($encode,$fields);

################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  38-39      2   38-39      2   Sequence Number (6)/Spare (8)
#  40-69     30                  Destination MMSI (6)
#  70         1                  Retransmit flag (6)   ** Not currently supported; set to 0
#  71         1                  Spare (6)
#  72-81     10   40-49     10   DAC
#  82-87      6   50-55      6   FI
#  88         1    56         1   Acknowledgement Required flag   ** Not currently supported; set to 0
#  89-99     11    57-67     11   Text Sequence Number
# 100-1005  906    68-1003  936   Text String (<=151 [6]/<=156 [8] six-bit characters)
#1006-1008    3  1004-1008    5   Spare (padding for octet-alignment, as needed)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",",$target);
  if ($type == 6) { $encode =  sprintf ("%02b",$fields[1]) . sprintf ("%030b",$fields[0]) . "00"; }
    else { $encode = "00"; }

# $info68 is a comma-delimited string containing "text_sequence_number,text_message"

  @fields = split (",",$info68);
  $encode .= sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . "0" . sprintf ("%011b", $fields [0]) . encode_string ($fields[1]);
  return $encode;
  }

# encode_06_001_02_03 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 2 (Interrogation for a specific functional message) [ITU M.1371]
# or FI = 3 (Capability interrogation) [ITU M.1371].
# The messages are 168 bits in length and have similar formats

sub encode_06_001_02_03
  {
# Parameter list
  my ($dac,$fi,$target,$info68) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-97     10   Requested DAC code
#  98-103     6   Requested FI code          [FI = 2]
# 104-167    64   Spare (set to 0)           [FI = 2]

#  98-167    70   Spare (set to 0)           [FI = 3]

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi)
  . sprintf ("%010b", int($info68/100));
  if ($fi == 2)
      { $encode .= sprintf ("%06b", int ($info68) % 100) . "0"x64; }
    else
      { $encode .= "0"x70; }

  return $encode;
  }

# encode_06_001_04 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 4 (Capability reply) [ITU M.1371].
# This message is 352 bits in length 

sub encode_06_001_04
  {
# Parameter list
  my ($dac,$fi,$target,$info68) = @_;
  my ($encode,$fields,$fi_list_len,$i,$x);
  my $fi_table = "00"x64;        # Initial FI capability table vector

################################################
#  Bits     Len   Description                  FIELDS
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-97     10   DAC code being interrogated     0
#  98-225   128   FI capability table             1
# 226-351   126   Spare (set to 0)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi);

# $info68 is a comma-delimited string containing the requested DAC ($fields[0]) and then the list of supports FIs [1]..[n]
# To prevent bad things from happening, limit the FI list length to 65 (64 FIs + DAC).
# Also, test that each entry in the list is an actual integer between 0-63.

  @fields = split (",",$info68);
  $encode .= sprintf ("%010b", $fields [0]);

  $fi_list_len = scalar @fields;
  if ($fi_list_len > 65) { $fi_list_len = 65; }
  for ($i = 1; $i < $fi_list_len; $i++)
    {
    $x = $fields [$i];
    if (looks_like_number ($x) && int($x) == $x && $x >= 0 && $x <=63) { substr ($fi_table,int($x)*2,1,"1"); }
    }
  for ($i = 0; $i < 128; $i++) { $encode .= sprintf ("%1b", substr ($fi_table,$i,1)); }
  $encode .= "0"x126;
      
  return $encode;
  }

# encode_06_001_16_40 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 16 (Number of Persons On Board - Addressed) [IMO 289]
# or FI = 16 [Number of Person on Board) [ITU M.1371].
# The DAC=16 message is 136 bits in length and the DAC=40 message is 104 bits in length.

sub encode_06_001_16_40
  {
# Parameter list
  my ($dac,$fi,$target,$pob) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-100    13   Number of persons
# 101-135    35   Spare (set to 0)          [DAC = 16]
# 101-103     3   Spare (set to 0)          [DAC = 40]

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi)
  . sprintf ("%013b", $pob);
  if ($dac == 16)
      { $encode .= "0"x35; }
    else
      { $encode .= "000"; }

  return $encode;
  }

# encode_08_001_16 -- Build Type 8, DAC 1, FI 16 (VTS Targets) message
# This message is variable length and can be 176-896 bits in length.

sub encode_08_001_16
  {
# Parameter list (NOTE: vts_targets array passed by reference)
  my ($dac,$fi,$vts_targets_ref) = @_;
  my @vts_targets = @{$vts_targets_ref};  # Dereference array
  my ($encode,$fields,$i,$id);

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-175   120   VTS Target 1
# 176-295   120   VTS Target 2 (optional)
# 296-415   120   VTS Target 3 (optional)
# 416-535   120   VTS Target 4 (optional)
# 536-655   120   VTS Target 5 (optional)
# 656-775   120   VTS Target 6 (optional)
# 776-895   120   VTS Target 7 (optional)

  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi);

# This message contains 1-7 VTS target reports; the $vts_targets array is a comma-delimited string containing the seven values

# Target Report
#-----------------------------------------------
#  Bits     Len   Description             FIELDS
#   0-1       2   Identifier Type            0
#   2-43     42   Target ID                  1
#  44-47      4   Spare (set to 0)
#  48-71     24   Latitude                   2
#  72-96     25   Longitude                  3
#  97-105     9   Course over ground         4
# 106-111     6   Timestamp                  5
# 112-119     8   Speed over ground          6

  for ($i=1; $i<=7; $i++)
    {
    if ($subarea [$i] eq "") { last; }

    @fields = split (",", $subarea [$i]);

    $id = $fields [0];
    $encode .= sprintf ("%02b", $id);
    if ($id == 2)
        { $encode .= encode_string (pad_string ($fields[1],7)); }
      else
        { $encode .= sprintf ("%042b", $fields [1]); }

    $encode .= "0000" . sprintf ("%024b", (int($fields[2]*60000) & (2**24-1))) . sprintf ("%025b", (int($fields[3]*60000) & (2**25-1)))
    . sprintf ("%09b", int($fields[4])) . sprintf ("%06b", int($fields[5])) . sprintf ("%08b", int($fields[6]));
    };

  return $encode;
  }

# encode_06_001_18 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 18 (Clearance Time to Enter Port - Addressed) [IMO 289].
# This message is 360 bits in length.

sub encode_06_001_18
  {
# Parameter list
  my ($dac,$fi,$target,$msglink,$month,$day,$hour,$minute,$info68,$dest,$long,$lat) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-97     10   Message Linkage ID
#  98-101     4   Clearance time to enter port -- UTC month
# 102-106     5   Clearance time to enter port -- UTC day
# 107-111     5   Clearance time to enter port -- UTC hour
# 112-117     6   Clearance time to enter port -- UTC minute
# 118-237   120   Name of Port and Berth
# 238-267    30   Destination
# 268-292    25   Longitude
# 293-316    24   Latitude
# 317-359    43   Spare (set to 0)

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi)
  . sprintf ("%010b", $msglink) . sprintf ("%04b", int ($month)) . sprintf ("%05b", int ($day)) . sprintf ("%05b", int ($hour))
  . sprintf ("%06b", int ($minute)) . encode_string (pad_string ($info68,20)) . encode_string (pad_string ($dest,5))
  . sprintf ("%025b", (int($long*60000) & (2**25-1))) . sprintf ("%024b", (int($lat*60000) & (2**24-1))) . "0"x43;

  return $encode;
  }

# encode_08_001_18 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 18 (Advice of Waypoints/VTS Route Plan) [ITU-R.M.1371-1].
# This message is up to 984 bits in length.

sub encode_08_001_18
  {
# Parameter list
  my ($dac,$fi,$info68) = @_;
  my ($encode,$fields,$i,$n);
  my @spare = (4,1,6,3,0,5,2,7,4,1,6,3,0);

################################################
#  Bits     Len   Description                        FIELDS
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-59      4   Number of waypoints                  0 [N]
#  60-126    67   Waypoint 1 detail (optional)
# 127-193    67   Waypoint 2 detail (optional)
# 194-260    67   Waypoint 3 detail (optional)
# 261-327    67   Waypoint 4 detail (optional)
# 328-394    67   Waypoint 5 detail (optional)
# 395-361    67   Waypoint 6 detail (optional)
# 462-528    67   Waypoint 7 detail (optional)
# 529-595    67   Waypoint 8 detail (optional)
# 596-662    67   Waypoint 9 detail (optional)
# 663-729    67   Waypoint 10 detail (optional)
# 730-796    67   Waypoint 11 detail (optional)
# 797-863    67   Waypoint 12 detail (optional)
# 864-983   120   Route description                   N*3 + 1

# Waypoint detail
#-----------------------------------------------
#  Bits     Len   Description
#   0-27     28   Longitude                           (I-1)*3 + 1, I = 1-N
#  28-54     27   Latitude                            (I-1)*3 + 2, I = 1-N
#  55-66     12   Turning radius                      (I-1)*3 + 3, I = 1-N <-- Yes, (I-1)*3+3 = I*3, but that interrupts the pattern...

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

# $info68 is a comma-delimited string containing waypoint information

  @fields = split (",", $info68);
  $n = int($fields [0]);
  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . sprintf ("%04b", $n);

  for ($i=1; $i<=$n; $i++)
    {
    $encode .= sprintf ("%028b", (int($fields[($i-1)*3+1]*600000) & (2**28-1))) . sprintf ("%027b", (int($fields[($i-1)*3+2]*600000) & (2**27-1)))
    . sprintf ("%012b", $fields [($i-1)*3+3]);
    }

  $encode .= encode_string (pad_string ($fields[$n*3+1],20)) . "0"x$spare[$n];

  return $encode;
  }

# encode_06_001_19 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 19 (Extended ship static and voyage related data) [ITU M.1371].
# This message is 104 bits in length.

sub encode_06_001_19
  {
# Parameter list
  my ($dac,$fi,$target,$info68) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-98     11   Height over keel
#  99-103     5   Spare (set to 0)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi)
  . sprintf ("%011b", $info68) . "0"x5;

  return $encode;
  }

# encode_06_001_20 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 20 [Berthing Data - Addressed) [IMO 289].
# This message is 360 bits in length.

sub encode_06_001_20
  {
# Parameter list
  my ($dac,$fi,$target,$msglink,$berth_len,$berth_dep,$mooring,$month,$day,$hour,$minute,$svc,$tos,$info68,$long,$lat) = @_;
  my ($encode,$fields,$i);

################################################
#  Bits     Len   Description
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-97     10   Message Linkage ID
#  98-106     9   Berth Length
# 107-114     8   Water depth at berth
# 115-117     3   Mooring Position
# 118-121     4   UTC month
# 122-126     5   UTC day
# 127-131     5   UTC hou
# 132-137     6   UTC minute
# 138         1   Services Availability
# 139-190    52   Type of Services Available
# 191-310   120   Name of Berth
# 311-335    25   Centre of Berth, Longitude
# 336-359    24   Centre of Berth, Latitude

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi)
  . sprintf ("%010b", $msglink) . sprintf ("%09b", int ($berth_len)) . sprintf ("%08b", int($berth_dep*10+0.5)) . sprintf ("%03b", int ($mooring))
  . sprintf ("%04b", int ($month)) . sprintf ("%05b", int ($day)) . sprintf ("%05b", int ($hour)) . sprintf ("%06b", int ($minute))
  . sprintf ("%01b", int ($svc));

  for ($i=0; $i<26; $i++)
    { $encode .= sprintf ("%02b", substr ($tos,$i,1)); }
 
  $encode .= encode_string (pad_string ($info68,20)) . sprintf ("%025b", (int($long*60000) & (2**25-1))) . sprintf ("%024b", (int($lat*60000) & (2**24-1)));

  return $encode;
  }

# encode_0608_001_23_22 -- Build Type 6, DAC 1, FI 23 (Area Notice - Addressed) and Type 8, DAC 1, FI 22 (Area Notice - Broadcast) [IMO Circular 289]
# This message is variable length and can be up to 1013 (6) or 981 (8) bits in length.

sub encode_0608_001_23_22
  {
# Parameter list (NOTE: subarea array passed by reference)
  my ($type,$dac,$fi,$target,$msglink,$notice,$month,$day,$hour,$minute,$duration,$subarea_ref) = @_;
  my @subarea = @{$subarea_ref};  # Dereference array
  my ($encode,$fields,$i);

################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  38-39      2    38-39      2   Sequence Number (6)/Spare (8)
#  40-69     30                   Destination MMSI
#  70         1                   Retransmit flag             ** Not currently supported; set to 0
#  71         1                   Spare
#  72-81     10    40-49     10   DAC
#  82-87      6    50-55      6   FI
#  88-97     10    56-65     10   Message Linkage ID
#  98-104     7    66-72      7   Notice Description
# 105-108     4    73-76      4   UTC month
# 109-113     5    77-81      5   UTC day
# 114-118     5    82-86      5   UTC hour
# 119-124     6    87-92      6   UTC minute
# 125-142    18    93-110    18   Duration
# 143-229    87   111-197    87   Sub-area 1
# 230-316    87   198-284    87   Sub-area 2 (if present)
# 317-403    87   285-371    87   Sub-area 3 (if present)
# 404-490    87   372-458    87   Sub-area 4 (if present)
# 491-577    87   459-545    87   Sub-area 5 (if present)
# 578-664    87   546-632    87   Sub-area 6 (if present)
# 665-751    87   633-719    87   Sub-area 7 (if present)
# 752-838    87   720-806    87   Sub-area 8 (if present)
# 839-925    87   807-893    87   Sub-area 9 (if present)
# 926-1012   87   894-980    87   Sub-area 10 (if present)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  if ($type == 6)
      {
      @fields = split (",", $target);
      $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00";
      }
    else
     { $encode = "00"; }

  $encode .= sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . sprintf ("%010b", $msglink) . sprintf ("%07b", $notice)
  . sprintf ("%04b", $month) . sprintf ("%05b", $day) . sprintf ("%05b", $hour) . sprintf ("%06b", $minute) . sprintf ("%018b", $duration);

# This message contains 1-10 subareas; the 10-element $subarea array contains one comma-delimited string per area, where each string
# lists the pertinent subarea parameters

  for ($i=1; $i<=10; $i++)
    {
    if ($subarea [$i] eq "")
      { last; }

    @fields = split (",", $subarea [$i]);

    SHAPE:
      {

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (0)      0
#   3-4       2   Scale Factor        1
#   5-29     25   Longitude           2
#  30-53     24   Latitude            3
#  54-56      3   Precision           4
#  57-68     12   Radius              5
#  69-86     18   Not used

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

      $fields [0] == 0 && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%025b", (int($fields[2]*60000) & (2**25-1)))
       . sprintf ("%024b", (int($fields[3]*60000) & (2**24-1))) . sprintf ("%03b", int($fields[4])) . sprintf ("%012b", int($fields[5])) . "0"x18;
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (1)      0
#   3-4       2   Scale Factor        1
#   5-29     25   Longitude           2
#  30-53     24   Latitude            3
#  54-56      3   Precision           4
#  57-64      8   E Dimension         5
#  65-72      8   N Dimension         6
#  73-81      9   Orientation         7
#  82-86      5   Not used

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

      $fields [0] == 1 && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%025b", (int($fields[2]*60000) & (2**25-1)))
       . sprintf ("%024b", (int($fields[3]*60000) & (2**24-1))) . sprintf ("%03b", int($fields[4])) . sprintf ("%08b", int($fields[5]))
       . sprintf ("%08b", int($fields[6])) . sprintf ("%09b", int($fields[7])) . "0"x5;
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (2)      0
#   3-4       2   Scale Factor        1
#   5-29     25   Longitude           2
#  30-53     24   Latitude            3
#  54-56      3   Precision           4
#  57-68     12   Radius              5
#  69-77      9   Left Boundary       6
#  78-86      9   Right Boundary      7

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

      $fields [0] == 2 && do
        {

        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%025b", (int($fields[2]*60000) & (2**25-1)))
       . sprintf ("%024b", (int($fields[3]*60000) & (2**24-1))) . sprintf ("%03b", int($fields[4])) . sprintf ("%012b", int($fields[5]))
       . sprintf ("%09b", int($fields[6])) . sprintf ("%09b", int($fields[7]));
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (3,4)    0
#   3-4       2   Scale Factor        1
#   5-14     10   Point 1 Angle       2
#  15-24     10   Point 1 Distance    3
#  25-34     10   Point 2 Angle       4
#  35-44     10   Point 2 Distance    5
#  45-54     10   Point 3 Angle       6
#  55-64     10   Point 3 Distance    7
#  65-74     10   Point 4 Angle       8
#  75-84     10   Point 4 Distance    9
#  85-86      2   Not used

      ($fields [0] == 3 || $fields [0] == 4) && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%010b", int($fields[2])) . sprintf ("%010b", int($fields[3]))
        . sprintf ("%010b", int($fields[4])) . sprintf ("%010b", int($fields[5])) . sprintf ("%010b", int($fields[6])) . sprintf ("%010b", int($fields[7]))
        . sprintf ("%010b", int($fields[8])) . sprintf ("%010b", int($fields[9])) . "00";
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (5)      0
#   3-86     84   Text                1   (14 6-bit characters)

      $fields [0] == 5 && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . encode_string (pad_string ($fields [1], 14));
        last SHAPE;
        };
      };
    };

  return $encode;
  }

# 0608_001_28_27 -- Build Type 6, DAC 1, FI 28 (Route Information - Addressed) and Type 8, DAC 1, FI 27 (Route Information - Broadcast) [IMO Circular 289]
# This message is variable length and can be up to 1030 (6) or 998 (8) bits in length.

sub encode_0608_001_28_27
  {
# Parameter list
  my ($type,$dac,$fi,$target,$msglink,$sndrclass,$routetype,$month,$day,$hour,$minute,$duration,$info68) = @_;
  my ($encode,$fields,$i,$numpoints);

################################################
#   == Type 6 ==    == Type 8 ==
#   Bits     Len    Bits     Len   Description
#   38-39      2    38-39      2   Sequence Number (6)/Spare (8)
#   40-69     30                   Destination MMSI
#   70         1                   Retransmit flag             ** Not currently supported; set to 0
#   71         1                   Spare
#   72-81     10    40-49     10   DAC
#   82-87      6    50-55      6   FI
#   88-97     10    56-65     10   Message Linkage ID
#   98-100     3    66-68      3   Sender Classification
#  101-105     5    69-73      5   Route Type
#  106-109     4    74-77      4   UTC month
#  110-114     5    78-82      5   UTC day
#  115-119     5    83-87      5   UTC hour
#  120-125     6    88-93      6   UTC minute
#  126-143    18    94-111    18   Duration
#  144-148     5   112-116     5   Number of waypoints
#  149-176    28   117-144    28   Longitude 1 (if present)
#  177-203    27   145-171    27   Latitude 1 (if present)
#  204-231    28   172-199    28   Longitude 2 (if present)
#  232-258    27   200-226    27   Latitude 2 (if present)
#  259-286    28   227-254    28   Longitude 3 (if present)
#  287-313    27   255-281    27   Latitude 3 (if present)
#  314-341    28   282-309    28   Longitude 4 (if present)
#  342-368    27   310-336    27   Latitude 4 (if present)
#  369-396    28   337-364    28   Longitude 5 (if present)
#  397-423    27   365-391    27   Latitude 5 (if present)
#  424-451    28   392-419    28   Longitude 6 (if present)
#  452-478    27   420-446    27   Latitude 6 (if present)
#  479-506    28   447-474    28   Longitude 7 (if present)
#  507-533    27   475-501    27   Latitude 7 (if present)
#  534-561    28   502-529    28   Longitude 8 (if present)
#  562-588    27   530-556    27   Latitude 8 (if present)
#  589-616    28   557-584    28   Longitude 9 (if present)
#  617-643    27   585-611    27   Latitude 9 (if present)
#  644-671    28   612-639    28   Longitude 10 (if present)
#  672-698    27   640-666    27   Latitude 10 (if present)
#  699-726    28   667-694    28   Longitude 11 (if present)
#  727-753    27   695-721    27   Latitude 11 (if present)
#  754-781    28   722-749    28   Longitude 12 (if present)
#  782-808    27   750-776    27   Latitude 12 (if present)
#  809-836    28   777-804    28   Longitude 13 (if present)
#  837-863    27   805-831    27   Latitude 13 (if present)
#  864-891    28   832-859    28   Longitude 14 (if present)
#  892-918    27   860-886    27   Latitude 14 (if present)
#  929-946    28   887-914    28   Longitude 15 (if present)
#  947-973    27   915-941    27   Latitude 15 (if present)
#  974-1001   28   942-969    28   Longitude 16 (if present)
# 1002-1028   27   970-996    27   Latitude 16 (if present)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  if ($type == 6)
      {
      @fields = split (",", $target);
      $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00";
      }
    else
     { $encode = "00"; }

  $encode .= sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . sprintf ("%010b", $msglink) . sprintf ("%03b", $sndrclass) . sprintf ("%05b", $routetype)
  . sprintf ("%04b", $month) . sprintf ("%05b", $day) . sprintf ("%05b", $hour) . sprintf ("%06b", $minute) . sprintf ("%018b", $duration);

# $info68 is a comma-delimited string containing "num_waypoints,long1,lat1,...,longN,latN" where num_waypoints (N) = 0-16

  @fields = split (",", $info68);
  $numpoints = $fields [0];
  $encode .= sprintf ("%05b", $numpoints);
  
  if ($numpoints > 0)
    {
    for ($i=1; $i<=$numpoints; $i++)
      {
      $encode .= sprintf ("%028b", (int($fields[$i*2-1]*600000) & (2**28-1))) . sprintf ("%027b", (int($fields[$i*2]*600000) & (2**27-1)));
      }
    }

  return $encode;
  }

# encode_0608_001_29_30__367_29 -- Build Type 6, DAC 1, FI 30 (Text Description - Addressed) and Type 8, DAC 1, FI 29 (Text Description - Broadcast) [IMO Circular 289]
# This message is variable length and can be up to 1028 (6) or 1032 (8) bits in length.

# Also supports Types 6 and 8, DAC 367, FI 29 (Linked Text -- Addressed/Broddcast) [USCG RDC]

sub encode_0608_001_29_30__367_29
  {
# Parameter list
  my ($type,$dac,$fi,$target,$msglink,$info68) = @_;
  my $encode;

################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  38-39      2    38-39      2   Sequence Number (6)/Spare (8)
#  40-69     30                   Destination MMSI
#  70         1                   Retransmit flag             ** Not currently supported; set to 0
#  71         1                   Spare
#  72-81     10    40-49     10   DAC
#  82-87      6    50-55      6   FI
#  88-97     10    56-65     10   Message Linkage ID
#  98-1027  930    66-1031  966   Text string

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  if ($type == 6)
      {
      @fields = split (",", $target);
      $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00";
      }
    else
     { $encode = "00"; }

  $encode .= sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . sprintf ("%010b", $msglink) . encode_string ($info68);
  return $encode;
  }

# encode_08_001_31 -- Designated Area Code (DAC) = 001 and Functional ID (FI) = 31 (Meteorological and Hydrological Data) [IMO 289].
# This message is 360 bits in length.

sub encode_08_001_31
  {
# Parameter list
  my ($dac,$fi,$long,$lat,$accuracy,$day,$hour,$minute,$wspeed,$wgust,$wdir,$wgustdir,$pressure,$presstend,$vis,$waveheight,$waveperiod,$wavedir,
      $swellheight,$swellperiod,$swelldir,$seastate,$precip,$ice) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-80     25   Longitude
#  81-104    24   Latitude
# 105         1   Position accuracy
# 106-110     5   UTC day
# 111-115     5   UTC hour
# 116-121     6   UTC minute
# 122-128     7   Average Wind Speed
# 129-135     7   Gust Speed
# 136-144     9   Wind Direction
# 145-153     9   Wind Gust Direction
# 154-164    11   Air Temperature
# 165-171     7   Relative Humidity             ** Not currently supported; set to 101
# 172-181    10   Dew Point                     ** Not currently supported; set to 501
# 182-190     9   Air Pressure
# 191-192     2   Pressure Tendency
# 193-200     8   Maximum Visibility
# 201-212    12   Water Level                   ** Not currently supported; set to 4001
# 213-214     2   Water Level Trend             ** Not currently supported; set to 3
# 215-222     8   Surface Current Speed         ** Not currently supported; set to 255
# 223-231     9   Surface Current Direction     ** Not currently supported; set to 360
# 232-239     8   Surface Current Speed #2      ** Not currently supported; set to 255
# 240-248     9   Surface Current Direction #2  ** Not currently supported; set to 360
# 249-253     5   Measurement Depth #2          ** Not currently supported; set to 31
# 254-261     8   Surface Current Speed #3      ** Not currently supported; set to 255
# 262-270     9   Surface Current Direction #3  ** Not currently supported; set to 360
# 271-275     5   Measurement Depth #3          ** Not currently supported; set to 31
# 276-283     8   Wave Height
# 284-289     6   Wave Period
# 290-298     9   Wave Direction
# 299-306     8   Swell Height
# 307-312     6   Swell Period
# 313-321     9   Swell Direction
# 322-325     4   Sea State
# 326-335    10   Water Temperature             ** Not currently supported; set to 501
# 336-338     3   Precipitation
# 339-347     9   Salinity                      ** Not currently supported; set to 510
# 348-349     2   Ice
# 350-359    10   Spare

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . sprintf ("%025b", (int($long*60000) & (2**25-1)))
  . sprintf ("%024b", (int($lat*60000) & (2**24-1))) . sprintf ("%01b", int($accuracy)) . sprintf ("%05b", int ($day)) . sprintf ("%05b", int ($hour))
  . sprintf ("%06b", int ($minute)) . sprintf ("%07b", int ($wspeed)) . sprintf ("%07b", int ($wgust)) . sprintf ("%09b", int ($wdir))
  . sprintf ("%09b", int ($wgustdir)) . sprintf ("%011b", (int($airtemp*10) & (2**11-1))) . "1100101"."0111110101" . sprintf ("%09b", int ($pressure))
  . sprintf ("%02b", int ($presstend)) . sprintf ("%08b", int($vis*10+0.5)) . "111110100001". "11" . "11111111" . "101101000" . "11111111"."101101000"."11111"
  . "11111111"."101101000"."11111" . sprintf ("%08b", int($waveheight*10+0.5)) . sprintf ("%06b", int ($waveperiod)) . sprintf ("%09b", int ($wavedir))
  . sprintf ("%08b", int($swellheight*10+0.5)) . sprintf ("%06b", int ($swellperiod)) . sprintf ("%09b", int ($swelldir)) . sprintf ("%04b", int ($seastate))
  . "0111110101" . sprintf ("%03b", int ($precip)) . "111111110" . sprintf ("%02b", int ($ice)) . "0"x10;

  return $encode;
  }

# 06_235_250_10 -- Build Type 6, DAC 235/250, FI 10 (Aids to navigation monitoring data) [UK Trinity House]
# This message is 136 bits in length.

sub encode_06_235_250_10
  {
# Parameter list
  my ($dac,$fi,$target,$info68,$offpos) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description                                                  FIELDS
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-97     10   Analogue (internal supply voltage to AIS unit)                  0
#  98-107    10   Analogue (external – from hardware analogue input No 1)         1
# 108-117    10   Analogue (external – from hardware analogue input No 2)         2
# 118-122     5   Status (internal)
# 123-130     8   Status (external)                                               3
# 131         1   Off position status
# 132-135     4   Spare (set to 0)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi);
  
# $info68 is a comma-delimited string containing four of the next five message parameters (not "internal" status)

  @fields = split (",", $info68);
  $encode .= sprintf ("%010b", int($fields [0])) . sprintf ("%010b", int($fields [1])) . sprintf ("%010b", int($fields [2])) . sprintf ("%05b", $aid_status)
  . sprintf ("%08b", $fields [3]) . sprintf ("%1b", $offpos) . "0000";
  return $encode;
  }

# 06_235_20 -- Build Type 6, DAC 235, FI 20 (Buoy position monitoring) [UK Trinity House]
# This message is 368 bits in length.

sub encode_06_235_20
  {
# Parameter list
  my ($dac,$fi,$target,$aid_name,$day,$hour,$minute,$long,$lat,$offpos) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-291   204   AtoN name
# 292-296     5   UTC day
# 297-301     5   UTC hour
# 302-307     6   UTC minute
# 308-335    28   Longitude
# 336-362    27   Latitude
# 363         1   Off position status
# 364-367     4   Spare (set to 0)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi)
    . encode_string (pad_string ($aid_name, 34)) . sprintf ("%05b", int($day)) . sprintf ("%05b", int($hour)) . sprintf ("%06b", int($minute))
    . sprintf ("%028b", (int($long*600000) & (2**28-1))) . sprintf ("%027b", (int($lat*600000) & (2**27-1)))
    . sprintf ("%1b", $offpos) . "0000";

  return $encode;
  }

# encode_08_366_20 -- Designated Area Code (DAC) = 366 and Functional ID (FI) = 20 (Dam Status) [USCG RDC, V.0].
# This message is 168-408 bits in length.

sub encode_08_366_20
  {
# Parameter list
  my ($dac,$fi,$msglink,$hour,$minute,$lockinfo,$info68) = @_;
  my ($encode,$i,$n,$tyui);

################################################
#  Bits     Len   Description                             FIELDS
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-58      3   Message version (0)
#  59-68     10   Message Linkage ID
#  69-86     18   UN-LOCODE                                  0 |
#  87-110    24   Navigation Unit Identifier                 1 |
# 111-115     5   UTC hour                                     |
# 116-121     6   UTC minute                                   |
# 122-134    13   Dam gate total                             2 |+--- lockinfo
# 135-146    12   Dam discharge                              3 |
# 147-148     2   Dam discharge source                       4 |
# 149-160    12   Hydroelectric plant discharge              5 |
# 161-162     2   Hydroelectric plant discharge source       6 |
# 163-167     5   Number of gates (N)                        0   |+--- info68
# 168-407 0-240   N octets of dam gate info                 1-30 |

# $lockinfo is a comma-delimited string containing several items of lock information

  @fields = split (",", $lockinfo);

  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . "000" . sprintf ("%10b", $msglink) . encode_string (pad_string ($fields[0],3))
  . encode_string (pad_string ($fields[1],4)) . sprintf ("%05b", int ($hour)) . sprintf ("%06b", int ($minute)) . sprintf ("%013b", int($fields[2]*10));
  for ($i = 0; $i < 2; $i++)
    { $encode .= sprintf ("%012b", int($fields [3+($i*2)] * 10)) . sprintf ("%02b", $fields [4+($i*2)]); }

# $info68 is a comma-delimited string containing number of gates (N) and gate info for N gates

  @fields = split (",", $info68);
  $n = $fields[0];  $encode .= sprintf ("%05b", $n);
  if ($n > 0)
    {
    for ($i = 1; $i <= $n; $i++)
      { $encode .= sprintf ("%08b", int($fields [$i] * 10)); }
    }
  return $encode;  
  }

# encode_06_367_16 -- Designated Area Code (DAC) = 367 and Functional ID (FI) = 16 (Passenger and Crew Count – Addressed ) [USCG RDC, V.1a].
# This message is 152 bits in length.

sub encode_06_367_16
  {
# Parameter list
  my ($dac,$fi,$target,$msglink,$day,$hour,$minute,$pob,$info68) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Sequence Number
#  40-69     30   Destination MMSI
#  70         1   Retransmit flag             ** Not currently supported; set to 0
#  71         1   Spare
#  72-81     10   DAC
#  82-87      6   FI
#  88-93      6   Message version (1)
#  94-103    10   Message Linkage ID
# 104-108     5   UTC day
# 109-113     5   UTC hour
# 114-119     6   UTC minute
# 120-132    13   Total crew members
# 133-147    15   Total passengers
# 148-151     4   Additional vessel type information

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  @fields = split (",", $target);
  $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . "000001"
  . sprintf ("%010b", $msglink) . sprintf ("%05b", int ($day)) . sprintf ("%05b", int ($hour)) . sprintf ("%06b", int ($minute));

# $pob is a comma-delimited string containing "crew count,passenger count"

  @fields = split (",", $pob);
  if ($info68 eq "") { $info68 = 0; }

  $encode .= sprintf ("%013b", $fields [0]) . sprintf ("%015b", $fields [1]) . sprintf ("%04b", int ($info68));

  return $encode;  
  }

# encode_0608_367_22 -- Designated Area Code (DAC) = 367, and Functional ID (FI) = 22 [Geographic Notice (USCG]].
# This message is variable length and can be up to 1016 (6) or 984 (8) bits in length.

# NOTE that this message is ALMOST identical to the Type 6, DAC 1, FI 23 and Type 8, DAC 1, FI 22 messages. It is different
# enough that I added this routine. Coded per USCG RDC document dated 23MAR2015.

sub encode_0608_367_22
  {
# Parameter list (NOTE: subarea array passed by reference)
  my ($type,$dac,$fi,$target,$msglink,$notice,$month,$day,$hour,$minute,$duration,$subarea_ref) = @_;
  my @subarea = @{$subarea_ref};  # Dereference array
  my ($encode,$fields,$i);

################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  38-39      2    38-39      2   Sequence Number (6)/Spare (8)
#  40-69     30                   Destination MMSI
#  70         1                   Retransmit flag             ** Not currently supported; set to 0
#  71         1                   Spare
#  72-81     10    40-49     10   DAC
#  82-87      6    50-55      6   FI
#  88-93      6    56-61      6   Message Version (2)
#  94-103    10    62-71     10   Message Linkage ID
# 104-110     7    72-78      7   Notice Description
# 111-114     4    79-82      4   UTC month
# 115-119     5    83-87      5   UTC day
# 120-124     5    88-92      5   UTC hour
# 125-130     6    93-98      6   UTC minute
# 131-148    18    99-116    18   Duration
#                 117         1   Action (set to 0)
# 149-151     3   118-119     2   Spare
# 152-247    96   120-215    96   Sub-area 1
# 248-343    96   216-311    96   Sub-area 2 (if present)
# 344-439    96   312-407    96   Sub-area 3 (if present)
# 440-535    96   408-503    96   Sub-area 4 (if present)
# 536-631    96   504-599    96   Sub-area 5 (if present)
# 632-727    96   600-695    96   Sub-area 6 (if present)
# 728-823    96   696-791    96   Sub-area 7 (if present)
# 824-919    96   792-887    96   Sub-area 8 (if present)
# 920-1015   96   888-983    96   Sub-area 9 (if present)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  if ($type == 6)
      {
      @fields = split (",", $target);      
      $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00";
      }
    else
     { $encode = "00"; }

  $encode .= sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . sprintf ("%06b", 2) . sprintf ("%010b", $msglink) . sprintf ("%07b", $notice)
  . sprintf ("%04b", $month) . sprintf ("%05b", $day) . sprintf ("%05b", $hour) . sprintf ("%06b", $minute) . sprintf ("%018b", $duration)
  . "0" . "00";

# This message contains 1-9 subareas; the 10-element $subarea array contains one comma-delimited string per area, where each string
# lists the pertinent subarea parameters

  for ($i=1; $i<=9; $i++)
    {
    if ($subarea [$i] eq "")
      { last; }

    @fields = split (",", $subarea [$i]);

    SHAPE:
      {

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (0)      0
#   3-4       2   Scale Factor        1
#   5-32     28   Longitude           2
#  33-59     27   Latitude            3
#  60-62      3   Precision           4
#  63-74     12   Radius              5
#  75-95     21   Not used

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

      $fields [0] == 0 && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%028b", (int($fields[2]*600000) & (2**28-1)))
       . sprintf ("%027b", (int($fields[3]*600000) & (2**27-1))) . sprintf ("%03b", int($fields[4])) . sprintf ("%012b", int($fields[5])) . "0"x21;
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (1)      0
#   3-4       2   Scale Factor        1
#   5-32     28   Longitude           2
#  33-59     27   Latitude            3
#  60-62      3   Precision           4
#  63-70      8   E Dimension         5
#  71-78      8   N Dimension         6
#  79-87      9   Orientation         7
#  88-95      8   Not used

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

      $fields [0] == 1 && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%028b", (int($fields[2]*600000) & (2**28-1)))
       . sprintf ("%027b", (int($fields[3]*600000) & (2**25-1))) . sprintf ("%03b", int($fields[4])) . sprintf ("%08b", int($fields[5]))
       . sprintf ("%08b", int($fields[6])) . sprintf ("%09b", int($fields[7])) . "0"x8;
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (2)      0
#   3-4       2   Scale Factor        1
#   5-32     28   Longitude           2
#  33-59     27   Latitude            3
#  60-62      3   Precision           4
#  63-74     12   Radius              5
#  75-83      9   Left Boundary       6
#  84-92      9   Right Boundary      7
#  93-95      3   Not used

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

      $fields [0] == 2 && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%028b", (int($fields[2]*600000) & (2**28-1)))
       . sprintf ("%027b", (int($fields[3]*600000) & (2**27-1))) . sprintf ("%03b", int($fields[4])) . sprintf ("%012b", int($fields[5]))
       . sprintf ("%09b", int($fields[6])) . sprintf ("%09b", int($fields[7])) . "000";
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (3,4)    0
#   3-4       2   Scale Factor        1
#   5-14     10   Point 1 Angle       2
#  15-25     11   Point 1 Distance    3
#  26-35     10   Point 2 Angle       4
#  36-46     11   Point 2 Distance    5
#  47-56     10   Point 3 Angle       6
#  57-67     11   Point 3 Distance    7
#  68-77     10   Point 4 Angle       8
#  78-88     11   Point 4 Distance    9
#  89-95      7   Not used

      ($fields [0] == 3 || $fields [0] == 4) && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . sprintf ("%02b", $fields [1]) . sprintf ("%010b", int($fields[2])) . sprintf ("%011b", int($fields[3]))
        . sprintf ("%010b", int($fields[4])) . sprintf ("%011b", int($fields[5])) . sprintf ("%010b", int($fields[6])) . sprintf ("%011b", int($fields[7]))
        . sprintf ("%010b", int($fields[8])) . sprintf ("%011b", int($fields[9])) . "0"x7;
        last SHAPE;
        };

################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (5)      0
#   3-92     90   Text                1   (15 6-bit characters)
#  93-95      3   Not used

      $fields [0] == 5 && do
        {
        $encode .= sprintf ("%03b", $fields [0]) . encode_string (pad_string ($fields [1], 15)) . "000";
        last SHAPE;
        };
      };
    };

  return $encode;
  }

# encode_08_367_23 -- Designated Area Code (DAC) = 367 and Functional ID (FI) = 23 (Satellite Ship Weather) [USCG RDC, V.0].
# This message is 168 bits in length.

sub encode_08_367_23
  {
# Parameter list
  my ($dac,$fi,$day,$hour,$minute,$long,$lat,$pressure,$airtemp,$wspeed,$wgust,$wdir) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-58      3   Message version (0)
#  59-63      5   UTC day
#  64-68      5   UTC hour
#  69-74      6   UTC minute
#  75-99     25   Longitude
# 100-123    24   Latitude
# 124-132     9   Air pressure
# 133-143    11   Air temperature
# 144-150     7   Wind speed
# 151-157     7   Wind gust
# 158-166     9   Wind direction
# 167         1   Spare

  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . "000" . sprintf ("%05b", int ($day)) . sprintf ("%05b", int ($hour)) . sprintf ("%06b", int ($minute))
  . sprintf ("%025b", (int($long*60000) & (2**25-1))) . sprintf ("%024b", (int($lat*60000) & (2**24-1))) . sprintf ("%09b", $pressure)
  . sprintf ("%011b", (int($airtemp*10) & (2**11-1))) . sprintf ("%07b", $wspeed) . sprintf ("%07b", $wgust) . sprintf ("%09b", $wdir) . "0";

  return $encode;  
  }

# encode_08_367_24 -- Designated Area Code (DAC) = 367 and Functional ID (FI) = 24 (Satellite Ship Weather Small) [USCG RDC, V.0].
# This message is 152 bits in length.

sub encode_08_367_24
  {
# Parameter list
  my ($dac,$fi,$hour,$minute,$long,$lat,$pressure) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-58      3   Message version (0)
#  59-63      5   UTC hour
#  64-69      6   UTC minute
#  70-94     25   Longitude
#  95-118    24   Latitude
# 119-127     9   Air pressure

  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . "000" . sprintf ("%05b", int ($hour)) . sprintf ("%06b", int ($minute))
  . sprintf ("%025b", (int($long*60000) & (2**25-1))) . sprintf ("%024b", (int($lat*60000) & (2**24-1))) . sprintf ("%09b", $pressure);

  return $encode;  
  }

# encode_08_367_25 -- Designated Area Code (DAC) = 367 and Functional ID (FI) = 25 (Satellite Ship Weather Tiny) [USCG RDC, V.0].
# This message is 96 bits in length.

sub encode_08_367_25
  {
# Parameter list
  my ($dac,$fi,$hour,$minute,$pressure,$wspeed,$wdir) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-58      3   Message version (0)
#  59-63      5   UTC hour
#  64-69      6   UTC minute
#  70-78      9   Air pressure
#  79-85      7   Wind speed
#  86-94      9   Wind direction
#  95         1   Spare

  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . "000" . sprintf ("%05b", int ($hour)) . sprintf ("%06b", int ($minute))
  . sprintf ("%09b", $pressure) . sprintf ("%07b", $wspeed) . sprintf ("%09b", $wdir) . "0";

  return $encode;  
  }

# encode_08_367_26 -- Designated Area Code (DAC) = 367 and Functional ID (FI) = 26 (GPS Jamming and Spoofing Report) [USCG RDC, V.0].
# This message is 128 bits in length.

sub encode_08_367_26
  {
# Parameter list
  my ($dac,$fi,$day,$hour,$minute,$long,$lat,$info68) = @_;
  my ($encode,$fields);

################################################
#  Bits     Len   Description              FIELDS
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-58      3   Message version (0)
#  59-63      5   UTC day
#  64-68      5   UTC hour
#  69-74      6   UTC minute
#  75-99     25   Longitude
# 100-123    24   Latitude
# 124         1   Type of event               0
# 125         1   Event status                1
# 126-127     2   Spare

  @fields = split (",", $info68);
  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . "000" . sprintf ("%05b", int ($day)) . sprintf ("%05b", int ($hour))
  . sprintf ("%06b", int ($minute)) . sprintf ("%025b", (int($long*60000) & (2**25-1))) . sprintf ("%024b", (int($lat*60000) & (2**24-1))) 
  . sprintf ("%01b", $fields [0]) . sprintf ("%01b", $fields [1]) . "00";
  return $encode;  
  }

# encode_0608_367_33 -- Build Type8, DAC 367, FI 33 (Environmental Management) [USCG RDC, V.3a]
# These messages are variable length and can be up to 952 bits in length.

sub encode_08_367_33
  {
# Parameter list (NOTE: sensor_info (nee subarea) array passed by reference)
  my ($type,$dac,$fi,$sensor_info_ref) = @_;
  my @sensor_info = @{$sensor_info_ref};  # Dereference array
  my ($encode,$fields,$n,$rep_type);

################################################
#  Bits     Len    Bits     Len   Description
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-167   112   Sensor Report 1
# 168-279   112   Sensor Report 2 (optional)
# 280-391   112   Sensor Report 3 (optional)
# 392-503   112   Sensor Report 4 (optional)
# 504-615   112   Sensor Report 5 (optional)
# 616-727   112   Sensor Report 6 (optional)
# 728-839   112   Sensor Report 7 (optional)
# 840-951   112   Sensor Report 8 (optional)

# Encode here the "header" block for the message...

  $encode = "00" . sprintf ("%010b", $dac) . sprintf ("%06b", $fi);

# Sensor Report
#-----------------------------------------------
#  Bits     Len   Description         FIELDS
#   0-3       4   Report Type            0
#   4-8       5   UTC day                1
#   9-13      5   UTC hour               2
#  14-19      6   UTC minute             3
#  20-26      7   Site ID                4
#  27-111    85   Sensor Data

# Now, cycle through 1 to 8 sensor reports... $sensor_info is a comma-delimited string of sensor data. Start by encoding the sensor report common header information
# (first 27 bits), and then get the sensor data (next 85 bits) that is specific to the report type

  $n = 1;
  while ($sensor_info [$n] ne "")
    {
    @fields = split (",", $sensor_info [$n]);
    $rep_type = $fields [0];
    $encode .= sprintf ("%04b", $rep_type) . sprintf ("%05b", $fields[1]) . sprintf ("%05b", $fields[2]) . sprintf ("%06b", $fields[3])
    . sprintf ("%07b", $fields[4]);
    $n++;

# Now get sensor data based on report type...

    SENSOR_TYPE_HANDLER:
      {
        
# Sensor Type 0 - Site Location
################################################
#  Bits     Len   Description          FIELDS
#   0-5       6   Message Version         5
#   6-33     28   Longitude               6
#  34-60     27   Latitude                7
#  61-63      3   Precision               8
#  64-75     12   Altitude                9
#  76-79      4   Sensor Owner           10
#  80-82      3   Data Timeout           11
#  83-84      2   Spare

      $rep_type == 0 && do
        {
        $encode .= sprintf ("%06b", $fields[5]) . sprintf ("%028b", (int($fields[6]*600000) & (2**28-1))) . sprintf ("%027b", (int($fields[7]*600000) & (2**27-1)))
        . sprintf ("%03b", $fields[8]) . sprintf ("%012b", $fields[9]) . sprintf ("%012b", (int($fields[9]*10) & (2**12-1))) . sprintf ("%04b", $fields[10])
        . sprintf ("%03b", $fields[11]) . "00";
        last SENSOR_TYPE_HANDLER;
        };
  
# Sensor Type 1 - Station ID
################################################
#  Bits     Len   Description          FIELDS
#   0-83     84   Name                    5
#  84         1  Spare

      $rep_type == 1 && do
        {
        $encode .= encode_string (pad_string ($fields[5],14)) . "0";
        last SENSOR_TYPE_HANDLER;
        };

# Sensor Type 2 - Wind Report
################################################
#  Bits     Len   Description               FIELDS
#   0-6       7   Wind Speed                   5
#   7-13      7   Wind Gust                    6
#  14-22      9   Wind Direction               7
#  23-31      9   Wind Gust Direction          8
#  32-34      3   Sensor Data Description      9
#  35-41      7   Forecast Wind Speed         10
#  48-48      7   Forecast Wind Gust          11
#  49-57      9   Forecast Wind Direction     12
#  58-62      5   UTC Date                    13
#  63-67      5   UTC Hour                    14
#  68-73      6   UTC Minute                  15
#  74-81      8   Duration of Forecast        16
#  82-84      3   Spare

      $rep_type == 2 && do
        {
        $encode .= sprintf ("%07b", $fields[5]) . sprintf ("%07b", $fields[6]) . sprintf ("%09b", $fields[7]) . sprintf ("%09b", $fields[8])
        . sprintf ("%03b", $fields[9]) . sprintf ("%07b", $fields[10]) . sprintf ("%07b", $fields[11]) . sprintf ("%09b", $fields[12])
        . sprintf ("%05b", $fields[13]) . sprintf ("%05b", $fields[14]) . sprintf ("%06b", $fields[15]) . sprintf ("%08b", $fields[16]) . "000";
        last SENSOR_TYPE_HANDLER;
        };

# Sensor Type 3	- Tide/River Stage (Water Level) Report
################################################
#  Bits     Len   Description                  FIELDS
#   0         1   Water Level Type                5
#   1-16     16   Water Level                     6
#  17-18      2   Water Level Trend               7
#  19-23      5   Vertical Reference Datum        8
#  24-26      3   Sensor Data Description         9
#  27         1   Forecast Water Level Type      10
#  28-43     16   Forecast Water Level           11
#  44-48      5   UTC Date                       12
#  49-53      5   UTC Hour                       13
#  54-59      6   UTC Minute                     14
#  60-67      8   Duration of Forecast           15
#  68-84     17   Spare

      $rep_type == 3 && do
        {
        $encode .= sprintf ("%01b", $fields[5]) . sprintf ("%016b",(int($fields[6]*100) & (2**16-1))) . sprintf ("%02b", $fields[7]) . sprintf ("%05b", $fields[8])
        . sprintf ("%03b", $fields[9]) . sprintf ("%01b", $fields[10]) . sprintf ("%016b", (int($fields[11]*100) & (2**16-1))) . sprintf ("%05b", $fields[12])
        . sprintf ("%05b", $fields[13]) . sprintf ("%06b", $fields[14]) . sprintf ("%08b", $fields[15]) . "0"x17 ;
        last SENSOR_TYPE_HANDLER;
        };

# Sensor Type 4	- Two-Dimensional (x & y) Vertical Current Profile Report
################################################
#  Bits     Len   Description                  FIELDS
#   0-7       8   Current 1 Speed                 5
#   8-16      9   Current 1 Direction             6
#  17-25      9   Current 1 Measuring Level       7
#  26-33      8   Current 2 Speed                 8
#  34-42      9   Current 2 Direction             9
#  43-51      9   Current 2 Measuring Level      10
#  52-59      8   Current 3 Speed                11
#  60-68      9   Current 3 Direction            12
#  69-77      9   Current 3 Measuring Level      13
#  78-80      3   Sensor Data Description        14
#  81-84      4   Spare


      $rep_type == 4 && do
        {
        $encode .= sprintf ("%08b", $fields[5]*10) . sprintf ("%09b", $fields[6]) . sprintf ("%09b", $fields[7]) . sprintf ("%08b", $fields[8]*10)
        . sprintf ("%09b", $fields[9]) . sprintf ("%09b", $fields[10]) . sprintf ("%08b", $fields[11]*10) . sprintf ("%09b", $fields[12])
        . sprintf ("%09b", $fields[13]) . sprintf ("%03b", $fields[14]) . "0000";
        last SENSOR_TYPE_HANDLER;
        };

# Sensor Type 5	- Three-Dimensional (x, y, & z) Vertical Current Profile Report
################################################
#  Bits     Len   Description                         FIELDS
#   0-8       9   Current 1: vector component North      5
#   9-17      9   Current 1: vector component East       6
#  18-26      9   Current 1: vector component Up         7
#  27-35      9   Current 1 Measuring Level              8
#  36-44      9   Current 2: vector component North      9
#  45-53      9   Current 2: vector component East      10
#  54-62      9   Current 2: vector component Up        11
#  63-71      9   Current 2 Measuring Level             12
#  72-74      3   Sensor Data Description               13
#  75-84     10   Spare

      $rep_type == 5 && do
        {
        $encode .= sprintf ("%09b", int($fields[5]*10 & (2**9-1))) . sprintf ("%09b", int($fields[6]*10 & (2**9-1))) . sprintf ("%09b", int($fields[7]*10 & (2**9-1)))
        . sprintf ("%09b", $fields[8]) . sprintf ("%09b", int($fields[9]*10 & (2**9-1))) . sprintf ("%09b", int($fields[10]*10 & (2**9-1)))
        . sprintf ("%09b", int($fields[11]*10 & (2**9-1))) . sprintf ("%09b", $fields[12]) . sprintf ("%03b", $fields[13]) . "0"x10;
        last SENSOR_TYPE_HANDLER;
        };
  
# Sensor Type 6	- Horizontal Current Profile
################################################
#  Bits     Len   Description                         FIELDS
#   0-8       9   Current Reading Bearing                5
#   9-17      9   Current 1 Distance                     6
#  18-25      8   Current 1 Speed                        7
#  26-34      9   Current 1 Direction                    8
#  35-43      9   Current 1 Measuring Level              9
#  44-52      9   Current 2 Distance                    10
#  53-60      8   Current 2 Speed                       11
#  61-69      9   Current 2 Direction                   12
#  70-78      9   Current 2 Measuring Level             13
#  79-81      3   Sensor Data Description               14
#  82-84      3   Spare

      $rep_type == 6 && do
        {
        $encode .= sprintf ("%09b", $fields[5]) . sprintf ("%09b", $fields[6]) . sprintf ("%08b", $fields[7]*10) . sprintf ("%09b", $fields[8])
        . sprintf ("%09b", $fields[9]) . sprintf ("%09b", $fields[10]) . sprintf ("%08b", $fields[11]*10) . sprintf ("%09b", $fields[12])
        . sprintf ("%09b", $fields[13]) . sprintf ("%03b", $fields[14]) . "000";
        last SENSOR_TYPE_HANDLER;
        };

# Sensor Type 7	- Sea State
################################################
#  Bits     Len   Description                                 FIELDS
#   0-7       8   Swell Height                                   5
#   8-13      6   Swell Period                                   6
#  14-22      9   Swell Direction                                7
#  23-26      4   Sea State                                      8
#  27-29      3   Swell Sensor Data Description                  9
#  30-39     10   Water Temperature                             10
#  40-46      7   Water Temperature Depth                       11
#  47-49      3   Water Temperature Sensor Data Description     12
#  50-57      8   Significant Wave Height                       13
#  58-63      6   Wave Period                                   14
#  64-72      9   Wave Direction                                15
#  73-75      3   Wave Sensor Data Description                  16
#  76-84      9   Salinity                                      17

  
      $rep_type == 7 && do
        {
        $encode .= sprintf ("%08b", int ($fields[5])*10) . sprintf ("%06b", int ($fields[6])) . sprintf ("%09b", int ($fields[7])) . sprintf ("%04b", int ($fields[8]))
        . sprintf ("%03b", int ($fields[9])) . sprintf ("%010b", int (($fields[10]+10)*10)) . sprintf ("%07b", int ($fields[11]*10)) . sprintf ("%03b", int ($fields[12]))
        . sprintf ("%08b", int ($fields[13]*10)) . sprintf ("%06b", int ($fields[14])) . sprintf ("%09b", int ($fields[15])) . sprintf ("%03b", int ($fields[16]))
        . sprintf ("%09b", int ($fields[17]*10));
        last SENSOR_TYPE_HANDLER;
        };
 
 # Sensor Type 8	- Salinity
################################################
#  Bits     Len   Description                           FIELDS
#   0-9      10   Water Temperature                        5
#  11-19     10   Conductivity                             6
#  20-35     16   Water Pressure                           7
#  36-44      9   Salinity                                 8
#  45-46      2   Salinity Type                            9
#  47-49      3   Salinity Sensor Data Description        10
#  50-84     35   Spare
  
      $rep_type == 8 && do
        {
        $encode .= sprintf ("%010b", int (($fields[5]+10)*10)) . sprintf ("%10b", int ($fields[6]*100)) . sprintf ("%016b", int ($fields[7]*10))
        . sprintf ("%09b", int ($fields[8]*10)) . sprintf ("%02b", int ($fields[9])) . sprintf ("%03b", int ($fields[10])) . "0"x35;
        last SENSOR_TYPE_HANDLER;
        };
  
# Sensor Type 9	- Weather
################################################
#  Bits     Len   Description                                FIELDS
#   0-10     11   Air Temperature                               5
#  11-13      3   Air Temperature Sensor Data Description       6
#  14-15      2   Precipitation Type                            7
#  16-23      8   Horizontal Visibility                         8
#  24-33     10   Dew Point                                     9
#  34-36      3   Dew Point Sensor Data Description            10
#  37-45      9   Air Pressure                                 11
#  46-47      2   Air Pressure Trend                           12
#  48-50      3   Air Pressure Sensor Data Description         13
#  51-59      9   Salinity                                     14
#  60-84     25   Spare

      $rep_type == 9 && do
        {
        $encode .= sprintf ("%011b", (int($fields[5]*10) & (2**11-1))) . sprintf ("%03b", int ($fields[6])) . sprintf ("%02b", int ($fields[7]))
        . sprintf ("%08b", int ($fields[8]*10)) . sprintf ("%010b", int (($fields[9]+20)*10)) . sprintf ("%03b", int ($fields[10]))
        . sprintf ("%09b", int ($fields[11])-799) . sprintf ("%02b", int ($fields[12])) . sprintf ("%03b", int ($fields[13]))
        . sprintf ("%09b", int ($fields[14]*10)) . "0"x25;
        last SENSOR_TYPE_HANDLER;
        };

# Sensor Type 10	- Air Gap
################################################
#  Bits     Len   Description                       FIELDS
#   0-12     13   Air Draught                          5
#  13-25     13   Air Gap                              6
#  26-27      2   Air Gap Trend                        7
#  28-40     13   Predicted Air Gap                    8
#  41-45      5   UTC Date                             9
#  46-50      5   UTC Hour                            10
#  51-56      6   UTC Minute                          11
#  57-59      3   Air Gap Sensor Data Description     12
#  68-84     25   Spare

      $rep_type == 10 && do
        {
        $encode .= sprintf ("%013b", int($fields[5])) . sprintf ("%013b", int($fields[6])) . sprintf ("%02b", int($fields[7])) . sprintf ("%013b", int($fields[8]))
        . sprintf ("%05b", int($fields[9])) . sprintf ("%05b", int($fields[10])) . sprintf ("%06b", int($fields[11])) . sprintf ("%03b", $fields[12]) . "0"x25;
        last SENSOR_TYPE_HANDLER;
        };
  
#Sensor Type 11	- Wind report v2
################################################
#  Bits     Len   Description                 FIELDS
#   0-6       7   Wind Speed                     5
#   7-13      7   Wind Gust                      6
#  14-22      9   Wind Direction                 7
#  23-28      6   Averaging Time                 8
#  29-31      3   Sensor Data Description        9
#  32-38      7   Forecast Wind Speed           10
#  39-45      7   Forecast Wind Gust            11
#  46-54      9   Forecast Wind Direction       12
#  55-59      5   UTC Hour                      13
#  60-65      6   UTC Minute                    14
#  66-73      8   Duration of Forecast          15
#  74-84     11   Spare

      $rep_type == 11 && do
        {
        $encode .= sprintf ("%07b", $fields[5]) . sprintf ("%07b", $fields[6]) . sprintf ("%09b", $fields[7]) . sprintf ("%06b", $fields[8])
        . sprintf ("%03b", $fields[9]) . sprintf ("%07b", $fields[10]) . sprintf ("%07b", $fields[11]) . sprintf ("%09b", $fields[12])
        . sprintf ("%05b", $fields[13]) . sprintf ("%06b", $fields[14]) . sprintf ("%08b", $fields[15]) . "0"x11;
        last SENSOR_TYPE_HANDLER;
        };
      };
    };
  return $encode;
  }

# encode_0608_367_35 -- Build Types 6 and 8, DAC 367, FI 35 (Watermays Management - Addressed/Broadcast) [USCG RDC, V.2c]
# These messages are variable length and can be up to 496 (6) or 464 (8) bits in length.

sub encode_0608_367_35
  {
# Parameter list (NOTE: vessel_info (nee subarea) array passed by reference)
  my ($type,$dac,$fi,$target,$msglink,$wmtype,$info68,$vessel_info_ref) = @_;
  my @vessel_info = @{$vessel_info_ref};  # Dereference array
  my ($encode,$fields,$i,$pad);

################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  38-39      2    38-39      2   Sequence Number (6)/Spare (8)
#  40-69     30                   Destination MMSI
#  70         1                   Retransmit flag             ** Not currently supported; set to 0
#  71         1                   Spare
#  72-81     10    40-49     10   DAC
#  82-87      6    50-55      6   FI
#  88-97     10    56-65     10   Message Linkage ID
#  98-101     4    66-69      4   Waterways Management message type
# 102-103     2    70-71      2   WM message subtype


# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  if ($type == 6)
      {
      @fields = split (",", $target);
      $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00";
      }
    else
     { $encode = "00"; }

  $encode .= sprintf ("%010b", $dac) . sprintf ("%06b", $fi) . sprintf ("%010b", $msglink) . sprintf ("%04b", int($wmtype/10))
  . sprintf ("%02b", ($wmtype % 10));
  
  SUBTYPE_HANDLER:
    {
  
 # Subtype 0 - Vessel Information
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
# 104-159    56    72-127    56   Vessel #1 information
# 160-215    56   128-183    56   Vessel #2 information (optional)
# 216-271    56   184-239    56   Vessel #3 information (optional)
# 272-327    56   240-295    56   Vessel #4 information (optional)
# 328-383    56   296-351    56   Vessel #5 information (optional)
# 384-439    56   352-407    56   Vessel #6 information (optional)
# 440-495    56   408-463    56   Vessel #7 information (optional)

# Vessel Information
#-----------------------------------------------
#  Bits     Len   Description         FIELDS
#   0-29     30   MMSI                   0
#  30-34      5   UTC day                1
#  35-39      5   UTC hour               2
#  40-45      6   UTC minute             3
#  46-47      2   Vessel status          4
#  48-50      3   Chamber                5
#  51         1   Direction              6
#  52-55      4   Lockage Type           7

# This message contains 1-7 sets of vessel informationn arrays, each with the eight elements above.

    ($wmtype % 10) == 0 && do
      {
      $i = 1;
      while ($vessel_info [$i] ne "")
        {
        @fields = split (",", $vessel_info [$i]);
        $encode .= sprintf ("%030b", $fields[0]) . sprintf ("%05b", $fields[1]) . sprintf ("%05b", $fields[2]) . sprintf ("%06b", $fields[3])
        . sprintf ("%02b", $fields[4]) . sprintf ("%03b", $fields[5]) . sprintf ("%01b", $fields[6]) . sprintf ("%04b", $fields[7]);
        $i++;
        }
      last SUBTYPE_HANDLER;
      };
    
# Subtype 1 - WM	Place	Identification
################################################
#  == Type 6 ==    == Type 8 ==                                      FIELDS
#  Bits     Len    Bits     Len   Description
# 104-109     6    72-77      6   Message version (2)
# 110-137    28    78-105    28   Longitude                             0
# 138-164    27   106-132    27   Latitude                              1
# 165-167     3   133-135     3   Precision                             2
# 168-176     9   136-144     9   Orientation                           3
# 177-386   210   145-354   210   Description (up to 35 characters)     4
# 387-391    <8   155-359    <8   Spare (octet-alignment)

    ($wmtype % 10) == 1 && do
      {
      @fields = split (",", $info68);
      $encode .= "000010" . sprintf ("%028b", (int($fields[0]*600000) & (2**28-1))) . sprintf ("%027b", (int($fields[1]*600000) & (2**27-1)))
      . sprintf ("%03b", $fields[2]) . sprintf ("%09b", $fields[3]) . encode_string ($fields[4]);
      last SUBTYPE_HANDLER;
      };

# Subtype 2 - Vessel Name
################################################
#  == Type 6 ==    == Type 8 ==                                      FIELDS
#  Bits     Len    Bits     Len   Description
# 104-113    10    72-81     10   Vessel number                         0
# 114-116     3    82-84      3   Descriptor type                       1
# 117-326   210    85-294   210   Vessel name (up to 35 characters)     2
# 327        <8   295        <8   Spare (octet-alignment)
    
    ($wmtype % 10) == 2 && do
      {
      @fields = split (",", $info68);
      $encode .= sprintf ("%010b", $fields[0]) . sprintf ("%03b", $fields[1]) . encode_string ($fields[2]);
      last SUBTYPE_HANDLER;
      };
    }

  return $encode;
  }

##### END OF TYPES 6 & 8 ######

# *** Encode_0713 builds a Type 7, Binary Acknowledge message or a Type 13, Safety-Related Acknowledgement ***
# Expected to be 72-168 bits (in 32-bit increments), occupying one AIS sentence.

sub encode_0713
  {
# Parameter list
  my ($target) = @_;
  my ($encode, $fields, $i);

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-69     30   Target MMSI #1
#  70-71      2   Sequence number for MMSI #1
#  72-101    30   Target MMSI #2 (if present)
# 101-103     2   Sequence number for MMSI #2
# 104-133    30   Target MMSI #3 (if present)
# 134-135     2   Sequence number for MMSI #3
# 136-165    30   Target MMSI #3 (if present)
# 166-167     2   Sequence number for MMSI #4

  $encode = "00";

# $target is a comma-delimited string containing "target_MMSI_1,sequence_number_1,...,target_MMSI_4,sequence_number_4"
  
  @fields = split (",", $target);

  for ($i=0; $i<@fields; $i=$i+2)
    { $encode .= sprintf ("%030b%02b", $fields [$i], $fields [$i+1]); }

  return $encode;
  }

# *** Encode_09 builds Type 9, Standard SAR Aircraft Position Report messages ***
# Expected to be 168 bits, occupying one AIS sentence.

sub encode_09
  {
# Parameter list
  my ($altitude,$airspeed,$accuracy,$long,$lat,$course,$ts,$dte,$raim) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-49     12   Altitude
#  50-59     10   Speed Over Ground (SOG)
#  60         1   Position Accuracy 
#  61-88     28   Longitude
#  89-115    27   Latitude
# 116-127    12   Course Over Ground (COG)
# 128-133     6   Time Stamp
# 134-141     8   Regional Reserved             ** Not currently supported; set to 0
# 142         1   DTE
# 143-145     3   Spare
# 146         1   Assigned                      ** Not currently supported; set to 0 = Autonomous mode (default)
# 147         1   RAIM flag
# 148-167    20   Radio Status                  ** Not currently supported; set to 0

  $encode = sprintf ("%012b", int($altitude)) . sprintf ("%010b", int($airspeed)) . sprintf ("%01b", int($accuracy))
  . sprintf ("%028b", (int($long*600000) & (2**28-1))) . sprintf ("%027b", (int($lat*600000) & (2**27-1))) . sprintf ("%012b", int($course*10+0.5))
  . sprintf ("%06b", int($ts)) . "00000000" . sprintf ("%1b", int ($dte)) . "0"x4 . sprintf ("%01b", int($raim)) . "0"x20 ;
  return $encode;
  }

# *** Encode_10 builds a Type 10, UTC/Date Inquiry message ***
# Message is 72 bits in length.

sub encode_10
  {
# Parameter list
  my ($target) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-69     30   Target MMSI
#  70-71      2   Spare

  $encode = "00" . sprintf ("%030b", int($target)) . "00";
  return $encode;
  }

# *** Encode_1214 builds a Type 12, Addressed Safety-Related Message or 
# Type 14, Safety-Related Broadcast message. Note the similarity between the two encodings....
# Maximum of 1,008 bits, occupying up to five AIS sentences.

sub encode_1214
  {
# Parameter list
  my ($type,$target,$message) = @_;
  my ($encode,$fields);

################################################
#  = TYPE 12 ==   = TYPE 14 ==
#  Bits     Len   Bits     Len  Description
#  38-39      2   38-39      2   Sequence Number (12)/Spare (14)
#  40-69     30                  Destination MMSI
#  70         1                  Retransmit flag [Set to 0 here]
#  71         1                  Spare
#  72-1007  936   40-1007  966   Message (<=156 [12]/<=161 [14] six-bit characters)

# $target is a comma-delimited string containing "target_MMSI,sequence_number"

  if ($type == 12)
      {
      @fields = split (",", $target);      
      $encode = sprintf ("%02b", $fields [1]) . sprintf ("%030b", $fields [0]) . "00";
      }
    else
     { $encode = "00"; }

  $encode .= encode_string ($message);
  return $encode;
  }

# *** Encode_15 builds a Type 15, Interrogation Message ***
# 88, 110/114, or 160 bits, occupying one AIS sentence.

sub encode_15
  {
# Parameter list
  my ($target) = @_;
  my ($encode,$fields,$n);

################################################
#  Bits     Len   Description
#  38-39      2   Spare (set to 0)
#  40-69     30   Interrogated MMSI 1
#  70-75      6   Message type 1.1
#  76-87     12   Slot offset 1.1         <-- End here if no add'l. queries
#  88-89      2   Spare (set to 0)
#  90-95      6   Message type 1.2
#  96-107    12   Slot offset 1.2
# 108-109     2   Spare (set to 0)        <-- End here if no 2nd MMSI (see notes below)
# 110-139    30   Interrogated MMSI 2
# 140-145     6   Message type 2.1
# 146-157    12   Slot offset 2.2
# 158-159     2   Spare (set to 0)

# $target is a comma-delimited string containing a list of MMSIs and message types/slot offsets, which are then split out into $fields
# $target format is "MMSI1,type1.1,offset1.1,type1.2,offset1.2,MMSI2,type2,offset2"

# There are four possible formats:
# 1) One MMSI and one interrogation/slot offset, values contained in $fields entries 0-2 [$n=3]
# 2) One MMSI and two interrogations/slot offsets, values contained in $fields entries 0-4 [$n=5]
# 3) Two MMSIs each with one interrogation/slot offset, values contained in $fields 0-5 [$n=6]
# 4) Two MMSIs where the first has two interrogations/slot offsets, values contained in $fields 0-7 [$n=8]

  @fields = split (",", $target);
  $n = scalar (@fields);

# All formats begin with a MMSI, message type, and slot offset

  $encode .= "00" . sprintf ("%030b", $fields [0]) . sprintf ("%06b", $fields [1]) . sprintf ("%012b", $fields [2]);

  BUILD_MSG15:
    {

# $n = 5 --> Format 2. This comes to 110 bits. Two additional bits are added at the end to bring to 8-bit alignment
    $n == 5 && do
      {
      $encode .= "00" . sprintf ("%06b", $fields [3]) . sprintf ("%012b", $fields [4]) . "00" . "00";
      last BUILD_MSG15;
      };

# $n == 6 --> Format 3. MMSI1's type and offset fields are set to 0
    $n == 6 && do
      {
      $encode .= "00" . "0"x18 . "00" . sprintf ("%030b", $fields [3]) . sprintf ("%06b", $fields [4]) . sprintf ("%012b", $fields [5]) . "00";
      last BUILD_MSG15;
      };

# $n == 8 --> Format 4.
    $n == 8 && do
      {
      $encode .= "00" . sprintf ("%06b", $fields [3]) . sprintf ("%012b", $fields [4]) . "00" . sprintf ("%030b", $fields [5])
       . sprintf ("%06b", $fields [6]) . sprintf ("%012b", $fields [7]) . "00";
      last BUILD_MSG15;
      };
    }
    
  return $encode;
  }

# *** Encode_16 builds a Type 16, Assignment Mode Command Message ***
# 96 or 144 bits, occupying one AIS sentence.

sub encode_16
  {
# Parameter list
  my ($target) = @_;
  my ($encode,$fields,$N);

################################################
#  Bits     Len   Description
#  38-39      2   Spare (set to 0)
#  40-69     30   Destination A MMSI
#  70-81     12   Offset A
#  82-91     10   Increment A            <-- End here if no add'l. queries
#  92-121    30   Destination B MMSI
# 122-133    12   Offset B
# 134-143    10   Increment B

# $target is a comma-delimited string containing a list of one or two MMSI/offset/increment tuples, which are then split out into $fields
# If there is one set, $n = 3; two sets, $n = 6

  @fields = split (",", $target);
  $n = scalar (@fields);

# There will always be one set of data...

  $encode .= "00" . sprintf ("%030b", $fields [0]) . sprintf ("%012b", $fields [1]) . sprintf ("%010b", $fields [2]);

# If there are three fields, add four bits for 8-bit alignment; otherwise, encode second data set

  if ($n == 3)
      { $encode .= "0000"; }
    else
      { $encode .= sprintf ("%030b", $fields [3]) . sprintf ("%012b", $fields [4]) . sprintf ("%010b", $fields [5]); }
           
  return $encode;
  }

# *** Encode_17 builds a Type 17, DGNSS Broadcast Binary Message ***
# 80 bits for the core message followed by 0-736 bits for an M.823 Differential DGNSS message (80-816 bits total, occupying 1-5 AIS sentence)

sub encode_17
  {
# Parameter list
  my ($long,$lat,$dgnss_msg) = @_;
  my ($encode,$fields,$i,$j,$len,$n);

################################################
#  Bits     Len   Description
#  38-39      2   Spare (set to 0)
#  40-57     18   Longitude
#  58-74     17   Latitude
#  75-79      5   Spare (set to 0)
#  80-815   736   Payload (M.823 message)

# Encode bits 0-80 here...

  $encode .= "00" . sprintf ("%018b", (int($long*600) & (2**18-1))) . sprintf ("%017b", (int($lat*600) & (2**17-1))) . "00000";

# ITU-R M.823 Message format
#************************************************************
# TYPE 17   M.823
#  Bits     Bits      Len   Description
#  80-85     0-5        6   Message Type
#  86-95     6-15      10   Station Identifier
#  96-108   16-28      13   Z Count
# 109-111   29-31       3   Sequence Number
# 112-116   32-36       5   Number of DGNSS Data Words (N)
# 117-119   37-39       3   Station Health
# 120-815   40-736  0-696   GNSS data words (in 24-bit increments)

  @fields = split (",",$dgnss_msg);
  $n = int($fields [4]);

  $encode .= sprintf ("%06b", int($fields [0])) . sprintf ("%010b", int($fields [1])) . sprintf ("%013b", int($fields [2])) . sprintf ("%03b", int($fields [3]))
  . sprintf ("%05b", $n) . sprintf ("%03b", int($fields [5]));

# Encode the data words only if the number of data words matches the value of N (there are six non-data fields...)
# This program accepts binary input (24 bits) or hex input (6 digits). If wrong lenth, put in a dummy binary value; otherwise, check the input

  if (scalar @fields > 6)
    {
    for ($i=1; $i<=$n; $i++)
      {
      $x = $fields [$i+5];
      $len = length ($x);
      DATA_WORDS:
        {
        $len != 6 && $len != 24  && do
          {
          $encode .= "0" x 24;
          last DATA_WORDS;
          };
        $len == 6 && do
          {
          for ($j=0; $j<=5; $j++)
            {
            if (substr ($x,$j,1) !~ /[0-9 A-Za-z.,-]/) { substr ($x,$j,1,"0"); } # Replace any invalid hex digit with a 0x0
            $encode .= sprintf ("%04b", hex (substr ($x,$j,1)));
            }
          last DATA_WORDS;
          };
        $len == 24 && do
          {
          for ($j=0; $j==23; $j++)
            {
            if (substr ($x,$j,1) !~ /[0-1]/) { substr ($x,$j,1,"0"); } # Replace any invalid binary digit with a 0
            }
          $encode .= $x;
          last DATA_WORDS;
          };
        };  
      };
    }
   
  return $encode;
  }

# *** Encode_18 builds a Type 18, Standard Class B CS Position Report ***
# 168 bits, occupying one AIS sentence.

sub encode_18
  {
# Parameter list
  my ($speed,$accuracy,$long,$lat,$course,$heading,$ts,$raim) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-45      8   Regional Reserved
#  46-55     10   Speed Over Ground (SOG)
#  56         1   Position Accuracy
#  57-84     28   Longitude
#  85-111    27   Latitude
# 112-123    12   Course Over Ground (COG)
# 124-132     9   True Heading (HDG)
# 133-138     6   Time Stamp
# 139-140     2   Regional Reserved             ** Not currently supported; set to 0
# 141         1   CS Unit                       ** Not currently supported; set to 1 (carrier sense Class B)
# 142         1   Display Flag                  ** Not currently supported; set to 0 (no visual display)
# 143         1   DSC Flag                      ** Not currently supported; set to 1 (unit attached to VHF voice radio with DSC capability)
# 144         1   Band Flag                     ** Not currently supported; set to 1 (unit can use any part of the marine channel)
# 145         1   Message 22 Flag               ** Not currently supported; set to 1 (unit can accept channel assignment via Message type 22)
# 146         1   Assigned                      ** Not currently supported; set to 0 = Autonomous mode (default)
# 147         1   RAIM flag
# 148-167    20   Radio Status                  ** Not currently supported; set to 11100000000000000110

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  $encode = "00000000" . sprintf ("%010b", int($speed*10+0.5)) . sprintf ("%01b", int($accuracy)) . sprintf ("%028b", (int($long*600000) & (2**28-1)))
  . sprintf ("%027b", (int($lat*600000) & (2**27-1))) . sprintf ("%012b", int($course*10+0.5)) . sprintf ("%09b", int($heading))
  . sprintf ("%06b", int($ts)) . "00101110" . sprintf ("%01b", int($raim)) . "11100000000000000110";
  return $encode;
  }

# *** Encode_19 builds a Type 19, Extended Class B CS Position Report ***
# 312 bits, occupying one-two AIS sentences.

sub encode_19
  {
# Parameter list
  my ($speed,$accuracy,$long,$lat,$course,$heading,$ts,$vname,$vtype,$vsize_a,$vsize_b,$vsize_c,$vsize_d,$epfd,$raim,$dte) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-45      8   Regional Reserved
#  46-55     10   Speed Over Ground (SOG)
#  56         1   Position Accuracy
#  57-84     28   Longitude
#  85-111    27   Latitude
# 112-123    12   Course Over Ground (COG)
# 124-132     9   True Heading (HDG)
# 133-138     6   Time Stamp
# 139-142     4   Regional Reserved             ** Not currently supported; set to 0
# 143-262   120   Name
# 263-270     8   Ship Type
# 271-279     9   Dimension to Bow
# 280-288     9   Dimension to Stern
# 289-294     6   Dimension to Port
# 295-300     6   Dimension to Starboard
# 301-304     4   Type of EPFD
# 305         1   RAIM flag
# 306         1   DTE
# 307         1   Assigned                     ** Not currently supported; set to 0 = Autonomous mode (default)
# 308-311     4   Spare

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  $encode = "00000000" . sprintf ("%010b", int($speed*10+0.5)) . sprintf ("%01b", int($accuracy)) . sprintf ("%028b", (int($long*600000) & (2**28-1)))
  . sprintf ("%027b", (int($lat*600000) & (2**27-1))) . sprintf ("%012b", int($course*10+0.5)) . sprintf ("%09b", int($heading))
  . sprintf ("%06b", int($ts)) . "0000" . encode_string (pad_string ($vname,20)) . sprintf ("%08b", int ($vtype))
  . sprintf ("%09b", int ($vsize_a)) . sprintf ("%09b", int ($vsize_b)) . sprintf ("%06b", int ($vsize_c)) . sprintf ("%06b", int ($vsize_d))
  . sprintf ("%04b", int ($epfd)) . sprintf ("%01b", int($raim)) . sprintf ("%1b", int ($dte)) . "0" . "0000";
  return $encode;
  }

# *** Encode_20 builds a Type 20, Data Link Management Message ***
# 72-160 bits in length, depending on the number of slot reservations (1 to 4) in the message, occupying one AIS sentence.
# Pad this message to a byte boundary (see below).

sub encode_20
  {
# Parameter list
  my ($fatdma) = @_;
  my ($encode,$fields,$i);

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-51     12   Offset number 1
#  52-55      4   Reserved slots
#  56-58      3   Time-out
#  59-69     11   Increment
#  70-81     12   Offset number 2
#  82-85      4   Reserved slots
#  86-88      3   Time-out
#  89-99     11   Increment
# 100-111    12   Offset number 3
# 112-115     4   Reserved slots
# 116-118     3   Time-out
# 119-129    11   Increment
# 130-141    12   Offset number 4
# 142-145     4   Reserved slots
# 146-148     3   Time-out
# 149-159    11   Increment

  $encode = "00";

# The $fatdma parameter should have 1-4 four-tuples of information

  @fields = split (",",$fatdma);
  for ($i=0; $i<@fields; $i+=4)
    {
    $encode .= sprintf ("%012b", int($fields [$i])) . sprintf ("%04b", int($fields [$i+1])) . sprintf ("%03b", int($fields [$i+2]))
    . sprintf ("%011b", int($fields [$i+3]));
    }
  return $encode;
  }

# *** Encode_21 builds a Type 21, Aid-to-Navigation Report ***
# 272-360 bits, occupying multiple AIS sentences.

sub encode_21
  {
# Parameter list
  my ($aid_type,$aid_name,$accuracy,$long,$lat,$vsize_a,$vsize_b,$vsize_c,$vsize_d,$epfd,$ts,$offpos,$aid_status,$raim,$v_aton) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-42      5   Aid type
#  43-162   120   Name
#  163        1   Position Accuracy
# 164-191    28   Longitude
# 192-218    27   Latitude
# 219-227     9   Dimension to Bow
# 228-236     9   Dimension to Stern
# 237-242     6   Dimension to Port
# 243-248     6   Dimension to Starboard
# 249-252     4   Type of EPFD
# 253-258     6   UTC seconds
# 259         1   Off-Position Indicator
# 260-267     8   AtonN Status
# 268         1   RAIM flag
# 269         1   Virtual-aid flag
# 270         1   Assigned                     ** Not currently supported; set to 0 = Autonomous mode (default)
# 271         1   Spare
# 272-360    88   Name Extension               ** Not currently supported

  $encode = sprintf ("%05b", int($aid_type)) . encode_string (pad_string ($aid_name, 20)) . sprintf ("%01b", int($accuracy))
  . sprintf ("%028b", (int($long*600000) & (2**28-1))) . sprintf ("%027b", (int($lat*600000) & (2**27-1))) . sprintf ("%09b", int ($vsize_a))
  . sprintf ("%09b", int ($vsize_b)) . sprintf ("%06b", int ($vsize_c)) . sprintf ("%06b", int ($vsize_d)) . sprintf ("%04b", int ($epfd))
  . sprintf ("%06b", int($ts)) . sprintf ("%01b", int($offpos)) . sprintf ("%08b", int($aid_status)) . sprintf ("%01b", int($raim)) . sprintf ("%1b", int ($v_aton)) . "0" . "0";
  return $encode;
  }

# *** Encode_22 builds a Type 22, Channel Management ***
# 168 bits, occupying one AIS sentence.

sub encode_22
  {
# Parameter list
  my ($channel_a, $channel_b, $ne_lon, $ne_lat, $sw_lon, $sw_lat) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-51     12   Channel A
#  52-63     12   Channel B
#  64-67      4   Tx/Rx mode                    ** Not currently supported; set to 0
#  68         1   Power                         ** Not currently supported; set to 0

#  If Addressed (bit 139) = 0
#  69-86     18   NE Longitude
#  87-103    17   NE Latitude
# 104-121    18   SW Longitude
# 122-138    17   SW Latitude

#  If Addressed (bit 139) = 1
#  69-98     30   MMSI_1                        ** Not currently supported
#  99-103     5   Spare                         ** Not currently supported
# 104-133    30   MMSI_2                        ** Not currently supported
# 134-138     5   Spare                         ** Not currently supported

# 139         1	  Addressed                     ** Not currently supported; set to 0
# 140         1   Channel A Band                ** Not currently supported; set to 0
# 141         1   Channel B Band                ** Not currently supported; set to 0
# 142-144     3   Zone size                     ** Not currently supported; set to 5
# 145-167    23   Spare

  $encode = "00" . sprintf ("%012b", $channel_a) . sprintf ("%012b", $channel_b) . "0000" . "0";

# As implemented, we only support an Addressed (bit 139) value of 0
# This break here is to make it obvious where to add if...then...else if $addressed becomes a user-specified variable

  $encode .= sprintf ("%018b", (int($ne_lon*600) & (2**18-1))) . sprintf ("%017b", (int($ne_lat*600) & (2**17-1)))
  . sprintf ("%018b", (int($sw_lon*600) & (2**18-1))) . sprintf ("%017b", (int($sw_lat*600) & (2**17-1)));

# Here is the Addressed bit... Obviously, this will be replaced with the $addressed value if that gets implemented

  $encode .= "0";

# Here is the rest of the sentence

  $encode .= "00" . "101" . "0"x23;

  return $encode;
  }

# *** Encode_23 builds a Type 23, Group Assignment Command ***
# 160 bits, occupying one AIS sentence.

sub encode_23
  {
# Parameter list
  my ($ne_lon, $ne_lat, $sw_lon, $sw_lat, $vtype, $interval, $quiet) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-57     18   NE Longitude
#  58-74     17   NE Latitude
#  75-92     18   SW Longitude
#  93-109    17   SW Latitude
# 110-113     4   Station Type                  ** Not currently supported; set to 0
# 114-121     8   Ship Type
# 122-143    22   Spare
# 144-145     2   Tx/Rx Mode                    ** Not currently supported; set to 0
# 146-149     4   Report Interval
# 150-153     4   Quiet Time
# 154-159     6   Spare

  $encode = "00". sprintf ("%018b", (int($ne_lon*600) & (2**18-1))) . sprintf ("%017b", (int($ne_lat*600) & (2**17-1)))
  . sprintf ("%018b", (int($sw_lon*600) & (2**18-1))) . sprintf ("%017b", (int($sw_lat*600) & (2**17-1))) . "0000"
  . sprintf ("%08b", int ($vtype)) . "0"x22 . "00". sprintf ("%04b", $interval) . sprintf ("%04b", $quiet) . "000000";

  return $encode;
  }

# *** Encode_24 builds a Type 24, Static Data Report ***
# Type 24 has two formats, part A (160 bits) and part B (168 bits), occupying one AIS sentence.
# Note that Format A is 168 bits by standard but is often only 160. Format B might also have two formats.

sub encode_24
  {
# Parameter list
  my ($mmsi, $part, $vname, $vtype, $vendor, $model, $serial, $callsign, $vsize_a, $vsize_b, $vsize_c, $vsize_d, $mothership) = @_;
  my ($part_b,$encode);

# PART A
################################################
#  Bits     Len   Description
#  38-39      2   Part Number (0=A,1=B)
#  40-159   120   Vessel Name (6-bit characters)
# 160-167     8   Spare

  if ($part eq "A")
      {
      $encode = "00" . encode_string (pad_string ($vname,20)) . "0"x8;
      return $encode;
      }
    else
      {

# PART B
################################################
#  Bits     Len   Description
#  38-39      2   Part Number (0=A,1=B)
#  40-47      8   Ship Type
#  48-65     18   Vendor ID (6-bit characters)
#  66-69      4   Unit Model Code
#  70-89     20   Serial Number
#  90-131    42   Call Sign (6-bit characters)

# If MMSI does not start with "98"
# 132-140     9   Dimension to Bow
# 141-149     9   Dimension to Stern
# 150-155     6   Dimension to Port
# 156-161     6   Dimension to Starboard

# If MMSI starts with "98"
# 132-161    30   Mothership MMSI (if MMSI starts with "98")
	
# 162-167     6   Spare
     
      $encode = "01" . sprintf ("%08b", int ($vtype)) . encode_string (pad_string ($vendor,3)) . sprintf ("%04b", int ($model))
      . sprintf ("%020b", int ($serial)) . encode_string (pad_string ($callsign,7));

# NOTE -- the next 30 bits are the dimensions of the ship... UNLESS the MMSI above indicates that this is an auxiliary craft (where the
# MMSI will start with a "98"); in that case, these 30 bits should be the MMSI of the mothership.

      if (substr ($mmsi,0,2) eq "98")
          {
          $encode .= sprintf ("%030b", int($mothership)) . "0"x6;
          return $encode;
          }
        else
          {
          $encode .= sprintf ("%09b", int ($vsize_a)) . sprintf ("%09b", int ($vsize_b)) . sprintf ("%06b", int ($vsize_c)) . sprintf ("%06b", int ($vsize_d)) . "0"x6;
          return $encode;
          }
      }
  }

# *** Encode_2526 builds Type 25: Single Slot Binary Message and Type 26: Multiple Slot Binary Message ***
# Type 25 messages can be up to 168 bits (one sentence) and Type 26 messages up to 1,064 bits (up to five sentences)

# There is also a special Type 26 - "R-Mode Navigation Message" (DAC = 1, FI = 10, 11, 14)

sub encode_2526
  {
# Parameter list
  my ($type,$target,$dac,$fi,$info68) = @_;
  my ($encode,$bin_len,$destination,$fields,$i,$max_data,$pad,$structured);

# TYPE 25
################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI
#  38         1   Destination Indicator
#  39         1   Binary Data flag
#
# Bits 38-39=00  Bits 38-39=01  Bits 38-39=10  Bits 38-39=11
#  Bits     Len   Bits     Len   Bits     Len   Bits     Len
#                                40-69     30   40-69     30  Destination MMSI
#                                70-71      2   70-71      2  Spare (00)
#                 40-49     10                  72-81     10  Application ID (DAC)
#                 50-55      6                  82-87      6  Application ID (FI)
#  40-167   128   56-167   112   72-167    96   88-167    80  Binary data (variable length, 0-128, octet-aligned)

# TYPE 26
################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI
#  38         1   Destination Indicator
#  39         1   Binary Data flag
#
# Bits 38-39=00  Bits 38-39=01  Bits 38-39=10  Bits 38-39=11
#  Bits     Len   Bits     Len   Bits     Len   Bits     Len
#                                40-69     30   40-69     30  Destination MMSI
#                                70-71      2   70-71      2  Spare (00)
#                 40-49     10                  72-81     10  Application ID (DAC)
#                 50-55      6                  82-87      6  Application ID (FI)
#  40-1039 1000   56-1039  984   72-1039  968   88-1039  952  Binary data (variable length, 0-1000, octet-aligned)
#  1040-1043  4   1040-1043  4   1040-1043  4   1040-1043  4  Spare
#  1044-1063 20   1044-1063 20   1044-1063 20   1044-1063 20  Radio Status   ** Not currently supported; set to 11100000000000000110

# TYPE 26 -- R-mode special message
################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI
#  38         1   Destination Indicator (0)
#  39         1   Binary Data flag (1)
#  40-49     10   Application Identifier (DAC)
#  50-55      6   Application Identifier (FI)
#  56-63      8   Station Health
#
#     FI=10          FI=11          FI=14
#  Bits     Len   Bits     Len   Bits     Len   Description
#  64-143    20   64-363   300   64-1031  968   Gold Code
# 144-147     4  364-371     8  1032-1043  12   Spare
# 148-167    20  372-391    20  1044-1063  20   Radio Status   ** Not currently supported; set to 00000000000000000000

# If no DAC and FI were provided in the command line, then DAC = 0, FI = 0 by default.
# $destination is the destination flag. If $target is null, $destination = 0 (broadcast), otherwise = 1 (addressed) and $target contains the destination MMSI
# $structured is the structured flag. If $dac or $fi != 0, then $structure = 1 (structured), otherwise = 1 (unstructured) [This is called Binary Data flag in R-mode messages]

  if ($type == 25) { $max_data = 128; }
    else { $max_data = 1000; }

# All messages have a common format (starting at bit 38) through the FI field

  $destination = 0; if ($target ne "") { $destination = 1; $max_data -= 32; }
  $structured = 0; if ($dac != 0 || $fi != 0) { $structured = 1; $max_data -= 16; }
  $encode = sprintf ("%01b", $destination) . sprintf ("%01b", $structured);
  
  if ($destination == 1) { $encode .= sprintf ("%030b", int($target)) . "00"; }
  if ($structured == 1) { $encode .= sprintf ("%010b", int($dac)) . sprintf ("%06b", int($fi)) ; }

# $info68 is a comma-delimited string where the first field is a T (for text messages [default]) or R (for R-mode hex messages).
# The second field is the message. Ensure that it is not too large....

  @fields = split (",", $info68);
  
  if ($fields [0] ne "R")
      {
      $encode .= encode_string (substr ($fields[1],0,$max_data));
      $pad = (8 - ((length ($encode)-2) % 8)) % 8;
      if ($pad > 0) { $encode .= "0" x $pad; }
      if ($type == 26) { $encode .= "0000" . "11100000000000000110"; }
      }
    else
      {
# We're here only for the R-mode messages, for space efficency, the R-mode data is stored in hex (complete with spare bits)

      $encode .= "0"x8;
      $bin_len = length ($fields [1]);
      for ($i=0; $i < $bin_len; $i++)
        {
        if (substr ($fields [1],$i,1) !~ /[0-9 A-Za-z.,-]/) { substr ($fields [1],$i,1,"0"); } # Replace any invalid hex digit with a 0x0
        $encode .= sprintf ("%04b", hex (substr ($fields [1],$i,1)));
        }
      $encode .= "0"x20;
      }

  return $encode;
  }

# *** Encode_27 builds Type 27: Long Range AIS Broadcast Messages ***
# Expected to be 96 bits, occupying one AIS sentence.

sub encode_27
  {
# Parameter list
  my ($accuracy,$raim,$navstat,$long,$lat,$speed27,$course27,$gnss_pos) = @_;
  my $encode;

################################################
#  Bits     Len   Description
#  38         1   Position Accuracy
#  39         1   RAIM flag
#  40-43      4   Navigation Status
#  44-61     18   Longitude
#  62-78     17   Latitude
#  79-84      6   Speed Over Ground (SOG)
#  85-93      9   Course Over Ground (COG)
#  94         1   GNSS Position status
#  95         1   Spare

  $encode = sprintf ("%01b", int($accuracy)) . sprintf ("%01b", int($raim)) . sprintf ("%04b", int($navstat)) . sprintf ("%018b", (int($long*600) & (2**18-1)))
  . sprintf ("%017b", (int($lat*600) & (2**17-1))) . sprintf ("%06b", int($speed27)) . sprintf ("%09b", int($course27+0.5))
  . sprintf ("%1b", int ($gnss_pos)) . "0";

  return $encode;
  }

# ********************************************************
# help
# ********************************************************

# Display the help information for the program

sub help
  {
print<< "EOT";

AIS_menu will create an AIS_ping command line that should only contain legal switches and values. The
user can change values by manipulating the AIS_ping command line.

For reference, the list below comprises all AIS_ping switches, default values (in square brackets), and
message types where the switch parameter is used. If the switch is not present and is required for the given
message type, AIS_ping will use the default value.

If the switch is a string with mutliple values, note the reference to check README.TXT.

Running AIS_menu will also give insights to the various parameters. For a complete understanding of the
possible values and their meanings, see the appropriate ITU or other specification.

EXAMPLE: AIS_ping.pl --type=1 --mmsi=338123456

--type             Message type (required)
                   1-3 = Position Report Class A
                     4 = Base Station Report
                     5 = Static and Voyage Related Data
                     6 = Binary Addressed Message
                     7 = Binary Acknowledge Message
                     8 = Binary Broadcast Message
                     9 = Standard SAR Aircraft Position Report
                    10 = UTC/Date Inquiry
                    11 = UTC/Date Response
                    12 = Addressed Safety-Related Message
                    13 = Safety-Related Acknowledgement
                    14 = Safety-Related Broadcast Message
                    15 = Interrogation Message
                    16 = Assignment Mode Command Message
                    17 = DGNSS Broadcast Binary Message
                    18 = Standard Class B CS Position Report
                    19 = Extended Class B CS Position Report
                    20 = Data Link Management Message
                    21 = Aid-to-Navigation Report
                    22 = Channel Management
                    23 = Group Assignment Command
                    24 = Static Data Report
                    25 = Single Slot Binary Message
                    26 = Multiple Slot Binary Message
                    27 = Long Range AIS Broadcast Message
--mmsi              All. MMSI (required)
--accuracy          1-3/4/8/9/11/18/19/21/27. Position accuracy (0-1) [0]
--course            1-3/9/18/19/27. Course over ground (0-360) [360]
--heading           1-3/18/19. True heading (0-359, 511) [511]
--lat               1-3/6/8/9/18/19/21/27. Latitude (-90:91) [91]
--long              1-3/6/8/9/18/19/21/27. Longitude (-180:181) [181]
--navstat           1-3/27. Navigation Status (0-15) [15]
--raim              1-3/9/11/18/19/21/27. RAIM flag (0-1) [0]
--rot               1-3. Rate of Turn (-128:127) [-128]
--speed             1-3/9/18/19. Speed over ground (0-102.3) [102.3]
--ts                1-3/9/18/19/21. Timestamp (0-63) [60]
--epfd              4/5/19/21. Type of Elec. Position Fixing Device (0-15) [0]
--year              4. UTC year (0-9999) [0]
--month             4/5/6. UTC month (0-12) [0]
--day               4/5/6/8. UTC day (0-31) [0]
--hour              4/5/6/8. UTC hour (0-24) [24]
--minute            4/5/6/8. UTC minute (0-60) [60]
--second            4. UTC second (0-60) [60]
--dest              5. Destination (1-20 chars) ["@@@...@@@"]
--draft             5. Draft ((0-25.5) [0]
--dte               5/19. Date Terminal Equipment state (0-1) [1]
--imo               5. IMO Number (0-1073741823) [0]
--callsign          5/24B. Call Sign (1-7 chars) ["@@@@@@@"]
--vsize_a           5/19/21/24B. AIS antenna to bow (A) (0-511) [0]
--vsize_b           5/19/21/24B. AIS antenna to stern (B) (0-511) [0]
--vsize_c           5/19/21/24B. AIS antenna to port (C) (0-63) [0]
--vsize_d           5/19/21/24B. AIS antenna to starboard (D) (0-63) [0]
--vtype             5/6/19/23/24B. Ship/cargo type (0-255) [0]
--vname             5/19/24A. Vessel Name (1-20 chars) ["@@@...@@@"]
--aid_name          6/21. Name of AtoN (1-34 chars) ["@@@...@@@"]
--aid_status        6/21. AtoN status (0-31)
--berth_dep         6. Berth depth (0-25.5) [0]
--berth_len         6. Berth length (0-511) [0]
--dac               6/8/25/26. Designated Area Code (required)
--fi                6/8/25/26. Functional ID (required)
--fid               6/8/25/26. Functional ID (OBSOLETE but still supported)
--duration          6/8. Duration of notice (0-262143) [262143]
--info68            6/8/25/26. Miscellaneous data for Types 6/8/25/26 messages [*** See README.TXT ***]
--mooring           6. Mooring position (0-5) [0]
--msglink           6/8. Message linkage identifier (0-1023) [0]
--offpos            6/21. Off-position indicator (0-1) [0]
--pob               6. Number of persons on board [*** See README.TXT ***]
--routetype         6/8. Route type (0-5, 31) [0]
--sa01..--sa10      6/8. Subarea N descriptors or vessel N descriptors [*** See README.TXT ***]
--sndrclass         6/8. Sender classification (0-1) [0]
--svc               6. Service availability (0-1) [0]
--target            6/7/10/12/13/15/16/25/26. Target MMSI [*** See README.TXT ***]
--tos               6. Type of service array; list of 22 digits -- each (0-3) [0]
--wmtype            6/8. Waterways Management message type/substype (0-15.0-2)
                         (i.e., 00, 01, 02, 10, 11, 12, ..., 150, 151, 152)
--airtemp           8. Air temperature (-60:60, -1024) [-1024]
--lockinfo          8. Lock and dam parameters [*** See README.TXT ***]
--ice               8. Ice status (0-3) [3]
--notice            8. Notice description (0-127) [127]
--precip            8. Precipitation type (1-7) [7]
--pressure          8. Air pressure (500-1500, 9999) [9999]
--presstend         8. Air pressure tendency (0-3) [3]
--seastate          8. Beaufort scale (0-13) [13]
--swelldir          8. Swell direction (0-360) [360]
--swellheight       8. Swell height (0-25.1, 25.5) [25.5]
--swellperiod       8. Swell period (0-60, 63) [63]
--vis               8. Horizontal visibility (0-12.7) [12.7]
--wavedir           8. Wave direction (0-360) [360]
--waveheight        8. Wave height (0-25.1, 25.5) [25.5]
--waveperiod        8. Wave period (0-60, 63) [63]
--wspeed            8. 10-min. avg. wind speed (0-127) [127]
--wgust             8. 10-min. max. wind speed (0-127) [127]
--wdir              8. Wind direction (0-360) [360]
--wgustdir          8. Wind gust direction (0-360) [360]
--altitude          9. Altitude (0-4095) [4095]
--sart_msg         12/14. SART alarm message (0-163 chars) ["SART ACTIVE"]
--dgnss_msg        17. M.823 DGNSS message [*** See README.TXT ***]
--fatdma           20. FATDMA parameters [*** See README.TXT ***]
 -aid_type         21. Type of AtoN (0-31) [0]
--v_AtoN           21. Virtual AtoN flag (0-1) [1]
--channel_a        22. Channel A frequency [2087] (87B = 161.975 MHz)
--channel_b        22. Channel B frequency [2088] (88B = 162.025 MHz)
--ne_lon           22/23. Long. of NE corner of jurisdiction area (-180:181) [181]
--ne_lat           22/23. Lat. of NE corner of jurisdiction area (-90:91) [91]
--sw_lon           22/23. Long. of SW corner of jurisdiction area (-180:181) [181]
--sw_lat           22/23. Lat. of SW corner of jurisdiction area (-90:91) [91]
--interval         23. Reporting interval (0-15) [1]
--quiet            23. Quiet interval (0-15) [0]
--part             24. Message type 24 part (0-1)
--model            24B. Unit model code (1-15)
--mothership       24B. Mothership MMSI (used if --mmsi starts with '98')
--serial           24B. Serial number (20 bits)
--vendor           24B. Vendor ID (1-7 chars) ["@@@@@@@"]
--course27         27. Course (0-359, 511) [511]
--gnss_pos         27. GNSS position status (0-1) [1]
--speed27          27. Speed (0-63) [63]

--batch            Batch mode; only the NMEA command gets displayed
--ais_sent         AIS sentence talker id. and encapsulation formatter (default = "!AIVDM")
--nmea             Create NMEA sentence on specified channel (default = A)
--tx               Create GnuRadio message on specified channel (default = A)
                   ** If both --nmea and --tx are present, use channel supplied with --nmea switch **
                   ** If neither --nmea nor --tx are present, use '--nmea=A'

EOT
  return;
  }
